
module Make

# 5 "src/parser/grammar.mly"
           ( C : Parser_config.ParseFile )
# 7 "src/parser/grammar.ml"

= struct
  
  module MenhirBasics = struct
    
    exception Error of int
    
    let _eRR =
      fun _s ->
        raise (Error _s)
    
    type token = Elpi_lexer_config.Tokens.token
    
  end
  
  include MenhirBasics
  
# 7 "src/parser/grammar.mly"
  
open Elpi_util
open Elpi_lexer_config

open Lexer_config
open Parser_config
open Ast
open Term

open TypeExpression

let loc (startpos, endpos) = {
  Util.Loc.client_payload = C.get_current_client_loc_payload ();
  source_name = startpos.Lexing.pos_fname;
  source_start = startpos.Lexing.pos_cnum;
  source_stop = endpos.Lexing.pos_cnum;
  line = startpos.Lexing.pos_lnum;
  line_starts_at = startpos.Lexing.pos_bol;
}

let desugar_multi_binder loc (t : Ast.Term.t) =
  match t.it with
  | App( { it = Const hd } as binder,args)
    when Func.equal hd Func.pif || Func.equal hd Func.sigmaf && List.length args > 1 ->
      let last, rev_rest = let l = List.rev args in List.hd l, List.tl l in
      let ty = match last.it with
        | Lam (_,ty,_) -> ty
        | Const x when Func.is_uvar_name x -> None
        | _ -> raise (ParseError(loc,"The last argument of 'pi' or 'sigma' must be a function or a unification variable, while it is: " ^ Ast.Term.show last)) in
      let names = List.map (function
        | { it = Const x; loc } -> Func.show x, loc
        | { it = (App _ | Lam _ | CData _ | Quoted _ | Cast _ | Parens _) } ->
            raise (ParseError(loc,"Only names are allowed after 'pi' or 'sigma'"))) rev_rest in
      let body = mkApp (Loc.merge binder.loc last.loc) [binder;last] in
      List.fold_left (fun bo (name,nloc) ->
        let loc = Loc.merge nloc bo.loc in
        mkApp loc [binder;mkLam loc name ty bo]) body names
  | (App _ | Const _ | Lam _ | CData _ | Quoted _ | Cast _ | Parens _) -> t
;;

let desugar_macro loc lhs rhs =
  match lhs, rhs with
  | { it = Const name }, body ->
      if ((Func.show name).[0] != '@') then
        raise (ParseError(loc,"Macro name must begin with '@'"));
      name, body
  | { it = App({ it = Const name },args) }, body ->
      if ((Func.show name).[0] != '@') then
        raise (ParseError(loc,"Macro name must begin with '@'"));
      let names = List.map (function
        | { it = Const x; loc } -> Func.show x, loc
        | { it = (App _ | Lam _ | CData _ | Quoted _ | Cast _ | Parens _) } ->
              raise (ParseError(loc,"Macro parameters must be names"))) args in
      name, List.fold_right (fun (name,nloc) b -> mkLam (Loc.merge nloc b.loc) name None b) names body
  | _ ->
        raise (ParseError(loc,"Illformed macro left hand side"))
;;

let mkParens_if_impl_or_conj loc t =
  match t.it with
  | App({ it = Const c},_) when Func.(equal c implf) -> mkParens loc t
  | App({ it = Const c},_) when Func.(equal c andf) -> mkParens loc t
  | _ -> t

let mkApp loc = function
  | { it = Const c; loc = cloc } :: a :: { it = App ({ it = Const c1 }, args) } :: [] when Func.(equal c andf && equal c1 andf) ->
      mkAppF loc (cloc,c) (a :: args)
  | l -> mkApp loc l

let rec unparen = function
  | [] -> []
  | { it = Parens { it = App ({ it = Const c1 }, args) } } as x :: xs when Func.(equal c1 implf) -> x :: unparen xs
  | { it = Parens x} :: xs -> x :: unparen xs
  | x :: xs -> x :: unparen xs

let mkAppF loc (cloc,c) l =
  if Func.(equal c implf) then
    match l with
    | { it = App ({ it = Const j; loc = jloc }, args) } :: rhs when Func.(equal j andf) ->
       begin match List.rev args with
       | last :: ( { loc = dloc } :: _ as rest_rev) ->
           let jloc = List.fold_left (fun x { loc } -> Loc.merge x loc) dloc rest_rev in
           let iloc = List.fold_left (fun x { loc } -> Loc.merge x loc) last.loc rhs in
           { it = App ({ it = Const j; loc = jloc },List.rev rest_rev @ [mkAppF iloc (cloc,c) (last :: rhs)]); loc = loc }
       | _ -> mkAppF loc (cloc,c) l
       end
    | _ -> mkAppF loc (cloc,c) (unparen l)
  else mkAppF loc (cloc,c) l

let binder l (loc,ty,b) =
  match List.rev l with
  | (name, bloc) :: rest ->
      let lloc = Loc.merge bloc b.loc in
      List.rev_map (fun (n,l) -> mkConst l n) rest @ [mkLam lloc (Func.show name) ty b]
  | _ -> raise (ParseError(loc,"bind '\\' operator must follow a name"))

let binder1 l = function
  | None -> l
  | Some (loc,ty,b) ->
      match List.rev l with
      | { it = Const name; loc = bloc } :: rest ->
        let lloc = Loc.merge bloc b.loc in
        List.rev rest @ [mkLam lloc (Func.show name) ty b]
      | _ -> raise (ParseError(loc,"bind '\\' operator must follow a name"))

;;

let underscore loc = { loc; it = Const Func.dummyname }

let decode_sequent loc t =
  match t.it with
  | App({ it = Const c },[hyps;bo]) when c == Func.sequentf -> hyps, bo
  | _ -> underscore loc, t

let fix_church x = if Func.show x = "o" then Func.propf else x

let mode_of_IO io =
  if io = 'i' then Mode.Input
  else if io = 'o' then Mode.Output
  else assert false


# 148 "src/parser/grammar.ml"
  
  type ('s, 'r) _menhir_state = 
    | MenhirState000 : ('s, _menhir_box_goal) _menhir_state
      (** State 000.
        Stack shape : .
        Start symbol: goal. *)

    | MenhirState003 : (('s, 'r) _menhir_cell1_SIGMA, 'r) _menhir_state
      (** State 003.
        Stack shape : SIGMA.
        Start symbol: <undetermined>. *)

    | MenhirState008 : (('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_state
      (** State 008.
        Stack shape : LPAREN.
        Start symbol: <undetermined>. *)

    | MenhirState057 : ((('s, 'r) _menhir_cell1_SIGMA, 'r) _menhir_cell1_nonempty_list_constant_w_loc_, 'r) _menhir_state
      (** State 057.
        Stack shape : SIGMA nonempty_list(constant_w_loc).
        Start symbol: <undetermined>. *)

    | MenhirState058 : ((('s, 'r) _menhir_cell1_nonempty_list_constant_w_loc_, 'r) _menhir_cell1_COLON, 'r) _menhir_state
      (** State 058.
        Stack shape : nonempty_list(constant_w_loc) COLON.
        Start symbol: <undetermined>. *)

    | MenhirState059 : (('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_state
      (** State 059.
        Stack shape : LPAREN.
        Start symbol: <undetermined>. *)

    | MenhirState060 : ((('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_cell1_PRED, 'r) _menhir_state
      (** State 060.
        Stack shape : LPAREN PRED.
        Start symbol: <undetermined>. *)

    | MenhirState061 : (('s, 'r) _menhir_cell1_IO_COLON, 'r) _menhir_state
      (** State 061.
        Stack shape : IO_COLON.
        Start symbol: <undetermined>. *)

    | MenhirState064 : (('s, 'r) _menhir_cell1_fotype_term _menhir_cell0_ARROW, 'r) _menhir_state
      (** State 064.
        Stack shape : fotype_term ARROW.
        Start symbol: <undetermined>. *)

    | MenhirState066 : (('s, 'r) _menhir_cell1_constant, 'r) _menhir_state
      (** State 066.
        Stack shape : constant.
        Start symbol: <undetermined>. *)

    | MenhirState067 : (('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_state
      (** State 067.
        Stack shape : LPAREN.
        Start symbol: <undetermined>. *)

    | MenhirState068 : ((('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_cell1_FUNC, 'r) _menhir_state
      (** State 068.
        Stack shape : LPAREN FUNC.
        Start symbol: <undetermined>. *)

    | MenhirState071 : (((('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_cell1_FUNC, 'r) _menhir_cell1_loption_separated_nonempty_list_CONJ_fotype_term__ _menhir_cell0_ARROW, 'r) _menhir_state
      (** State 071.
        Stack shape : LPAREN FUNC loption(separated_nonempty_list(CONJ,fotype_term)) ARROW.
        Start symbol: <undetermined>. *)

    | MenhirState074 : (('s, 'r) _menhir_cell1_fotype_term _menhir_cell0_CONJ, 'r) _menhir_state
      (** State 074.
        Stack shape : fotype_term CONJ.
        Start symbol: <undetermined>. *)

    | MenhirState082 : (('s, 'r) _menhir_cell1_atype_term, 'r) _menhir_state
      (** State 082.
        Stack shape : atype_term.
        Start symbol: <undetermined>. *)

    | MenhirState087 : (('s, 'r) _menhir_cell1_pred_item _menhir_cell0_option_CONJ_, 'r) _menhir_state
      (** State 087.
        Stack shape : pred_item option(CONJ).
        Start symbol: <undetermined>. *)

    | MenhirState095 : (((('s, 'r) _menhir_cell1_nonempty_list_constant_w_loc_, 'r) _menhir_cell1_COLON, 'r) _menhir_cell1_type_term _menhir_cell0_BIND, 'r) _menhir_state
      (** State 095.
        Stack shape : nonempty_list(constant_w_loc) COLON type_term BIND.
        Start symbol: <undetermined>. *)

    | MenhirState097 : (('s, 'r) _menhir_cell1_PI, 'r) _menhir_state
      (** State 097.
        Stack shape : PI.
        Start symbol: <undetermined>. *)

    | MenhirState098 : ((('s, 'r) _menhir_cell1_PI, 'r) _menhir_cell1_nonempty_list_constant_w_loc_, 'r) _menhir_state
      (** State 098.
        Stack shape : PI nonempty_list(constant_w_loc).
        Start symbol: <undetermined>. *)

    | MenhirState099 : ((('s, 'r) _menhir_cell1_nonempty_list_constant_w_loc_, 'r) _menhir_cell1_BIND, 'r) _menhir_state
      (** State 099.
        Stack shape : nonempty_list(constant_w_loc) BIND.
        Start symbol: <undetermined>. *)

    | MenhirState100 : (('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_state
      (** State 100.
        Stack shape : LPAREN.
        Start symbol: <undetermined>. *)

    | MenhirState101 : ((('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_cell1_SIGMA, 'r) _menhir_state
      (** State 101.
        Stack shape : LPAREN SIGMA.
        Start symbol: <undetermined>. *)

    | MenhirState102 : (('s, 'r) _menhir_cell1_constant_w_loc, 'r) _menhir_state
      (** State 102.
        Stack shape : constant_w_loc.
        Start symbol: <undetermined>. *)

    | MenhirState105 : ((('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_cell1_PI, 'r) _menhir_state
      (** State 105.
        Stack shape : LPAREN PI.
        Start symbol: <undetermined>. *)

    | MenhirState106 : (('s, 'r) _menhir_cell1_LCURLY, 'r) _menhir_state
      (** State 106.
        Stack shape : LCURLY.
        Start symbol: <undetermined>. *)

    | MenhirState107 : (('s, 'r) _menhir_cell1_LBRACKET, 'r) _menhir_state
      (** State 107.
        Stack shape : LBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState108 : (('s, 'r) _menhir_cell1_SIGMA, 'r) _menhir_state
      (** State 108.
        Stack shape : SIGMA.
        Start symbol: <undetermined>. *)

    | MenhirState109 : ((('s, 'r) _menhir_cell1_SIGMA, 'r) _menhir_cell1_nonempty_list_constant_w_loc_, 'r) _menhir_state
      (** State 109.
        Stack shape : SIGMA nonempty_list(constant_w_loc).
        Start symbol: <undetermined>. *)

    | MenhirState112 : (('s, 'r) _menhir_cell1_PIPE, 'r) _menhir_state
      (** State 112.
        Stack shape : PIPE.
        Start symbol: <undetermined>. *)

    | MenhirState113 : (('s, 'r) _menhir_cell1_PI, 'r) _menhir_state
      (** State 113.
        Stack shape : PI.
        Start symbol: <undetermined>. *)

    | MenhirState114 : ((('s, 'r) _menhir_cell1_PI, 'r) _menhir_cell1_nonempty_list_constant_w_loc_, 'r) _menhir_state
      (** State 114.
        Stack shape : PI nonempty_list(constant_w_loc).
        Start symbol: <undetermined>. *)

    | MenhirState118 : (('s, 'r) _menhir_cell1_FAMILY_TILDE, 'r) _menhir_state
      (** State 118.
        Stack shape : FAMILY_TILDE.
        Start symbol: <undetermined>. *)

    | MenhirState122 : (('s, 'r) _menhir_cell1_head_term, 'r) _menhir_state
      (** State 122.
        Stack shape : head_term.
        Start symbol: <undetermined>. *)

    | MenhirState123 : ((('s, 'r) _menhir_cell1_head_term, 'r) _menhir_cell1_nonempty_list_closed_term_, 'r) _menhir_state
      (** State 123.
        Stack shape : head_term nonempty_list(closed_term).
        Start symbol: <undetermined>. *)

    | MenhirState124 : (((('s, 'r) _menhir_cell1_head_term, 'r) _menhir_cell1_nonempty_list_closed_term_, 'r) _menhir_cell1_BIND, 'r) _menhir_state
      (** State 124.
        Stack shape : head_term nonempty_list(closed_term) BIND.
        Start symbol: <undetermined>. *)

    | MenhirState125 : (('s, 'r) _menhir_cell1_FAMILY_TILDE, 'r) _menhir_state
      (** State 125.
        Stack shape : FAMILY_TILDE.
        Start symbol: <undetermined>. *)

    | MenhirState129 : (('s, 'r) _menhir_cell1_head_term, 'r) _menhir_state
      (** State 129.
        Stack shape : head_term.
        Start symbol: <undetermined>. *)

    | MenhirState130 : ((('s, 'r) _menhir_cell1_head_term, 'r) _menhir_cell1_nonempty_list_closed_term_, 'r) _menhir_state
      (** State 130.
        Stack shape : head_term nonempty_list(closed_term).
        Start symbol: <undetermined>. *)

    | MenhirState135 : (('s, 'r) _menhir_cell1_closed_term, 'r) _menhir_state
      (** State 135.
        Stack shape : closed_term.
        Start symbol: <undetermined>. *)

    | MenhirState138 : (('s, 'r) _menhir_cell1_constant _menhir_cell0_COLON, 'r) _menhir_state
      (** State 138.
        Stack shape : constant COLON.
        Start symbol: <undetermined>. *)

    | MenhirState140 : ((('s, 'r) _menhir_cell1_constant _menhir_cell0_COLON, 'r) _menhir_cell1_type_term _menhir_cell0_BIND, 'r) _menhir_state
      (** State 140.
        Stack shape : constant COLON type_term BIND.
        Start symbol: <undetermined>. *)

    | MenhirState142 : (('s, 'r) _menhir_cell1_term _menhir_cell0_VDASH, 'r) _menhir_state
      (** State 142.
        Stack shape : term VDASH.
        Start symbol: <undetermined>. *)

    | MenhirState144 : (('s, 'r) _menhir_cell1_term _menhir_cell0_SLASH, 'r) _menhir_state
      (** State 144.
        Stack shape : term SLASH.
        Start symbol: <undetermined>. *)

    | MenhirState146 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_SHARP, 'r) _menhir_state
      (** State 146.
        Stack shape : term FAMILY_SHARP.
        Start symbol: <undetermined>. *)

    | MenhirState150 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_OR, 'r) _menhir_state
      (** State 150.
        Stack shape : term FAMILY_OR.
        Start symbol: <undetermined>. *)

    | MenhirState152 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_AND, 'r) _menhir_state
      (** State 152.
        Stack shape : term FAMILY_AND.
        Start symbol: <undetermined>. *)

    | MenhirState154 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_MINUS, 'r) _menhir_state
      (** State 154.
        Stack shape : term FAMILY_MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState156 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_EQ, 'r) _menhir_state
      (** State 156.
        Stack shape : term FAMILY_EQ.
        Start symbol: <undetermined>. *)

    | MenhirState158 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_BTICK, 'r) _menhir_state
      (** State 158.
        Stack shape : term FAMILY_BTICK.
        Start symbol: <undetermined>. *)

    | MenhirState160 : (('s, 'r) _menhir_cell1_term _menhir_cell0_OR, 'r) _menhir_state
      (** State 160.
        Stack shape : term OR.
        Start symbol: <undetermined>. *)

    | MenhirState162 : (('s, 'r) _menhir_cell1_term _menhir_cell0_MOD, 'r) _menhir_state
      (** State 162.
        Stack shape : term MOD.
        Start symbol: <undetermined>. *)

    | MenhirState164 : (('s, 'r) _menhir_cell1_term _menhir_cell0_MINUSs, 'r) _menhir_state
      (** State 164.
        Stack shape : term MINUSs.
        Start symbol: <undetermined>. *)

    | MenhirState166 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_TIMES, 'r) _menhir_state
      (** State 166.
        Stack shape : term FAMILY_TIMES.
        Start symbol: <undetermined>. *)

    | MenhirState168 : (('s, 'r) _menhir_cell1_term _menhir_cell0_DIV, 'r) _menhir_state
      (** State 168.
        Stack shape : term DIV.
        Start symbol: <undetermined>. *)

    | MenhirState170 : (('s, 'r) _menhir_cell1_term _menhir_cell0_MINUSr, 'r) _menhir_state
      (** State 170.
        Stack shape : term MINUSr.
        Start symbol: <undetermined>. *)

    | MenhirState172 : (('s, 'r) _menhir_cell1_term _menhir_cell0_MINUSi, 'r) _menhir_state
      (** State 172.
        Stack shape : term MINUSi.
        Start symbol: <undetermined>. *)

    | MenhirState174 : (('s, 'r) _menhir_cell1_term _menhir_cell0_MINUS, 'r) _menhir_state
      (** State 174.
        Stack shape : term MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState176 : (('s, 'r) _menhir_cell1_term _menhir_cell0_IS, 'r) _menhir_state
      (** State 176.
        Stack shape : term IS.
        Start symbol: <undetermined>. *)

    | MenhirState178 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_TICK, 'r) _menhir_state
      (** State 178.
        Stack shape : term FAMILY_TICK.
        Start symbol: <undetermined>. *)

    | MenhirState180 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_PLUS, 'r) _menhir_state
      (** State 180.
        Stack shape : term FAMILY_PLUS.
        Start symbol: <undetermined>. *)

    | MenhirState182 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_EXP, 'r) _menhir_state
      (** State 182.
        Stack shape : term FAMILY_EXP.
        Start symbol: <undetermined>. *)

    | MenhirState184 : (('s, 'r) _menhir_cell1_term _menhir_cell0_CONS, 'r) _menhir_state
      (** State 184.
        Stack shape : term CONS.
        Start symbol: <undetermined>. *)

    | MenhirState186 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_LT, 'r) _menhir_state
      (** State 186.
        Stack shape : term FAMILY_LT.
        Start symbol: <undetermined>. *)

    | MenhirState188 : (('s, 'r) _menhir_cell1_term _menhir_cell0_FAMILY_GT, 'r) _menhir_state
      (** State 188.
        Stack shape : term FAMILY_GT.
        Start symbol: <undetermined>. *)

    | MenhirState190 : (('s, 'r) _menhir_cell1_term _menhir_cell0_EQ2, 'r) _menhir_state
      (** State 190.
        Stack shape : term EQ2.
        Start symbol: <undetermined>. *)

    | MenhirState192 : (('s, 'r) _menhir_cell1_term _menhir_cell0_EQ, 'r) _menhir_state
      (** State 192.
        Stack shape : term EQ.
        Start symbol: <undetermined>. *)

    | MenhirState194 : (('s, 'r) _menhir_cell1_term _menhir_cell0_DDARROW, 'r) _menhir_state
      (** State 194.
        Stack shape : term DDARROW.
        Start symbol: <undetermined>. *)

    | MenhirState196 : (('s, 'r) _menhir_cell1_term _menhir_cell0_DARROW, 'r) _menhir_state
      (** State 196.
        Stack shape : term DARROW.
        Start symbol: <undetermined>. *)

    | MenhirState198 : (('s, 'r) _menhir_cell1_term _menhir_cell0_CONJ2, 'r) _menhir_state
      (** State 198.
        Stack shape : term CONJ2.
        Start symbol: <undetermined>. *)

    | MenhirState200 : (('s, 'r) _menhir_cell1_term _menhir_cell0_CONJ, 'r) _menhir_state
      (** State 200.
        Stack shape : term CONJ.
        Start symbol: <undetermined>. *)

    | MenhirState202 : (('s, 'r) _menhir_cell1_term _menhir_cell0_ARROW, 'r) _menhir_state
      (** State 202.
        Stack shape : term ARROW.
        Start symbol: <undetermined>. *)

    | MenhirState204 : (('s, 'r) _menhir_cell1_term _menhir_cell0_QDASH, 'r) _menhir_state
      (** State 204.
        Stack shape : term QDASH.
        Start symbol: <undetermined>. *)

    | MenhirState206 : (('s, 'r) _menhir_cell1_constant _menhir_cell0_BIND, 'r) _menhir_state
      (** State 206.
        Stack shape : constant BIND.
        Start symbol: <undetermined>. *)

    | MenhirState211 : (('s, 'r) _menhir_cell1_constant _menhir_cell0_COLON, 'r) _menhir_state
      (** State 211.
        Stack shape : constant COLON.
        Start symbol: <undetermined>. *)

    | MenhirState213 : ((('s, 'r) _menhir_cell1_constant _menhir_cell0_COLON, 'r) _menhir_cell1_type_term _menhir_cell0_BIND, 'r) _menhir_state
      (** State 213.
        Stack shape : constant COLON type_term BIND.
        Start symbol: <undetermined>. *)

    | MenhirState215 : (('s, 'r) _menhir_cell1_constant _menhir_cell0_BIND, 'r) _menhir_state
      (** State 215.
        Stack shape : constant BIND.
        Start symbol: <undetermined>. *)

    | MenhirState220 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_VDASH, 'r) _menhir_state
      (** State 220.
        Stack shape : term_noconj VDASH.
        Start symbol: <undetermined>. *)

    | MenhirState222 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_SLASH, 'r) _menhir_state
      (** State 222.
        Stack shape : term_noconj SLASH.
        Start symbol: <undetermined>. *)

    | MenhirState224 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_SHARP, 'r) _menhir_state
      (** State 224.
        Stack shape : term_noconj FAMILY_SHARP.
        Start symbol: <undetermined>. *)

    | MenhirState226 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_OR, 'r) _menhir_state
      (** State 226.
        Stack shape : term_noconj FAMILY_OR.
        Start symbol: <undetermined>. *)

    | MenhirState228 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_AND, 'r) _menhir_state
      (** State 228.
        Stack shape : term_noconj FAMILY_AND.
        Start symbol: <undetermined>. *)

    | MenhirState230 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_MINUS, 'r) _menhir_state
      (** State 230.
        Stack shape : term_noconj FAMILY_MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState232 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_EQ, 'r) _menhir_state
      (** State 232.
        Stack shape : term_noconj FAMILY_EQ.
        Start symbol: <undetermined>. *)

    | MenhirState234 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_BTICK, 'r) _menhir_state
      (** State 234.
        Stack shape : term_noconj FAMILY_BTICK.
        Start symbol: <undetermined>. *)

    | MenhirState236 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_OR, 'r) _menhir_state
      (** State 236.
        Stack shape : term_noconj OR.
        Start symbol: <undetermined>. *)

    | MenhirState238 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_MOD, 'r) _menhir_state
      (** State 238.
        Stack shape : term_noconj MOD.
        Start symbol: <undetermined>. *)

    | MenhirState240 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_MINUSs, 'r) _menhir_state
      (** State 240.
        Stack shape : term_noconj MINUSs.
        Start symbol: <undetermined>. *)

    | MenhirState242 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_TIMES, 'r) _menhir_state
      (** State 242.
        Stack shape : term_noconj FAMILY_TIMES.
        Start symbol: <undetermined>. *)

    | MenhirState244 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_DIV, 'r) _menhir_state
      (** State 244.
        Stack shape : term_noconj DIV.
        Start symbol: <undetermined>. *)

    | MenhirState246 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_MINUSr, 'r) _menhir_state
      (** State 246.
        Stack shape : term_noconj MINUSr.
        Start symbol: <undetermined>. *)

    | MenhirState248 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_MINUSi, 'r) _menhir_state
      (** State 248.
        Stack shape : term_noconj MINUSi.
        Start symbol: <undetermined>. *)

    | MenhirState250 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_MINUS, 'r) _menhir_state
      (** State 250.
        Stack shape : term_noconj MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState252 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_IS, 'r) _menhir_state
      (** State 252.
        Stack shape : term_noconj IS.
        Start symbol: <undetermined>. *)

    | MenhirState254 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_TICK, 'r) _menhir_state
      (** State 254.
        Stack shape : term_noconj FAMILY_TICK.
        Start symbol: <undetermined>. *)

    | MenhirState256 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_PLUS, 'r) _menhir_state
      (** State 256.
        Stack shape : term_noconj FAMILY_PLUS.
        Start symbol: <undetermined>. *)

    | MenhirState258 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_EXP, 'r) _menhir_state
      (** State 258.
        Stack shape : term_noconj FAMILY_EXP.
        Start symbol: <undetermined>. *)

    | MenhirState260 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_CONS, 'r) _menhir_state
      (** State 260.
        Stack shape : term_noconj CONS.
        Start symbol: <undetermined>. *)

    | MenhirState262 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_LT, 'r) _menhir_state
      (** State 262.
        Stack shape : term_noconj FAMILY_LT.
        Start symbol: <undetermined>. *)

    | MenhirState264 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_GT, 'r) _menhir_state
      (** State 264.
        Stack shape : term_noconj FAMILY_GT.
        Start symbol: <undetermined>. *)

    | MenhirState266 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_EQ2, 'r) _menhir_state
      (** State 266.
        Stack shape : term_noconj EQ2.
        Start symbol: <undetermined>. *)

    | MenhirState268 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_EQ, 'r) _menhir_state
      (** State 268.
        Stack shape : term_noconj EQ.
        Start symbol: <undetermined>. *)

    | MenhirState270 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_DDARROW, 'r) _menhir_state
      (** State 270.
        Stack shape : term_noconj DDARROW.
        Start symbol: <undetermined>. *)

    | MenhirState272 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_DARROW, 'r) _menhir_state
      (** State 272.
        Stack shape : term_noconj DARROW.
        Start symbol: <undetermined>. *)

    | MenhirState274 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_CONJ2, 'r) _menhir_state
      (** State 274.
        Stack shape : term_noconj CONJ2.
        Start symbol: <undetermined>. *)

    | MenhirState276 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_ARROW, 'r) _menhir_state
      (** State 276.
        Stack shape : term_noconj ARROW.
        Start symbol: <undetermined>. *)

    | MenhirState279 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_QDASH, 'r) _menhir_state
      (** State 279.
        Stack shape : term_noconj QDASH.
        Start symbol: <undetermined>. *)

    | MenhirState283 : (('s, 'r) _menhir_cell1_term_noconj, 'r) _menhir_state
      (** State 283.
        Stack shape : term_noconj.
        Start symbol: <undetermined>. *)

    | MenhirState286 : (('s, 'r) _menhir_cell1_term_noconj _menhir_cell0_CONJ, 'r) _menhir_state
      (** State 286.
        Stack shape : term_noconj CONJ.
        Start symbol: <undetermined>. *)

    | MenhirState293 : ((('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_cell1_FAMILY_TILDE, 'r) _menhir_state
      (** State 293.
        Stack shape : LPAREN FAMILY_TILDE.
        Start symbol: <undetermined>. *)

    | MenhirState296 : ((('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_cell1_term _menhir_cell0_COLON, 'r) _menhir_state
      (** State 296.
        Stack shape : LPAREN term COLON.
        Start symbol: <undetermined>. *)

    | MenhirState299 : ((('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_cell1_term _menhir_cell0_AS, 'r) _menhir_state
      (** State 299.
        Stack shape : LPAREN term AS.
        Start symbol: <undetermined>. *)

    | MenhirState303 : ((('s, 'r) _menhir_cell1_LPAREN, 'r) _menhir_cell1_constant _menhir_cell0_COLON, 'r) _menhir_state
      (** State 303.
        Stack shape : LPAREN constant COLON.
        Start symbol: <undetermined>. *)

    | MenhirState353 : ('s, _menhir_box_program) _menhir_state
      (** State 353.
        Stack shape : .
        Start symbol: program. *)

    | MenhirState355 : (('s, _menhir_box_program) _menhir_cell1_USEONLY, _menhir_box_program) _menhir_state
      (** State 355.
        Stack shape : USEONLY.
        Start symbol: program. *)

    | MenhirState356 : ((('s, _menhir_box_program) _menhir_cell1_USEONLY, _menhir_box_program) _menhir_cell1_separated_nonempty_list_CONJ_constant_, _menhir_box_program) _menhir_state
      (** State 356.
        Stack shape : USEONLY separated_nonempty_list(CONJ,constant).
        Start symbol: program. *)

    | MenhirState359 : (('s, _menhir_box_program) _menhir_cell1_constant _menhir_cell0_CONJ, _menhir_box_program) _menhir_state
      (** State 359.
        Stack shape : constant CONJ.
        Start symbol: program. *)

    | MenhirState361 : (('s, _menhir_box_program) _menhir_cell1_TYPEABBREV, _menhir_box_program) _menhir_state
      (** State 361.
        Stack shape : TYPEABBREV.
        Start symbol: program. *)

    | MenhirState362 : ((('s, _menhir_box_program) _menhir_cell1_TYPEABBREV, _menhir_box_program) _menhir_cell1_LPAREN, _menhir_box_program) _menhir_state
      (** State 362.
        Stack shape : TYPEABBREV LPAREN.
        Start symbol: program. *)

    | MenhirState363 : (((('s, _menhir_box_program) _menhir_cell1_TYPEABBREV, _menhir_box_program) _menhir_cell1_LPAREN, _menhir_box_program) _menhir_cell1_constant, _menhir_box_program) _menhir_state
      (** State 363.
        Stack shape : TYPEABBREV LPAREN constant.
        Start symbol: program. *)

    | MenhirState367 : ((('s, _menhir_box_program) _menhir_cell1_TYPEABBREV, _menhir_box_program) _menhir_cell1_abbrevform, _menhir_box_program) _menhir_state
      (** State 367.
        Stack shape : TYPEABBREV abbrevform.
        Start symbol: program. *)

    | MenhirState369 : (('s, _menhir_box_program) _menhir_cell1_SIG, _menhir_box_program) _menhir_state
      (** State 369.
        Stack shape : SIG.
        Start symbol: program. *)

    | MenhirState371 : (('s, _menhir_box_program) _menhir_cell1_SHORTEN, _menhir_box_program) _menhir_state
      (** State 371.
        Stack shape : SHORTEN.
        Start symbol: program. *)

    | MenhirState377 : ((('s, _menhir_box_program) _menhir_cell1_SHORTEN, _menhir_box_program) _menhir_cell1_constant _menhir_cell0_FULLSTOP _menhir_cell0_LCURLY, _menhir_box_program) _menhir_state
      (** State 377.
        Stack shape : SHORTEN constant FULLSTOP LCURLY.
        Start symbol: program. *)

    | MenhirState379 : (('s, _menhir_box_program) _menhir_cell1_subtrie _menhir_cell0_CONJ, _menhir_box_program) _menhir_state
      (** State 379.
        Stack shape : subtrie CONJ.
        Start symbol: program. *)

    | MenhirState383 : (('s, _menhir_box_program) _menhir_cell1_constant _menhir_cell0_FULLSTOP _menhir_cell0_LCURLY, _menhir_box_program) _menhir_state
      (** State 383.
        Stack shape : constant FULLSTOP LCURLY.
        Start symbol: program. *)

    | MenhirState389 : (('s, _menhir_box_program) _menhir_cell1_NAMESPACE, _menhir_box_program) _menhir_state
      (** State 389.
        Stack shape : NAMESPACE.
        Start symbol: program. *)

    | MenhirState392 : (('s, _menhir_box_program) _menhir_cell1_MODULE, _menhir_box_program) _menhir_state
      (** State 392.
        Stack shape : MODULE.
        Start symbol: program. *)

    | MenhirState394 : (('s, _menhir_box_program) _menhir_cell1_MACRO, _menhir_box_program) _menhir_state
      (** State 394.
        Stack shape : MACRO.
        Start symbol: program. *)

    | MenhirState396 : ((('s, _menhir_box_program) _menhir_cell1_MACRO, _menhir_box_program) _menhir_cell1_term _menhir_cell0_VDASH, _menhir_box_program) _menhir_state
      (** State 396.
        Stack shape : MACRO term VDASH.
        Start symbol: program. *)

    | MenhirState398 : (('s, _menhir_box_program) _menhir_cell1_LOCALKIND, _menhir_box_program) _menhir_state
      (** State 398.
        Stack shape : LOCALKIND.
        Start symbol: program. *)

    | MenhirState400 : (('s, _menhir_box_program) _menhir_cell1_LOCAL, _menhir_box_program) _menhir_state
      (** State 400.
        Stack shape : LOCAL.
        Start symbol: program. *)

    | MenhirState401 : ((('s, _menhir_box_program) _menhir_cell1_LOCAL, _menhir_box_program) _menhir_cell1_separated_nonempty_list_CONJ_constant_, _menhir_box_program) _menhir_state
      (** State 401.
        Stack shape : LOCAL separated_nonempty_list(CONJ,constant).
        Start symbol: program. *)

    | MenhirState406 : (('s, _menhir_box_program) _menhir_cell1_KIND, _menhir_box_program) _menhir_state
      (** State 406.
        Stack shape : KIND.
        Start symbol: program. *)

    | MenhirState407 : ((('s, _menhir_box_program) _menhir_cell1_KIND, _menhir_box_program) _menhir_cell1_separated_nonempty_list_CONJ_constant_, _menhir_box_program) _menhir_state
      (** State 407.
        Stack shape : KIND separated_nonempty_list(CONJ,constant).
        Start symbol: program. *)

    | MenhirState409 : (('s, _menhir_box_program) _menhir_cell1_TYPE _menhir_cell0_ARROW, _menhir_box_program) _menhir_state
      (** State 409.
        Stack shape : TYPE ARROW.
        Start symbol: program. *)

    | MenhirState413 : (('s, _menhir_box_program) _menhir_cell1_FIXITY, _menhir_box_program) _menhir_state
      (** State 413.
        Stack shape : FIXITY.
        Start symbol: program. *)

    | MenhirState419 : (('s, _menhir_box_program) _menhir_cell1_EXPORTDEF, _menhir_box_program) _menhir_state
      (** State 419.
        Stack shape : EXPORTDEF.
        Start symbol: program. *)

    | MenhirState420 : ((('s, _menhir_box_program) _menhir_cell1_EXPORTDEF, _menhir_box_program) _menhir_cell1_separated_nonempty_list_CONJ_constant_, _menhir_box_program) _menhir_state
      (** State 420.
        Stack shape : EXPORTDEF separated_nonempty_list(CONJ,constant).
        Start symbol: program. *)

    | MenhirState423 : (('s, _menhir_box_program) _menhir_cell1_CONSTRAINT, _menhir_box_program) _menhir_state
      (** State 423.
        Stack shape : CONSTRAINT.
        Start symbol: program. *)

    | MenhirState425 : ((('s, _menhir_box_program) _menhir_cell1_CONSTRAINT, _menhir_box_program) _menhir_cell1_list_constant_ _menhir_cell0_QDASH, _menhir_box_program) _menhir_state
      (** State 425.
        Stack shape : CONSTRAINT list(constant) QDASH.
        Start symbol: program. *)

    | MenhirState428 : (('s, _menhir_box_program) _menhir_cell1_constant, _menhir_box_program) _menhir_state
      (** State 428.
        Stack shape : constant.
        Start symbol: program. *)

    | MenhirState431 : (('s, _menhir_box_program) _menhir_cell1_COLON, _menhir_box_program) _menhir_state
      (** State 431.
        Stack shape : COLON.
        Start symbol: program. *)

    | MenhirState440 : (('s, _menhir_box_program) _menhir_cell1_INDEX _menhir_cell0_LPAREN, _menhir_box_program) _menhir_state
      (** State 440.
        Stack shape : INDEX LPAREN.
        Start symbol: program. *)

    | MenhirState447 : (('s, _menhir_box_program) _menhir_cell1_indexing, _menhir_box_program) _menhir_state
      (** State 447.
        Stack shape : indexing.
        Start symbol: program. *)

    | MenhirState459 : (('s, _menhir_box_program) _menhir_cell1_attribute _menhir_cell0_COLON, _menhir_box_program) _menhir_state
      (** State 459.
        Stack shape : attribute COLON.
        Start symbol: program. *)

    | MenhirState461 : (('s, _menhir_box_program) _menhir_cell1_CLOSED, _menhir_box_program) _menhir_state
      (** State 461.
        Stack shape : CLOSED.
        Start symbol: program. *)

    | MenhirState480 : (('s, _menhir_box_program) _menhir_cell1_decl, _menhir_box_program) _menhir_state
      (** State 480.
        Stack shape : decl.
        Start symbol: program. *)

    | MenhirState486 : (('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_state
      (** State 486.
        Stack shape : attributes.
        Start symbol: program. *)

    | MenhirState487 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_TYPE, _menhir_box_program) _menhir_state
      (** State 487.
        Stack shape : attributes TYPE.
        Start symbol: program. *)

    | MenhirState488 : (((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_TYPE, _menhir_box_program) _menhir_cell1_separated_nonempty_list_CONJ_constant_, _menhir_box_program) _menhir_state
      (** State 488.
        Stack shape : attributes TYPE separated_nonempty_list(CONJ,constant).
        Start symbol: program. *)

    | MenhirState490 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_SIGMA, _menhir_box_program) _menhir_state
      (** State 490.
        Stack shape : attributes SIGMA.
        Start symbol: program. *)

    | MenhirState491 : (((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_SIGMA, _menhir_box_program) _menhir_cell1_nonempty_list_constant_w_loc_, _menhir_box_program) _menhir_state
      (** State 491.
        Stack shape : attributes SIGMA nonempty_list(constant_w_loc).
        Start symbol: program. *)

    | MenhirState493 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_RULE, _menhir_box_program) _menhir_state
      (** State 493.
        Stack shape : attributes RULE.
        Start symbol: program. *)

    | MenhirState494 : (('s, _menhir_box_program) _menhir_cell1_LPAREN, _menhir_box_program) _menhir_state
      (** State 494.
        Stack shape : LPAREN.
        Start symbol: program. *)

    | MenhirState497 : ((('s, _menhir_box_program) _menhir_cell1_LPAREN, _menhir_box_program) _menhir_cell1_closed_term, _menhir_box_program) _menhir_state
      (** State 497.
        Stack shape : LPAREN closed_term.
        Start symbol: program. *)

    | MenhirState500 : (('s, _menhir_box_program) _menhir_cell1_sequent, _menhir_box_program) _menhir_state
      (** State 500.
        Stack shape : sequent.
        Start symbol: program. *)

    | MenhirState504 : (((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_RULE, _menhir_box_program) _menhir_cell1_list_sequent_ _menhir_cell0_BIND, _menhir_box_program) _menhir_state
      (** State 504.
        Stack shape : attributes RULE list(sequent) BIND.
        Start symbol: program. *)

    | MenhirState505 : (('s, _menhir_box_program) _menhir_cell1_sequent, _menhir_box_program) _menhir_state
      (** State 505.
        Stack shape : sequent.
        Start symbol: program. *)

    | MenhirState509 : (((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_RULE, _menhir_box_program) _menhir_cell1_list_sequent_ _menhir_cell0_option_preceded_BIND_nonempty_list_sequent___, _menhir_box_program) _menhir_state
      (** State 509.
        Stack shape : attributes RULE list(sequent) option(preceded(BIND,nonempty_list(sequent))).
        Start symbol: program. *)

    | MenhirState512 : (((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_RULE, _menhir_box_program) _menhir_cell1_list_sequent_ _menhir_cell0_option_preceded_BIND_nonempty_list_sequent___ _menhir_cell0_option_preceded_PIPE_term__, _menhir_box_program) _menhir_state
      (** State 512.
        Stack shape : attributes RULE list(sequent) option(preceded(BIND,nonempty_list(sequent))) option(preceded(PIPE,term)).
        Start symbol: program. *)

    | MenhirState515 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_PRED, _menhir_box_program) _menhir_state
      (** State 515.
        Stack shape : attributes PRED.
        Start symbol: program. *)

    | MenhirState516 : (((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_PRED, _menhir_box_program) _menhir_cell1_constant, _menhir_box_program) _menhir_state
      (** State 516.
        Stack shape : attributes PRED constant.
        Start symbol: program. *)

    | MenhirState518 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_PI, _menhir_box_program) _menhir_state
      (** State 518.
        Stack shape : attributes PI.
        Start symbol: program. *)

    | MenhirState519 : (((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_PI, _menhir_box_program) _menhir_cell1_nonempty_list_constant_w_loc_, _menhir_box_program) _menhir_state
      (** State 519.
        Stack shape : attributes PI nonempty_list(constant_w_loc).
        Start symbol: program. *)

    | MenhirState521 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_LPAREN, _menhir_box_program) _menhir_state
      (** State 521.
        Stack shape : attributes LPAREN.
        Start symbol: program. *)

    | MenhirState524 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_FUNC, _menhir_box_program) _menhir_state
      (** State 524.
        Stack shape : attributes FUNC.
        Start symbol: program. *)

    | MenhirState525 : (((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_FUNC, _menhir_box_program) _menhir_cell1_constant, _menhir_box_program) _menhir_state
      (** State 525.
        Stack shape : attributes FUNC constant.
        Start symbol: program. *)

    | MenhirState527 : ((((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_FUNC, _menhir_box_program) _menhir_cell1_constant, _menhir_box_program) _menhir_cell1_loption_separated_nonempty_list_CONJ_fotype_term__ _menhir_cell0_ARROW, _menhir_box_program) _menhir_state
      (** State 527.
        Stack shape : attributes FUNC constant loption(separated_nonempty_list(CONJ,fotype_term)) ARROW.
        Start symbol: program. *)

    | MenhirState529 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_FAMILY_TILDE, _menhir_box_program) _menhir_state
      (** State 529.
        Stack shape : attributes FAMILY_TILDE.
        Start symbol: program. *)

    | MenhirState531 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_head_term, _menhir_box_program) _menhir_state
      (** State 531.
        Stack shape : attributes head_term.
        Start symbol: program. *)

    | MenhirState532 : (((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_head_term, _menhir_box_program) _menhir_cell1_nonempty_list_closed_term_, _menhir_box_program) _menhir_state
      (** State 532.
        Stack shape : attributes head_term nonempty_list(closed_term).
        Start symbol: program. *)

    | MenhirState536 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_VDASH, _menhir_box_program) _menhir_state
      (** State 536.
        Stack shape : attributes clause_hd_term VDASH.
        Start symbol: program. *)

    | MenhirState538 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_SLASH, _menhir_box_program) _menhir_state
      (** State 538.
        Stack shape : attributes clause_hd_term SLASH.
        Start symbol: program. *)

    | MenhirState540 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_QDASH, _menhir_box_program) _menhir_state
      (** State 540.
        Stack shape : attributes clause_hd_term QDASH.
        Start symbol: program. *)

    | MenhirState542 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_OR, _menhir_box_program) _menhir_state
      (** State 542.
        Stack shape : attributes clause_hd_term OR.
        Start symbol: program. *)

    | MenhirState544 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_MOD, _menhir_box_program) _menhir_state
      (** State 544.
        Stack shape : attributes clause_hd_term MOD.
        Start symbol: program. *)

    | MenhirState546 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_MINUSs, _menhir_box_program) _menhir_state
      (** State 546.
        Stack shape : attributes clause_hd_term MINUSs.
        Start symbol: program. *)

    | MenhirState548 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_MINUSr, _menhir_box_program) _menhir_state
      (** State 548.
        Stack shape : attributes clause_hd_term MINUSr.
        Start symbol: program. *)

    | MenhirState550 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_MINUSi, _menhir_box_program) _menhir_state
      (** State 550.
        Stack shape : attributes clause_hd_term MINUSi.
        Start symbol: program. *)

    | MenhirState552 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_MINUS, _menhir_box_program) _menhir_state
      (** State 552.
        Stack shape : attributes clause_hd_term MINUS.
        Start symbol: program. *)

    | MenhirState554 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_IS, _menhir_box_program) _menhir_state
      (** State 554.
        Stack shape : attributes clause_hd_term IS.
        Start symbol: program. *)

    | MenhirState556 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_TIMES, _menhir_box_program) _menhir_state
      (** State 556.
        Stack shape : attributes clause_hd_term FAMILY_TIMES.
        Start symbol: program. *)

    | MenhirState558 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_TICK, _menhir_box_program) _menhir_state
      (** State 558.
        Stack shape : attributes clause_hd_term FAMILY_TICK.
        Start symbol: program. *)

    | MenhirState560 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_SHARP, _menhir_box_program) _menhir_state
      (** State 560.
        Stack shape : attributes clause_hd_term FAMILY_SHARP.
        Start symbol: program. *)

    | MenhirState563 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_PLUS, _menhir_box_program) _menhir_state
      (** State 563.
        Stack shape : attributes clause_hd_term FAMILY_PLUS.
        Start symbol: program. *)

    | MenhirState565 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_OR, _menhir_box_program) _menhir_state
      (** State 565.
        Stack shape : attributes clause_hd_term FAMILY_OR.
        Start symbol: program. *)

    | MenhirState567 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_MINUS, _menhir_box_program) _menhir_state
      (** State 567.
        Stack shape : attributes clause_hd_term FAMILY_MINUS.
        Start symbol: program. *)

    | MenhirState569 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_LT, _menhir_box_program) _menhir_state
      (** State 569.
        Stack shape : attributes clause_hd_term FAMILY_LT.
        Start symbol: program. *)

    | MenhirState571 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_GT, _menhir_box_program) _menhir_state
      (** State 571.
        Stack shape : attributes clause_hd_term FAMILY_GT.
        Start symbol: program. *)

    | MenhirState573 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_EXP, _menhir_box_program) _menhir_state
      (** State 573.
        Stack shape : attributes clause_hd_term FAMILY_EXP.
        Start symbol: program. *)

    | MenhirState575 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_EQ, _menhir_box_program) _menhir_state
      (** State 575.
        Stack shape : attributes clause_hd_term FAMILY_EQ.
        Start symbol: program. *)

    | MenhirState577 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_BTICK, _menhir_box_program) _menhir_state
      (** State 577.
        Stack shape : attributes clause_hd_term FAMILY_BTICK.
        Start symbol: program. *)

    | MenhirState579 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_AND, _menhir_box_program) _menhir_state
      (** State 579.
        Stack shape : attributes clause_hd_term FAMILY_AND.
        Start symbol: program. *)

    | MenhirState581 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_EQ2, _menhir_box_program) _menhir_state
      (** State 581.
        Stack shape : attributes clause_hd_term EQ2.
        Start symbol: program. *)

    | MenhirState583 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_EQ, _menhir_box_program) _menhir_state
      (** State 583.
        Stack shape : attributes clause_hd_term EQ.
        Start symbol: program. *)

    | MenhirState585 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_DIV, _menhir_box_program) _menhir_state
      (** State 585.
        Stack shape : attributes clause_hd_term DIV.
        Start symbol: program. *)

    | MenhirState587 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_DDARROW, _menhir_box_program) _menhir_state
      (** State 587.
        Stack shape : attributes clause_hd_term DDARROW.
        Start symbol: program. *)

    | MenhirState589 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_DARROW, _menhir_box_program) _menhir_state
      (** State 589.
        Stack shape : attributes clause_hd_term DARROW.
        Start symbol: program. *)

    | MenhirState591 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_CONS, _menhir_box_program) _menhir_state
      (** State 591.
        Stack shape : attributes clause_hd_term CONS.
        Start symbol: program. *)

    | MenhirState593 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_CONJ2, _menhir_box_program) _menhir_state
      (** State 593.
        Stack shape : attributes clause_hd_term CONJ2.
        Start symbol: program. *)

    | MenhirState595 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_CONJ, _menhir_box_program) _menhir_state
      (** State 595.
        Stack shape : attributes clause_hd_term CONJ.
        Start symbol: program. *)

    | MenhirState597 : ((('s, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_ARROW, _menhir_box_program) _menhir_state
      (** State 597.
        Stack shape : attributes clause_hd_term ARROW.
        Start symbol: program. *)

    | MenhirState601 : (('s, _menhir_box_program) _menhir_cell1_accumulate, _menhir_box_program) _menhir_state
      (** State 601.
        Stack shape : accumulate.
        Start symbol: program. *)

    | MenhirState606 : (('s, _menhir_box_program) _menhir_cell1_filename _menhir_cell0_CONJ, _menhir_box_program) _menhir_state
      (** State 606.
        Stack shape : filename CONJ.
        Start symbol: program. *)

  
  and ('s, 'r) _menhir_cell1_abbrevform = 
    | MenhirCell1_abbrevform of 's * ('s, 'r) _menhir_state * (Ast.Func.t * (Ast.Func.t * Ast.Loc.t) list)
  
  and ('s, 'r) _menhir_cell1_accumulate = 
    | MenhirCell1_accumulate of 's * ('s, 'r) _menhir_state * (string) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_attribute = 
    | MenhirCell1_attribute of 's * ('s, 'r) _menhir_state * (Ast.raw_attribute) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_attributes = 
    | MenhirCell1_attributes of 's * ('s, 'r) _menhir_state * (Ast.raw_attribute list) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_atype_term = 
    | MenhirCell1_atype_term of 's * ('s, 'r) _menhir_state * (Ast.raw_attribute list Ast.TypeExpression.t) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_clause_hd_term = 
    | MenhirCell1_clause_hd_term of 's * ('s, 'r) _menhir_state * (Ast.Term.t) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_closed_term = 
    | MenhirCell1_closed_term of 's * ('s, 'r) _menhir_state * (Ast.Term.t) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_constant = 
    | MenhirCell1_constant of 's * ('s, 'r) _menhir_state * (Ast.Func.t) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_constant_w_loc = 
    | MenhirCell1_constant_w_loc of 's * ('s, 'r) _menhir_state * (Ast.Func.t * Ast.Loc.t)
  
  and ('s, 'r) _menhir_cell1_decl = 
    | MenhirCell1_decl of 's * ('s, 'r) _menhir_state * (Ast.Program.decl)
  
  and ('s, 'r) _menhir_cell1_filename = 
    | MenhirCell1_filename of 's * ('s, 'r) _menhir_state * (string)
  
  and ('s, 'r) _menhir_cell1_fotype_term = 
    | MenhirCell1_fotype_term of 's * ('s, 'r) _menhir_state * (Ast.raw_attribute list Ast.TypeExpression.t) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_head_term = 
    | MenhirCell1_head_term of 's * ('s, 'r) _menhir_state * (Ast.Term.t) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_indexing = 
    | MenhirCell1_indexing of 's * ('s, 'r) _menhir_state * (int)
  
  and ('s, 'r) _menhir_cell1_list_constant_ = 
    | MenhirCell1_list_constant_ of 's * ('s, 'r) _menhir_state * (Ast.Func.t list)
  
  and ('s, 'r) _menhir_cell1_list_sequent_ = 
    | MenhirCell1_list_sequent_ of 's * ('s, 'r) _menhir_state * (Ast.Term.t Ast.Chr.sequent list)
  
  and ('s, 'r) _menhir_cell1_loption_separated_nonempty_list_CONJ_fotype_term__ = 
    | MenhirCell1_loption_separated_nonempty_list_CONJ_fotype_term__ of 's * ('s, 'r) _menhir_state * (Ast.raw_attribute list Ast.TypeExpression.t list) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_nonempty_list_closed_term_ = 
    | MenhirCell1_nonempty_list_closed_term_ of 's * ('s, 'r) _menhir_state * (Ast.Term.t list)
  
  and ('s, 'r) _menhir_cell1_nonempty_list_constant_w_loc_ = 
    | MenhirCell1_nonempty_list_constant_w_loc_ of 's * ('s, 'r) _menhir_state * ((Ast.Func.t * Ast.Loc.t) list)
  
  and ('s, 'r) _menhir_cell1_nonempty_list_indexing_ = 
    | MenhirCell1_nonempty_list_indexing_ of 's * ('s, 'r) _menhir_state * (int list)
  
  and 's _menhir_cell0_option_CONJ_ = 
    | MenhirCell0_option_CONJ_ of 's * (unit option)
  
  and 's _menhir_cell0_option_preceded_BIND_nonempty_list_sequent___ = 
    | MenhirCell0_option_preceded_BIND_nonempty_list_sequent___ of 's * (Ast.Term.t Ast.Chr.sequent list option)
  
  and 's _menhir_cell0_option_preceded_PIPE_term__ = 
    | MenhirCell0_option_preceded_PIPE_term__ of 's * (Ast.Term.t option)
  
  and ('s, 'r) _menhir_cell1_pred_item = 
    | MenhirCell1_pred_item of 's * ('s, 'r) _menhir_state * (Ast.Mode.t * Ast.raw_attribute list Ast.TypeExpression.t) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_separated_nonempty_list_CONJ_constant_ = 
    | MenhirCell1_separated_nonempty_list_CONJ_constant_ of 's * ('s, 'r) _menhir_state * (Ast.Func.t list) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_sequent = 
    | MenhirCell1_sequent of 's * ('s, 'r) _menhir_state * (Ast.Term.t Ast.Chr.sequent) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_subtrie = 
    | MenhirCell1_subtrie of 's * ('s, 'r) _menhir_state * ((string * string) list)
  
  and ('s, 'r) _menhir_cell1_term = 
    | MenhirCell1_term of 's * ('s, 'r) _menhir_state * (Ast.Term.t) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_term_noconj = 
    | MenhirCell1_term_noconj of 's * ('s, 'r) _menhir_state * (Ast.Term.t) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_type_term = 
    | MenhirCell1_type_term of 's * ('s, 'r) _menhir_state * (Ast.raw_attribute list Ast.TypeExpression.t) * Lexing.position
  
  and 's _menhir_cell0_ARROW = 
    | MenhirCell0_ARROW of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_AS = 
    | MenhirCell0_AS of 's * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_BIND = 
    | MenhirCell1_BIND of 's * ('s, 'r) _menhir_state * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_BIND = 
    | MenhirCell0_BIND of 's * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_CLOSED = 
    | MenhirCell1_CLOSED of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_COLON = 
    | MenhirCell1_COLON of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_COLON = 
    | MenhirCell0_COLON of 's * Lexing.position
  
  and 's _menhir_cell0_CONJ = 
    | MenhirCell0_CONJ of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_CONJ2 = 
    | MenhirCell0_CONJ2 of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_CONS = 
    | MenhirCell0_CONS of 's * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_CONSTRAINT = 
    | MenhirCell1_CONSTRAINT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_DARROW = 
    | MenhirCell0_DARROW of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_DDARROW = 
    | MenhirCell0_DDARROW of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_DIV = 
    | MenhirCell0_DIV of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_EQ = 
    | MenhirCell0_EQ of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_EQ2 = 
    | MenhirCell0_EQ2 of 's * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_EXPORTDEF = 
    | MenhirCell1_EXPORTDEF of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_FAMILY_AND = 
    | MenhirCell0_FAMILY_AND of 's * (
# 90 "src/parser/tokens.mly"
       (string)
# 1317 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_FAMILY_BTICK = 
    | MenhirCell0_FAMILY_BTICK of 's * (
# 86 "src/parser/tokens.mly"
       (string)
# 1324 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_FAMILY_EQ = 
    | MenhirCell0_FAMILY_EQ of 's * (
# 84 "src/parser/tokens.mly"
       (string)
# 1331 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_FAMILY_EXP = 
    | MenhirCell0_FAMILY_EXP of 's * (
# 81 "src/parser/tokens.mly"
       (string)
# 1338 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_FAMILY_GT = 
    | MenhirCell0_FAMILY_GT of 's * (
# 83 "src/parser/tokens.mly"
       (string)
# 1345 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_FAMILY_LT = 
    | MenhirCell0_FAMILY_LT of 's * (
# 82 "src/parser/tokens.mly"
       (string)
# 1352 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_FAMILY_MINUS = 
    | MenhirCell0_FAMILY_MINUS of 's * (
# 80 "src/parser/tokens.mly"
       (string)
# 1359 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_FAMILY_OR = 
    | MenhirCell0_FAMILY_OR of 's * (
# 91 "src/parser/tokens.mly"
       (string)
# 1366 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_FAMILY_PLUS = 
    | MenhirCell0_FAMILY_PLUS of 's * (
# 78 "src/parser/tokens.mly"
       (string)
# 1373 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_FAMILY_SHARP = 
    | MenhirCell0_FAMILY_SHARP of 's * (
# 88 "src/parser/tokens.mly"
       (string)
# 1380 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_FAMILY_TICK = 
    | MenhirCell0_FAMILY_TICK of 's * (
# 87 "src/parser/tokens.mly"
       (string)
# 1387 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_FAMILY_TILDE = 
    | MenhirCell1_FAMILY_TILDE of 's * ('s, 'r) _menhir_state * (
# 89 "src/parser/tokens.mly"
       (string)
# 1394 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_FAMILY_TIMES = 
    | MenhirCell0_FAMILY_TIMES of 's * (
# 79 "src/parser/tokens.mly"
       (string)
# 1401 "src/parser/grammar.ml"
  ) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_FIXITY = 
    | MenhirCell1_FIXITY of 's * ('s, 'r) _menhir_state * (
# 56 "src/parser/tokens.mly"
       (string)
# 1408 "src/parser/grammar.ml"
  ) * Lexing.position
  
  and 's _menhir_cell0_FULLSTOP = 
    | MenhirCell0_FULLSTOP of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_FUNC = 
    | MenhirCell1_FUNC of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_INDEX = 
    | MenhirCell1_INDEX of 's * ('s, 'r) _menhir_state * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_IO_COLON = 
    | MenhirCell1_IO_COLON of 's * ('s, 'r) _menhir_state * (
# 22 "src/parser/tokens.mly"
       (char)
# 1424 "src/parser/grammar.ml"
  )
  
  and 's _menhir_cell0_IS = 
    | MenhirCell0_IS of 's * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_KIND = 
    | MenhirCell1_KIND of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LBRACKET = 
    | MenhirCell1_LBRACKET of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LCURLY = 
    | MenhirCell1_LCURLY of 's * ('s, 'r) _menhir_state * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_LCURLY = 
    | MenhirCell0_LCURLY of 's * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LOCAL = 
    | MenhirCell1_LOCAL of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LOCALKIND = 
    | MenhirCell1_LOCALKIND of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LPAREN = 
    | MenhirCell1_LPAREN of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_LPAREN = 
    | MenhirCell0_LPAREN of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_MACRO = 
    | MenhirCell1_MACRO of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_MINUS = 
    | MenhirCell0_MINUS of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_MINUSi = 
    | MenhirCell0_MINUSi of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_MINUSr = 
    | MenhirCell0_MINUSr of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_MINUSs = 
    | MenhirCell0_MINUSs of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_MOD = 
    | MenhirCell0_MOD of 's * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_MODULE = 
    | MenhirCell1_MODULE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_NAMESPACE = 
    | MenhirCell1_NAMESPACE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_OR = 
    | MenhirCell0_OR of 's * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_PI = 
    | MenhirCell1_PI of 's * ('s, 'r) _menhir_state * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_PIPE = 
    | MenhirCell1_PIPE of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_PRED = 
    | MenhirCell1_PRED of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_QDASH = 
    | MenhirCell0_QDASH of 's * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_RPAREN = 
    | MenhirCell0_RPAREN of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RULE = 
    | MenhirCell1_RULE of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_SHORTEN = 
    | MenhirCell1_SHORTEN of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_SIG = 
    | MenhirCell1_SIG of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_SIGMA = 
    | MenhirCell1_SIGMA of 's * ('s, 'r) _menhir_state * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_SLASH = 
    | MenhirCell0_SLASH of 's * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_TYPE = 
    | MenhirCell1_TYPE of 's * ('s, 'r) _menhir_state * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_TYPEABBREV = 
    | MenhirCell1_TYPEABBREV of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_USEONLY = 
    | MenhirCell1_USEONLY of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_VDASH = 
    | MenhirCell0_VDASH of 's * Lexing.position * Lexing.position
  
  and _menhir_box_program = 
    | MenhirBox_program of (Ast.Program.t) [@@unboxed]
  
  and _menhir_box_prefix_SYMB = 
    | MenhirBox_prefix_SYMB of (Ast.Func.t) [@@unboxed]
  
  and _menhir_box_postfix_SYMB = 
    | MenhirBox_postfix_SYMB of (Ast.Func.t) [@@unboxed]
  
  and _menhir_box_infix_SYMB = 
    | MenhirBox_infix_SYMB of (Ast.Func.t) [@@unboxed]
  
  and _menhir_box_goal = 
    | MenhirBox_goal of (Ast.Term.t) [@@unboxed]
  
  let _menhir_action_005 =
    fun c ->
      (
# 285 "src/parser/grammar.mly"
               ( c, [] )
# 1543 "src/parser/grammar.ml"
       : (Ast.Func.t * (Ast.Func.t * Ast.Loc.t) list))
  
  let _menhir_action_006 =
    fun args hd ->
      (
# 286 "src/parser/grammar.mly"
                                                                      ( hd, args  )
# 1551 "src/parser/grammar.ml"
       : (Ast.Func.t * (Ast.Func.t * Ast.Loc.t) list))
  
  let _menhir_action_007 =
    fun () ->
      (
# 189 "src/parser/grammar.mly"
             ( ".elpi" )
# 1559 "src/parser/grammar.ml"
       : (string))
  
  let _menhir_action_008 =
    fun () ->
      (
# 190 "src/parser/grammar.mly"
         ( ".mod" )
# 1567 "src/parser/grammar.ml"
       : (string))
  
  let _menhir_action_009 =
    fun () ->
      (
# 191 "src/parser/grammar.mly"
            ( ".sig" )
# 1575 "src/parser/grammar.ml"
       : (string))
  
  let _menhir_action_010 =
    fun () ->
      (
# 192 "src/parser/grammar.mly"
          ( ".sig" )
# 1583 "src/parser/grammar.ml"
       : (string))
  
  let _menhir_action_011 =
    fun _endpos_xs_ _startpos__1_ xs ->
      let args = 
# 241 "<standard.mly>"
    ( xs )
# 1591 "src/parser/grammar.ml"
       in
      let _endpos_args_ = _endpos_xs_ in
      let _endpos = _endpos_args_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 227 "src/parser/grammar.mly"
                                                      ( { tloc = loc _loc; tit = TPred ([], args) } )
# 1600 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_012 =
    fun _endpos_xs_inlined1_ _startpos__1_ xs xs_inlined1 ->
      let out_args =
        let xs = xs_inlined1 in
        
# 241 "<standard.mly>"
    ( xs )
# 1610 "src/parser/grammar.ml"
        
      in
      let _endpos_out_args_ = _endpos_xs_inlined1_ in
      let in_args = 
# 241 "<standard.mly>"
    ( xs )
# 1617 "src/parser/grammar.ml"
       in
      let _endpos = _endpos_out_args_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 228 "src/parser/grammar.mly"
                                                                                                       (
    let args = List.map (fun x -> Mode.Input,x) in_args @ List.map (fun x -> Mode.Output,x) out_args in
    { tloc = loc _loc; tit = TPred ([Functional], args) }
  )
# 1628 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_013 =
    fun _endpos_xs_ _startpos__1_ xs ->
      let in_args = 
# 241 "<standard.mly>"
    ( xs )
# 1636 "src/parser/grammar.ml"
       in
      let _endpos_in_args_ = _endpos_xs_ in
      let _endpos = _endpos_in_args_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 232 "src/parser/grammar.mly"
                                                   (
    let args = List.map (fun x -> Mode.Input,x) in_args in
    { tloc = loc _loc; tit = TPred ([Functional], args) }
  )
# 1648 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_014 =
    fun s ->
      (
# 355 "src/parser/grammar.mly"
                 ( If s )
# 1656 "src/parser/grammar.ml"
       : (Ast.raw_attribute))
  
  let _menhir_action_015 =
    fun s ->
      (
# 356 "src/parser/grammar.mly"
                    ( Name s )
# 1664 "src/parser/grammar.ml"
       : (Ast.raw_attribute))
  
  let _menhir_action_016 =
    fun s ->
      (
# 357 "src/parser/grammar.mly"
                    ( After s )
# 1672 "src/parser/grammar.ml"
       : (Ast.raw_attribute))
  
  let _menhir_action_017 =
    fun s ->
      (
# 358 "src/parser/grammar.mly"
                     ( Before s )
# 1680 "src/parser/grammar.ml"
       : (Ast.raw_attribute))
  
  let _menhir_action_018 =
    fun s ->
      (
# 359 "src/parser/grammar.mly"
                      ( Replace s )
# 1688 "src/parser/grammar.ml"
       : (Ast.raw_attribute))
  
  let _menhir_action_019 =
    fun s ->
      (
# 360 "src/parser/grammar.mly"
                     ( Remove s )
# 1696 "src/parser/grammar.ml"
       : (Ast.raw_attribute))
  
  let _menhir_action_020 =
    fun () ->
      (
# 361 "src/parser/grammar.mly"
           ( External )
# 1704 "src/parser/grammar.ml"
       : (Ast.raw_attribute))
  
  let _menhir_action_021 =
    fun () ->
      (
# 362 "src/parser/grammar.mly"
             ( Functional )
# 1712 "src/parser/grammar.ml"
       : (Ast.raw_attribute))
  
  let _menhir_action_022 =
    fun () ->
      (
# 363 "src/parser/grammar.mly"
          ( Untyped )
# 1720 "src/parser/grammar.ml"
       : (Ast.raw_attribute))
  
  let _menhir_action_023 =
    fun l o ->
      (
# 364 "src/parser/grammar.mly"
                                                                          ( Index (l,o) )
# 1728 "src/parser/grammar.ml"
       : (Ast.raw_attribute))
  
  let _menhir_action_024 =
    fun () ->
      (
# 350 "src/parser/grammar.mly"
  ( [] )
# 1736 "src/parser/grammar.ml"
       : (Ast.raw_attribute list))
  
  let _menhir_action_025 =
    fun () ->
      (
# 351 "src/parser/grammar.mly"
           ( [ External ] )
# 1744 "src/parser/grammar.ml"
       : (Ast.raw_attribute list))
  
  let _menhir_action_026 =
    fun l ->
      (
# 352 "src/parser/grammar.mly"
                                                       ( l )
# 1752 "src/parser/grammar.ml"
       : (Ast.raw_attribute list))
  
  let _menhir_action_027 =
    fun _endpos_c_ _startpos_c_ c ->
      let _endpos = _endpos_c_ in
      let _startpos = _startpos_c_ in
      let _loc = (_startpos, _endpos) in
      (
# 250 "src/parser/grammar.mly"
               ( { tloc = loc _loc; tit = TConst (fix_church c) } )
# 1763 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_028 =
    fun t ->
      (
# 251 "src/parser/grammar.mly"
                                ( t )
# 1771 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_029 =
    fun t ->
      (
# 252 "src/parser/grammar.mly"
                                     ( t )
# 1779 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_030 =
    fun _endpos_bind_ _startpos_bind_ b bind ->
      let _loc_bind_ = (_startpos_bind_, _endpos_bind_) in
      (
# 419 "src/parser/grammar.mly"
                        ( (loc _loc_bind_, None, b) )
# 1788 "src/parser/grammar.ml"
       : (Ast.Loc.t * Ast.Term.typ option * Ast.Term.t))
  
  let _menhir_action_031 =
    fun _endpos_bind_ _startpos_bind_ b bind ty ->
      let _loc_bind_ = (_startpos_bind_, _endpos_bind_) in
      (
# 420 "src/parser/grammar.mly"
                                               ( (loc _loc_bind_, Some ty, b) )
# 1797 "src/parser/grammar.ml"
       : (Ast.Loc.t * Ast.Term.typ option * Ast.Term.t))
  
  let _menhir_action_032 =
    fun _endpos_bind_ _startpos_bind_ b bind ->
      let _loc_bind_ = (_startpos_bind_, _endpos_bind_) in
      (
# 416 "src/parser/grammar.mly"
                        ( (loc _loc_bind_, None, b) )
# 1806 "src/parser/grammar.ml"
       : (Ast.Loc.t * Ast.Term.typ option * Ast.Term.t))
  
  let _menhir_action_033 =
    fun _endpos_b_ _startpos_t_ b t ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_t_ in
      let _loc = (_startpos, _endpos) in
      (
# 412 "src/parser/grammar.mly"
                               ( mkLam (loc _loc) (Func.show t) None b )
# 1817 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_034 =
    fun _endpos_b_ _startpos_t_ b t ty ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_t_ in
      let _loc = (_startpos, _endpos) in
      (
# 413 "src/parser/grammar.mly"
                                                      ( mkLam (loc _loc) (Func.show t) (Some ty) b )
# 1828 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_035 =
    fun _endpos_b_ _startpos_t_ b t ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_t_ in
      let _loc = (_startpos, _endpos) in
      (
# 423 "src/parser/grammar.mly"
                               ( mkLam (loc _loc) (Func.show t) None b )
# 1839 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_036 =
    fun _endpos_b_ _startpos_t_ b t ty ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_t_ in
      let _loc = (_startpos, _endpos) in
      (
# 424 "src/parser/grammar.mly"
                                                      ( mkLam (loc _loc) (Func.show t) (Some ty) b )
# 1850 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_037 =
    fun _endpos_new_goal_ _startpos_attributes_ attributes guard new_goal to_match to_remove ->
      let _endpos = _endpos_new_goal_ in
      let _startpos = _startpos_attributes_ in
      let _loc = (_startpos, _endpos) in
      (
# 203 "src/parser/grammar.mly"
                                    (
    { Chr.to_match; to_remove = Util.option_default [] to_remove; guard; new_goal; attributes; loc=(loc _loc) }
  )
# 1863 "src/parser/grammar.ml"
       : ((Ast.raw_attribute list, Ast.Term.t) Ast.Chr.t))
  
  let _menhir_action_038 =
    fun _endpos_attributes_ _endpos_body_ _startpos_attributes_ _startpos_body_ attributes body ->
      let _endpos = _endpos_body_ in
      let _symbolstartpos = if _startpos_attributes_ != _endpos_attributes_ then
        _startpos_attributes_
      else
        _startpos_body_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 334 "src/parser/grammar.mly"
                                                  (
    { Clause.loc = loc _sloc;
      attributes;
      body;
      needs_spilling = ();
    }
  )
# 1883 "src/parser/grammar.ml"
       : ((Ast.Term.t, Ast.raw_attribute list, unit) Ast.Clause.t))
  
  let _menhir_action_039 =
    fun _endpos_attributes_ _endpos_r_ _endpos_v_ _startpos_attributes_ _startpos_l_ _startpos_v_ attributes l r v ->
      let _endpos = _endpos_r_ in
      let _symbolstartpos = if _startpos_attributes_ != _endpos_attributes_ then
        _startpos_attributes_
      else
        _startpos_l_ in
      let _loc_v_ = (_startpos_v_, _endpos_v_) in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 341 "src/parser/grammar.mly"
                                                                   ( 
    { Clause.loc = loc _sloc;
      attributes;
      body = mkApp (loc _sloc) [mkConst (loc _loc_v_) Func.rimplf;l;r];
      needs_spilling = ();
    }
)
# 1904 "src/parser/grammar.ml"
       : ((Ast.Term.t, Ast.raw_attribute list, unit) Ast.Clause.t))
  
  let _menhir_action_040 =
    fun _endpos_t_ _startpos_t_ t ->
      let _endpos = _endpos_t_ in
      let _symbolstartpos = _startpos_t_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 456 "src/parser/grammar.mly"
               ( mkConst (loc _sloc) t )
# 1915 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_041 =
    fun _endpos__3_ _startpos__1_ t ->
      let _endpos = _endpos__3_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 457 "src/parser/grammar.mly"
                           ( mkParens_if_impl_or_conj (loc _loc) t )
# 1926 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_042 =
    fun _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_hd_ in
      let _loc_hd_ = (_startpos_hd_, _endpos_hd_) in
      let _loc = (_startpos, _endpos) in
      (
# 460 "src/parser/grammar.mly"
                                                                 ( desugar_multi_binder (loc _loc) @@ mkApp (loc _loc) (mkConst (loc _loc_hd_) (Func.from_string "pi") :: binder args b) )
# 1938 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_043 =
    fun _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_hd_ in
      let _loc_hd_ = (_startpos_hd_, _endpos_hd_) in
      let _loc = (_startpos, _endpos) in
      (
# 461 "src/parser/grammar.mly"
                                                                    ( desugar_multi_binder (loc _loc) @@ mkApp (loc _loc) (mkConst (loc _loc_hd_) (Func.from_string "sigma") :: binder args b) )
# 1950 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_044 =
    fun _endpos_b_ _startpos_hd_ args b hd ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_hd_ in
      let _loc = (_startpos, _endpos) in
      (
# 462 "src/parser/grammar.mly"
                                                                                   (
    let args = binder1 args b in
    let t = mkApp (loc _loc) (hd :: args) in
    t
)
# 1965 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_045 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 530 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 1974 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 1979 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 1989 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_046 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 531 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 1998 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 2003 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2013 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_047 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 532 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 2022 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 2027 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2037 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_048 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 533 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 2046 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 2051 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2061 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_049 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 534 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 2070 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 2075 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2085 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_050 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 535 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 2094 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 2099 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2109 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_051 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 536 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 2118 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 2123 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2133 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_052 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 537 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 2142 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 2147 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2157 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_053 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 538 "src/parser/grammar.mly"
                  ( Func.from_string s )
# 2166 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 2171 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2181 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_054 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 539 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 2190 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 2195 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2205 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_055 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 540 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 2214 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 2219 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2229 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_056 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 541 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 2238 "src/parser/grammar.ml"
         in
        
# 526 "src/parser/grammar.mly"
                       ( x )
# 2243 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2253 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_057 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 544 "src/parser/grammar.mly"
         ( Func.consf )
# 2263 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2268 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2274 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2285 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_058 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 545 "src/parser/grammar.mly"
         ( Func.eqf )
# 2295 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2300 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2306 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2317 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_059 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 546 "src/parser/grammar.mly"
         ( Func.from_string "-" )
# 2327 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2332 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2338 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2349 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_060 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 547 "src/parser/grammar.mly"
         ( Func.from_string "r-" )
# 2359 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2364 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2370 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2381 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_061 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 548 "src/parser/grammar.mly"
         ( Func.from_string "i-" )
# 2391 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2396 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2402 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2413 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_062 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 549 "src/parser/grammar.mly"
         ( Func.from_string "s-" )
# 2423 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2428 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2434 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2445 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_063 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 550 "src/parser/grammar.mly"
         ( Func.from_string "==" )
# 2455 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2460 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2466 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2477 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_064 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 551 "src/parser/grammar.mly"
         ( Func.orf )
# 2487 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2492 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2498 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2509 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_065 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 552 "src/parser/grammar.mly"
         ( Func.isf )
# 2519 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2524 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2530 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2541 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_066 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 553 "src/parser/grammar.mly"
         ( Func.from_string "mod" )
# 2551 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2556 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2562 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2573 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_067 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 554 "src/parser/grammar.mly"
         ( Func.from_string "div" )
# 2583 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2588 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2594 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2605 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_068 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 555 "src/parser/grammar.mly"
         ( Func.arrowf )
# 2615 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2620 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2626 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2637 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_069 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 556 "src/parser/grammar.mly"
         ( Func.implf )
# 2647 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2652 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2658 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2669 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_070 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 557 "src/parser/grammar.mly"
          ( Func.implf )
# 2679 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2684 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2690 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2701 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_071 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 558 "src/parser/grammar.mly"
         ( Func.sequentf )
# 2711 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2716 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2722 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2733 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_072 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 559 "src/parser/grammar.mly"
         ( Func.from_string "/" )
# 2743 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2748 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2754 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2765 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_073 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 560 "src/parser/grammar.mly"
         ( Func.andf )
# 2775 "src/parser/grammar.ml"
           in
          
# 563 "src/parser/grammar.mly"
                                          ( x )
# 2780 "src/parser/grammar.ml"
          
        in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2786 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2797 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_074 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x = 
# 564 "src/parser/grammar.mly"
         ( Func.andf )
# 2806 "src/parser/grammar.ml"
         in
        
# 527 "src/parser/grammar.mly"
                                   ( x )
# 2811 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 467 "src/parser/grammar.mly"
                                                  ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 2822 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_075 =
    fun _endpos_r_ _endpos_s_ _startpos_s_ r s ->
      let s =
        let x = 
# 575 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 2831 "src/parser/grammar.ml"
         in
        
# 512 "src/parser/grammar.mly"
                        ( x )
# 2836 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_s_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 468 "src/parser/grammar.mly"
                       ( mkAppF (loc _loc) (loc _loc_s_,s) [r] )
# 2846 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_076 =
    fun _endpos_s_ _startpos_l_ _startpos_s_ l s ->
      let s =
        let x = 
# 578 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 2855 "src/parser/grammar.ml"
         in
        
# 515 "src/parser/grammar.mly"
                         ( x )
# 2860 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_s_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 469 "src/parser/grammar.mly"
                                   ( mkAppF (loc _loc) (loc _loc_s_,s) [l] )
# 2870 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_077 =
    fun t ->
      (
# 452 "src/parser/grammar.mly"
                          ( t )
# 2878 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_078 =
    fun t ->
      (
# 453 "src/parser/grammar.mly"
                            ( t )
# 2886 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_079 =
    fun _endpos_x_ _startpos_x_ x ->
      let _endpos = _endpos_x_ in
      let _startpos = _startpos_x_ in
      let _loc = (_startpos, _endpos) in
      (
# 381 "src/parser/grammar.mly"
              ( mkC (loc _loc) (cint.Util.CData.cin x))
# 2897 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_080 =
    fun _endpos_x_ _startpos_x_ x ->
      let _endpos = _endpos_x_ in
      let _startpos = _startpos_x_ in
      let _loc = (_startpos, _endpos) in
      (
# 382 "src/parser/grammar.mly"
            ( mkC (loc _loc) (cfloat.Util.CData.cin x))
# 2908 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_081 =
    fun _endpos_x_ _startpos_x_ x ->
      let _endpos = _endpos_x_ in
      let _startpos = _startpos_x_ in
      let _loc = (_startpos, _endpos) in
      (
# 383 "src/parser/grammar.mly"
             ( mkC (loc _loc) (cstring.Util.CData.cin x))
# 2919 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_082 =
    fun _endpos_x_ _startpos_x_ x ->
      let _endpos = _endpos_x_ in
      let _startpos = _startpos_x_ in
      let _loc = (_startpos, _endpos) in
      (
# 384 "src/parser/grammar.mly"
             ( mkQuoted (loc _loc) (fst x) (snd x) )
# 2930 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_083 =
    fun _endpos__5_ _endpos_a_ _startpos__1_ _startpos_a_ a c t ->
      let _endpos = _endpos__5_ in
      let _startpos = _startpos__1_ in
      let _loc_a_ = (_startpos_a_, _endpos_a_) in
      let _loc = (_startpos, _endpos) in
      (
# 385 "src/parser/grammar.mly"
                                             ( mkApp (loc _loc) [mkCon (loc _loc_a_) "as";t;c] )
# 2942 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_084 =
    fun _endpos_l_ _startpos__1_ l ->
      let _endpos = _endpos_l_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 386 "src/parser/grammar.mly"
                           ( mkSeq ~loc:(loc _loc) l )
# 2953 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_085 =
    fun _endpos_l_ _startpos__1_ l ->
      let _endpos = _endpos_l_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 387 "src/parser/grammar.mly"
                                  (  mkSeq ~loc:(loc _loc) l )
# 2964 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_086 =
    fun _endpos__3_ _endpos_l_ _startpos_l_ l t ->
      let _endpos = _endpos__3_ in
      let _startpos = _startpos_l_ in
      let _loc_l_ = (_startpos_l_, _endpos_l_) in
      let _loc = (_startpos, _endpos) in
      (
# 388 "src/parser/grammar.mly"
                               ( mkAppF (loc _loc) (loc _loc_l_,Func.spillf) [t] )
# 2976 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_087 =
    fun t ->
      (
# 389 "src/parser/grammar.mly"
                ( t )
# 2984 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_088 =
    fun c ->
      (
# 472 "src/parser/grammar.mly"
               (
    if c = "nil" then Func.nilf
    else if c = "cons" then Func.consf
    else Func.from_string c )
# 2995 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_089 =
    fun () ->
      (
# 476 "src/parser/grammar.mly"
     ( Func.from_string "if" )
# 3003 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_090 =
    fun () ->
      (
# 477 "src/parser/grammar.mly"
       ( Func.from_string "name" )
# 3011 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_091 =
    fun () ->
      (
# 478 "src/parser/grammar.mly"
         ( Func.from_string "before" )
# 3019 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_092 =
    fun () ->
      (
# 479 "src/parser/grammar.mly"
        ( Func.from_string "after" )
# 3027 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_093 =
    fun () ->
      (
# 480 "src/parser/grammar.mly"
          ( Func.from_string "replace" )
# 3035 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_094 =
    fun () ->
      (
# 481 "src/parser/grammar.mly"
         ( Func.from_string "remove" )
# 3043 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_095 =
    fun () ->
      (
# 482 "src/parser/grammar.mly"
        ( Func.from_string "index" )
# 3051 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_096 =
    fun () ->
      (
# 483 "src/parser/grammar.mly"
             ( Func.from_string "functional" )
# 3059 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_097 =
    fun () ->
      (
# 484 "src/parser/grammar.mly"
          ( Func.from_string "untyped" )
# 3067 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_098 =
    fun c ->
      (
# 485 "src/parser/grammar.mly"
         ( Func.from_string @@ String.make 1 c )
# 3075 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_099 =
    fun () ->
      (
# 486 "src/parser/grammar.mly"
      ( Func.cutf )
# 3083 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_100 =
    fun () ->
      (
# 487 "src/parser/grammar.mly"
          ( Func.dummyname )
# 3091 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_101 =
    fun s ->
      (
# 488 "src/parser/grammar.mly"
                                  ( s )
# 3099 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_102 =
    fun () ->
      (
# 489 "src/parser/grammar.mly"
                      ( Func.from_string "as" )
# 3107 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_103 =
    fun () ->
      (
# 490 "src/parser/grammar.mly"
      ( Func.nilf )
# 3115 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_104 =
    fun _endpos_c_ _startpos_c_ c ->
      let _endpos = _endpos_c_ in
      let _startpos = _startpos_c_ in
      let _loc = (_startpos, _endpos) in
      (
# 493 "src/parser/grammar.mly"
               ( c, loc _loc )
# 3126 "src/parser/grammar.ml"
       : (Ast.Func.t * Ast.Loc.t))
  
  let _menhir_action_105 =
    fun c ->
      (
# 165 "src/parser/grammar.mly"
                       ( Program.Clause c )
# 3134 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_106 =
    fun r ->
      (
# 166 "src/parser/grammar.mly"
                         ( Program.Chr r )
# 3142 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_107 =
    fun p ->
      (
# 167 "src/parser/grammar.mly"
                     ( Program.Pred p )
# 3150 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_108 =
    fun t ->
      (
# 168 "src/parser/grammar.mly"
                      ( Program.Type t )
# 3158 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_109 =
    fun t ->
      (
# 169 "src/parser/grammar.mly"
                     ( Program.Kind t )
# 3166 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_110 =
    fun m ->
      (
# 170 "src/parser/grammar.mly"
                      ( Program.Macro m )
# 3174 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_111 =
    fun _endpos__5_ _startpos__1_ cl hyps ->
      let _endpos = _endpos__5_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 171 "src/parser/grammar.mly"
                                                                        ( Program.Constraint(loc _sloc, hyps, cl) )
# 3185 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_112 =
    fun _endpos__3_ _startpos__1_ cl ->
      let _endpos = _endpos__3_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 172 "src/parser/grammar.mly"
                                          ( Program.Constraint(loc _sloc, [], cl) )
# 3196 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_113 =
    fun _endpos__3_ _startpos__1_ c ->
      let _endpos = _endpos__3_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 173 "src/parser/grammar.mly"
                                  ( Program.Namespace(loc _sloc, c ))
# 3207 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_114 =
    fun _endpos__3_ _startpos__1_ s ->
      let _endpos = _endpos__3_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 174 "src/parser/grammar.mly"
                                 ( Program.Shorten(loc _sloc, s) )
# 3218 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_115 =
    fun a ->
      (
# 175 "src/parser/grammar.mly"
                           ( Program.TypeAbbreviation a )
# 3226 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_116 =
    fun _endpos__1_ _startpos__1_ ->
      let _endpos = _endpos__1_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 176 "src/parser/grammar.mly"
         ( Program.Begin (loc _sloc) )
# 3237 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_117 =
    fun _endpos__1_ _startpos__1_ ->
      let _endpos = _endpos__1_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 177 "src/parser/grammar.mly"
         ( Program.End (loc _sloc) )
# 3248 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_118 =
    fun _endpos__3_ _startpos_ext_ ext l ->
      let _endpos = _endpos__3_ in
      let _symbolstartpos = _startpos_ext_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 178 "src/parser/grammar.mly"
                                                                         (
    Program.Accumulated(loc _sloc,List.(concat (map (fun x ->
      let cwd = Filename.dirname (loc _sloc).source_name in
      C.parse_file ~cwd (x ^ ext)) l)))
  )
# 3263 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_119 =
    fun _endpos__4_ _startpos__1_ l ->
      let _endpos = _endpos__4_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 183 "src/parser/grammar.mly"
                                                                                 (
    raise (ParseError(loc _loc,"local keyword is no longer supported"))  )
# 3275 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_120 =
    fun _endpos__2_ _startpos__1_ ->
      let _endpos = _endpos__2_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 185 "src/parser/grammar.mly"
                    ( Program.Ignored (loc _sloc) )
# 3286 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_121 =
    fun _endpos__2_ _startpos_f_ f ->
      let _endpos = _endpos__2_ in
      let _startpos = _startpos_f_ in
      let _loc = (_startpos, _endpos) in
      (
# 186 "src/parser/grammar.mly"
                       ( error_mixfix (loc _loc) )
# 3297 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_122 =
    fun c ->
      (
# 195 "src/parser/grammar.mly"
               ( Func.show c )
# 3305 "src/parser/grammar.ml"
       : (string))
  
  let _menhir_action_123 =
    fun s ->
      (
# 196 "src/parser/grammar.mly"
             ( s )
# 3313 "src/parser/grammar.ml"
       : (string))
  
  let _menhir_action_124 =
    fun _endpos_c_ _startpos_c_ c f i ->
      let _loc_c_ = (_startpos_c_, _endpos_c_) in
      (
# 301 "src/parser/grammar.mly"
                                        ( (fixity_of_string f,c,i,loc _loc_c_) )
# 3322 "src/parser/grammar.ml"
       : (Elpi_lexer_config.Lexer_config.fixity * Ast.Func.t * int * Ast.Loc.t))
  
  let _menhir_action_125 =
    fun _endpos_c_ _startpos_c_ c f i ->
      let _loc_c_ = (_startpos_c_, _endpos_c_) in
      (
# 302 "src/parser/grammar.mly"
                                           ( (fixity_of_string f,c,i,loc _loc_c_) )
# 3331 "src/parser/grammar.ml"
       : (Elpi_lexer_config.Lexer_config.fixity * Ast.Func.t * int * Ast.Loc.t))
  
  let _menhir_action_126 =
    fun _endpos_c_ _startpos_c_ c ->
      let _endpos = _endpos_c_ in
      let _startpos = _startpos_c_ in
      let _loc = (_startpos, _endpos) in
      (
# 254 "src/parser/grammar.mly"
               ( { tloc = loc _loc; tit = TConst (fix_church c) } )
# 3342 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_127 =
    fun _endpos_args_ _startpos_hd_ args hd ->
      let _endpos = _endpos_args_ in
      let _startpos = _startpos_hd_ in
      let _loc = (_startpos, _endpos) in
      (
# 255 "src/parser/grammar.mly"
                                                  ( { tloc = loc _loc; tit = TApp (hd, List.hd args, List.tl args) } )
# 3353 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_128 =
    fun t ->
      (
# 256 "src/parser/grammar.mly"
                                     ( t )
# 3361 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_129 =
    fun t ->
      (
# 257 "src/parser/grammar.mly"
                                ( t )
# 3369 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_130 =
    fun g ->
      (
# 330 "src/parser/grammar.mly"
                ( g )
# 3377 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_131 =
    fun g ->
      (
# 331 "src/parser/grammar.mly"
                     ( g )
# 3385 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_132 =
    fun _endpos_t_ _startpos_t_ t ->
      let _endpos = _endpos_t_ in
      let _startpos = _startpos_t_ in
      let _loc = (_startpos, _endpos) in
      (
# 396 "src/parser/grammar.mly"
                                  ( mkConst (loc _loc) t )
# 3396 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_133 =
    fun _endpos__3_ _startpos__1_ t ->
      let _endpos = _endpos__3_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 397 "src/parser/grammar.mly"
                           ( mkParens_if_impl_or_conj (loc _loc) t )
# 3407 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_134 =
    fun _endpos__5_ _endpos_t_ _startpos__1_ _startpos_t_ t ty ->
      let _endpos = _endpos__5_ in
      let _startpos = _startpos__1_ in
      let _loc_t_ = (_startpos_t_, _endpos_t_) in
      let _loc = (_startpos, _endpos) in
      (
# 398 "src/parser/grammar.mly"
                                                     ( mkCast (loc _loc) (mkConst (loc _loc_t_) t) ty )
# 3419 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_135 =
    fun _endpos__5_ _startpos__1_ t ty ->
      let _endpos = _endpos__5_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 399 "src/parser/grammar.mly"
                                                 ( mkCast (loc _loc) t ty )
# 3430 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_136 =
    fun _endpos__2_ _startpos__1_ ->
      let _endpos = _endpos__2_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 298 "src/parser/grammar.mly"
  ( Program.Ignored (loc _sloc) )
# 3441 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_137 =
    fun _endpos__2_ _startpos__1_ ->
      let _endpos = _endpos__2_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 298 "src/parser/grammar.mly"
  ( Program.Ignored (loc _sloc) )
# 3452 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_138 =
    fun _endpos__2_ _startpos__1_ ->
      let _endpos = _endpos__2_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 298 "src/parser/grammar.mly"
  ( Program.Ignored (loc _sloc) )
# 3463 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_139 =
    fun _endpos__3_ _startpos__1_ ->
      let _endpos = _endpos__3_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 298 "src/parser/grammar.mly"
  ( Program.Ignored (loc _sloc) )
# 3474 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_140 =
    fun _endpos__2_ _startpos__1_ ->
      let _endpos = _endpos__2_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 298 "src/parser/grammar.mly"
  ( Program.Ignored (loc _sloc) )
# 3485 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_141 =
    fun _endpos__2_ _startpos__1_ ->
      let _endpos = _endpos__2_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 298 "src/parser/grammar.mly"
  ( Program.Ignored (loc _sloc) )
# 3496 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_142 =
    fun _endpos__3_ _startpos__1_ ->
      let _endpos = _endpos__3_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 298 "src/parser/grammar.mly"
  ( Program.Ignored (loc _sloc) )
# 3507 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_143 =
    fun _endpos__2_ _startpos__1_ ->
      let _endpos = _endpos__2_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 298 "src/parser/grammar.mly"
  ( Program.Ignored (loc _sloc) )
# 3518 "src/parser/grammar.ml"
       : (Ast.Program.decl))
  
  let _menhir_action_144 =
    fun () ->
      (
# 367 "src/parser/grammar.mly"
          ( 0 )
# 3526 "src/parser/grammar.ml"
       : (int))
  
  let _menhir_action_145 =
    fun i ->
      (
# 368 "src/parser/grammar.mly"
              ( i )
# 3534 "src/parser/grammar.ml"
       : (int))
  
  let _menhir_action_146 =
    fun s ->
      let x =
        let x = 
# 530 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 3543 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3548 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3554 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_147 =
    fun s ->
      let x =
        let x = 
# 531 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 3563 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3568 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3574 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_148 =
    fun s ->
      let x =
        let x = 
# 532 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 3583 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3588 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3594 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_149 =
    fun s ->
      let x =
        let x = 
# 533 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 3603 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3608 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3614 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_150 =
    fun s ->
      let x =
        let x = 
# 534 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 3623 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3628 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3634 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_151 =
    fun s ->
      let x =
        let x = 
# 535 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 3643 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3648 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3654 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_152 =
    fun s ->
      let x =
        let x = 
# 536 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 3663 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3668 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3674 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_153 =
    fun s ->
      let x =
        let x = 
# 537 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 3683 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3688 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3694 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_154 =
    fun s ->
      let x =
        let x = 
# 538 "src/parser/grammar.mly"
                  ( Func.from_string s )
# 3703 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3708 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3714 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_155 =
    fun s ->
      let x =
        let x = 
# 539 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 3723 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3728 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3734 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_156 =
    fun s ->
      let x =
        let x = 
# 540 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 3743 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3748 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3754 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_157 =
    fun s ->
      let x =
        let x = 
# 541 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 3763 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 3768 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3774 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_158 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 544 "src/parser/grammar.mly"
         ( Func.consf )
# 3785 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 3790 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 3796 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 3802 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3808 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_159 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 545 "src/parser/grammar.mly"
         ( Func.eqf )
# 3819 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 3824 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 3830 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 3836 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3842 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_160 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 546 "src/parser/grammar.mly"
         ( Func.from_string "-" )
# 3853 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 3858 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 3864 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 3870 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3876 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_161 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 547 "src/parser/grammar.mly"
         ( Func.from_string "r-" )
# 3887 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 3892 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 3898 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 3904 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3910 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_162 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 548 "src/parser/grammar.mly"
         ( Func.from_string "i-" )
# 3921 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 3926 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 3932 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 3938 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3944 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_163 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 549 "src/parser/grammar.mly"
         ( Func.from_string "s-" )
# 3955 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 3960 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 3966 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 3972 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 3978 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_164 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 550 "src/parser/grammar.mly"
         ( Func.from_string "==" )
# 3989 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 3994 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4000 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4006 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4012 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_165 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 551 "src/parser/grammar.mly"
         ( Func.orf )
# 4023 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4028 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4034 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4040 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4046 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_166 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 552 "src/parser/grammar.mly"
         ( Func.isf )
# 4057 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4062 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4068 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4074 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4080 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_167 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 553 "src/parser/grammar.mly"
         ( Func.from_string "mod" )
# 4091 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4096 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4102 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4108 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4114 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_168 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 554 "src/parser/grammar.mly"
         ( Func.from_string "div" )
# 4125 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4130 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4136 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4142 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4148 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_169 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 555 "src/parser/grammar.mly"
         ( Func.arrowf )
# 4159 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4164 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4170 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4176 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4182 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_170 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 556 "src/parser/grammar.mly"
         ( Func.implf )
# 4193 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4198 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4204 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4210 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4216 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_171 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 557 "src/parser/grammar.mly"
          ( Func.implf )
# 4227 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4232 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4238 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4244 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4250 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_172 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 558 "src/parser/grammar.mly"
         ( Func.sequentf )
# 4261 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4266 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4272 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4278 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4284 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_173 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 559 "src/parser/grammar.mly"
         ( Func.from_string "/" )
# 4295 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4300 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4306 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4312 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4318 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_174 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 560 "src/parser/grammar.mly"
         ( Func.andf )
# 4329 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4334 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4340 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4346 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4352 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_175 =
    fun () ->
      let x =
        let x =
          let x = 
# 564 "src/parser/grammar.mly"
         ( Func.andf )
# 4362 "src/parser/grammar.ml"
           in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4367 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4373 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4379 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_176 =
    fun () ->
      let x =
        let x = 
# 572 "src/parser/grammar.mly"
         ( Func.rimplf )
# 4388 "src/parser/grammar.ml"
         in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4393 "src/parser/grammar.ml"
        
      in
      (
# 503 "src/parser/grammar.mly"
            ( x )
# 4399 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_177 =
    fun _endpos_k_ _startpos__1_ k names ->
      let _endpos = _endpos_k_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 238 "src/parser/grammar.mly"
                                                                      (
    names |> List.map (fun n->
     { Type.loc=loc _sloc; attributes=[]; name =  n; ty = k })
  )
# 4413 "src/parser/grammar.ml"
       : ((Ast.raw_attribute list, Ast.raw_attribute list) Ast.Type.t list))
  
  let _menhir_action_178 =
    fun _endpos__1_ _startpos__1_ ->
      let _endpos = _endpos__1_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 263 "src/parser/grammar.mly"
       ( { tloc = loc _loc; tit = TConst (Func.from_string "type") } )
# 4424 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_179 =
    fun _endpos_t_ _endpos_x_ _startpos_x_ t x ->
      let _endpos = _endpos_t_ in
      let _startpos = _startpos_x_ in
      let _loc_x_ = (_startpos_x_, _endpos_x_) in
      let _loc = (_startpos, _endpos) in
      (
# 264 "src/parser/grammar.mly"
                                 ( { tloc = loc _loc; tit = TArr ({ tloc = loc _loc_x_; tit = TConst (Func.from_string "type") }, t) } )
# 4436 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_180 =
    fun () ->
      (
# 216 "<standard.mly>"
    ( [] )
# 4444 "src/parser/grammar.ml"
       : (Ast.Func.t list))
  
  let _menhir_action_181 =
    fun x xs ->
      (
# 219 "<standard.mly>"
    ( x :: xs )
# 4452 "src/parser/grammar.ml"
       : (Ast.Func.t list))
  
  let _menhir_action_182 =
    fun () ->
      (
# 216 "<standard.mly>"
    ( [] )
# 4460 "src/parser/grammar.ml"
       : (Ast.Term.t Ast.Chr.sequent list))
  
  let _menhir_action_183 =
    fun x xs ->
      (
# 219 "<standard.mly>"
    ( x :: xs )
# 4468 "src/parser/grammar.ml"
       : (Ast.Term.t Ast.Chr.sequent list))
  
  let _menhir_action_184 =
    fun _endpos__1_ _startpos__1_ ->
      let _endpos = _endpos__1_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 402 "src/parser/grammar.mly"
           ( [mkNil (loc _loc)] )
# 4479 "src/parser/grammar.ml"
       : (Ast.Term.t list))
  
  let _menhir_action_185 =
    fun _endpos_r_ _startpos_r_ r x ->
      let _loc_r_ = (_startpos_r_, _endpos_r_) in
      (
# 403 "src/parser/grammar.mly"
                                ( [x;mkNil (loc _loc_r_)] )
# 4488 "src/parser/grammar.ml"
       : (Ast.Term.t list))
  
  let _menhir_action_186 =
    fun x xs ->
      (
# 404 "src/parser/grammar.mly"
                                         ( x :: xs )
# 4496 "src/parser/grammar.ml"
       : (Ast.Term.t list))
  
  let _menhir_action_187 =
    fun x ->
      (
# 407 "src/parser/grammar.mly"
                                  ( [x] )
# 4504 "src/parser/grammar.ml"
       : (Ast.Term.t list))
  
  let _menhir_action_188 =
    fun x xs ->
      (
# 408 "src/parser/grammar.mly"
                                                    ( [x;xs] )
# 4512 "src/parser/grammar.ml"
       : (Ast.Term.t list))
  
  let _menhir_action_189 =
    fun x xs ->
      (
# 409 "src/parser/grammar.mly"
                                              ( x :: xs )
# 4520 "src/parser/grammar.ml"
       : (Ast.Term.t list))
  
  let _menhir_action_190 =
    fun () ->
      (
# 145 "<standard.mly>"
    ( [] )
# 4528 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t list))
  
  let _menhir_action_191 =
    fun x ->
      (
# 148 "<standard.mly>"
    ( x )
# 4536 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t list))
  
  let _menhir_action_192 =
    fun () ->
      (
# 145 "<standard.mly>"
    ( [] )
# 4544 "src/parser/grammar.ml"
       : ((Ast.Mode.t * Ast.raw_attribute list Ast.TypeExpression.t) list))
  
  let _menhir_action_193 =
    fun x ->
      (
# 148 "<standard.mly>"
    ( x )
# 4552 "src/parser/grammar.ml"
       : ((Ast.Mode.t * Ast.raw_attribute list Ast.TypeExpression.t) list))
  
  let _menhir_action_194 =
    fun _endpos_b_ _startpos__1_ b m ->
      let _endpos = _endpos_b_ in
      let _symbolstartpos = _startpos__1_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 267 "src/parser/grammar.mly"
                                   (
  let name, body = desugar_macro (loc _sloc) m b in
  { Macro.loc = loc _sloc; name; body }
)
# 4566 "src/parser/grammar.ml"
       : ((Ast.Func.t, Ast.Term.t) Ast.Macro.t))
  
  let _menhir_action_195 =
    fun s ->
      let x =
        let x = 
# 530 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 4575 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4580 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4586 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_196 =
    fun s ->
      let x =
        let x = 
# 531 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 4595 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4600 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4606 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_197 =
    fun s ->
      let x =
        let x = 
# 532 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 4615 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4620 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4626 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_198 =
    fun s ->
      let x =
        let x = 
# 533 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 4635 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4640 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4646 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_199 =
    fun s ->
      let x =
        let x = 
# 534 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 4655 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4660 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4666 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_200 =
    fun s ->
      let x =
        let x = 
# 535 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 4675 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4680 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4686 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_201 =
    fun s ->
      let x =
        let x = 
# 536 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 4695 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4700 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4706 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_202 =
    fun s ->
      let x =
        let x = 
# 537 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 4715 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4720 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4726 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_203 =
    fun s ->
      let x =
        let x = 
# 538 "src/parser/grammar.mly"
                  ( Func.from_string s )
# 4735 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4740 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4746 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_204 =
    fun s ->
      let x =
        let x = 
# 539 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 4755 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4760 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4766 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_205 =
    fun s ->
      let x =
        let x = 
# 540 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 4775 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4780 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4786 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_206 =
    fun s ->
      let x =
        let x = 
# 541 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 4795 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 4800 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4806 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_207 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 544 "src/parser/grammar.mly"
         ( Func.consf )
# 4817 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4822 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4828 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4834 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4840 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_208 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 545 "src/parser/grammar.mly"
         ( Func.eqf )
# 4851 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4856 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4862 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4868 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4874 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_209 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 546 "src/parser/grammar.mly"
         ( Func.from_string "-" )
# 4885 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4890 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4896 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4902 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4908 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_210 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 547 "src/parser/grammar.mly"
         ( Func.from_string "r-" )
# 4919 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4924 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4930 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4936 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4942 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_211 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 548 "src/parser/grammar.mly"
         ( Func.from_string "i-" )
# 4953 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4958 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4964 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 4970 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 4976 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_212 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 549 "src/parser/grammar.mly"
         ( Func.from_string "s-" )
# 4987 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 4992 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 4998 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5004 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5010 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_213 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 550 "src/parser/grammar.mly"
         ( Func.from_string "==" )
# 5021 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5026 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5032 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5038 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5044 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_214 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 551 "src/parser/grammar.mly"
         ( Func.orf )
# 5055 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5060 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5066 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5072 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5078 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_215 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 552 "src/parser/grammar.mly"
         ( Func.isf )
# 5089 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5094 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5100 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5106 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5112 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_216 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 553 "src/parser/grammar.mly"
         ( Func.from_string "mod" )
# 5123 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5128 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5134 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5140 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5146 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_217 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 554 "src/parser/grammar.mly"
         ( Func.from_string "div" )
# 5157 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5162 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5168 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5174 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5180 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_218 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 555 "src/parser/grammar.mly"
         ( Func.arrowf )
# 5191 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5196 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5202 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5208 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5214 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_219 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 556 "src/parser/grammar.mly"
         ( Func.implf )
# 5225 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5230 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5236 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5242 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5248 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_220 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 557 "src/parser/grammar.mly"
          ( Func.implf )
# 5259 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5264 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5270 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5276 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5282 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_221 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 558 "src/parser/grammar.mly"
         ( Func.sequentf )
# 5293 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5298 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5304 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5310 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5316 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_222 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 559 "src/parser/grammar.mly"
         ( Func.from_string "/" )
# 5327 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5332 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5338 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5344 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5350 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_223 =
    fun () ->
      let x =
        let x =
          let x =
            let x = 
# 560 "src/parser/grammar.mly"
         ( Func.andf )
# 5361 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5366 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5372 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5378 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5384 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_224 =
    fun () ->
      let x =
        let x =
          let x = 
# 564 "src/parser/grammar.mly"
         ( Func.andf )
# 5394 "src/parser/grammar.ml"
           in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5399 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5405 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5411 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_225 =
    fun () ->
      let x =
        let x = 
# 572 "src/parser/grammar.mly"
         ( Func.rimplf )
# 5420 "src/parser/grammar.ml"
         in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5425 "src/parser/grammar.ml"
        
      in
      (
# 496 "src/parser/grammar.mly"
            ( x )
# 5431 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_226 =
    fun s ->
      let x =
        let x = 
# 575 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5440 "src/parser/grammar.ml"
         in
        
# 512 "src/parser/grammar.mly"
                        ( x )
# 5445 "src/parser/grammar.ml"
        
      in
      (
# 497 "src/parser/grammar.mly"
             ( x )
# 5451 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_227 =
    fun s ->
      let x =
        let x = 
# 578 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5460 "src/parser/grammar.ml"
         in
        
# 515 "src/parser/grammar.mly"
                         ( x )
# 5465 "src/parser/grammar.ml"
        
      in
      (
# 498 "src/parser/grammar.mly"
              ( x )
# 5471 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_228 =
    fun () ->
      (
# 499 "src/parser/grammar.mly"
     ( Func.from_string "pi" )
# 5479 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_229 =
    fun () ->
      (
# 500 "src/parser/grammar.mly"
        ( Func.from_string "sigma" )
# 5487 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_230 =
    fun x ->
      (
# 228 "<standard.mly>"
    ( [ x ] )
# 5495 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t list))
  
  let _menhir_action_231 =
    fun x xs ->
      (
# 231 "<standard.mly>"
    ( x :: xs )
# 5503 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t list))
  
  let _menhir_action_232 =
    fun x ->
      (
# 228 "<standard.mly>"
    ( [ x ] )
# 5511 "src/parser/grammar.ml"
       : (Ast.Term.t list))
  
  let _menhir_action_233 =
    fun x xs ->
      (
# 231 "<standard.mly>"
    ( x :: xs )
# 5519 "src/parser/grammar.ml"
       : (Ast.Term.t list))
  
  let _menhir_action_234 =
    fun x ->
      (
# 228 "<standard.mly>"
    ( [ x ] )
# 5527 "src/parser/grammar.ml"
       : ((Ast.Func.t * Ast.Loc.t) list))
  
  let _menhir_action_235 =
    fun x xs ->
      (
# 231 "<standard.mly>"
    ( x :: xs )
# 5535 "src/parser/grammar.ml"
       : ((Ast.Func.t * Ast.Loc.t) list))
  
  let _menhir_action_236 =
    fun x ->
      (
# 228 "<standard.mly>"
    ( [ x ] )
# 5543 "src/parser/grammar.ml"
       : (int list))
  
  let _menhir_action_237 =
    fun x xs ->
      (
# 231 "<standard.mly>"
    ( x :: xs )
# 5551 "src/parser/grammar.ml"
       : (int list))
  
  let _menhir_action_238 =
    fun x ->
      (
# 228 "<standard.mly>"
    ( [ x ] )
# 5559 "src/parser/grammar.ml"
       : (Ast.Term.t Ast.Chr.sequent list))
  
  let _menhir_action_239 =
    fun x xs ->
      (
# 231 "<standard.mly>"
    ( x :: xs )
# 5567 "src/parser/grammar.ml"
       : (Ast.Term.t Ast.Chr.sequent list))
  
  let _menhir_action_240 =
    fun _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_hd_ in
      let _loc_hd_ = (_startpos_hd_, _endpos_hd_) in
      let _loc = (_startpos, _endpos) in
      (
# 427 "src/parser/grammar.mly"
                                                                 ( desugar_multi_binder (loc _loc) @@ mkApp (loc _loc) (mkConst (loc _loc_hd_) (Func.from_string "pi") :: binder args b) )
# 5579 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_241 =
    fun _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_hd_ in
      let _loc_hd_ = (_startpos_hd_, _endpos_hd_) in
      let _loc = (_startpos, _endpos) in
      (
# 428 "src/parser/grammar.mly"
                                                                    ( desugar_multi_binder (loc _loc) @@ mkApp (loc _loc) (mkConst (loc _loc_hd_) (Func.from_string "sigma") :: binder args b) )
# 5591 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_242 =
    fun _endpos_b_ _startpos_hd_ args b hd ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_hd_ in
      let _loc = (_startpos, _endpos) in
      (
# 429 "src/parser/grammar.mly"
                                                                                    (
    let args = binder1 args b in
    let t = mkApp (loc _loc) (hd :: args) in
    t
)
# 5606 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_243 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 530 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5615 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5620 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5630 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_244 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 531 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5639 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5644 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5654 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_245 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 532 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5663 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5668 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5678 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_246 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 533 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5687 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5692 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5702 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_247 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 534 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5711 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5716 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5726 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_248 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 535 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5735 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5740 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5750 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_249 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 536 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5759 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5764 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5774 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_250 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 537 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5783 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5788 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5798 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_251 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 538 "src/parser/grammar.mly"
                  ( Func.from_string s )
# 5807 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5812 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5822 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_252 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 539 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5831 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5836 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5846 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_253 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 540 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5855 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5860 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5870 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_254 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 541 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 5879 "src/parser/grammar.ml"
         in
        
# 518 "src/parser/grammar.mly"
                       ( x )
# 5884 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5894 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_255 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 544 "src/parser/grammar.mly"
         ( Func.consf )
# 5905 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5910 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5916 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5922 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5933 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_256 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 545 "src/parser/grammar.mly"
         ( Func.eqf )
# 5944 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5949 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5955 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 5961 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 5972 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_257 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 546 "src/parser/grammar.mly"
         ( Func.from_string "-" )
# 5983 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 5988 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 5994 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6000 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6011 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_258 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 547 "src/parser/grammar.mly"
         ( Func.from_string "r-" )
# 6022 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6027 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6033 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6039 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6050 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_259 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 548 "src/parser/grammar.mly"
         ( Func.from_string "i-" )
# 6061 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6066 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6072 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6078 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6089 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_260 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 549 "src/parser/grammar.mly"
         ( Func.from_string "s-" )
# 6100 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6105 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6111 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6117 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6128 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_261 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 550 "src/parser/grammar.mly"
         ( Func.from_string "==" )
# 6139 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6144 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6150 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6156 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6167 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_262 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 551 "src/parser/grammar.mly"
         ( Func.orf )
# 6178 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6183 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6189 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6195 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6206 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_263 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 552 "src/parser/grammar.mly"
         ( Func.isf )
# 6217 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6222 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6228 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6234 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6245 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_264 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 553 "src/parser/grammar.mly"
         ( Func.from_string "mod" )
# 6256 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6261 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6267 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6273 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6284 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_265 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 554 "src/parser/grammar.mly"
         ( Func.from_string "div" )
# 6295 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6300 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6306 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6312 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6323 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_266 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 555 "src/parser/grammar.mly"
         ( Func.arrowf )
# 6334 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6339 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6345 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6351 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6362 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_267 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 556 "src/parser/grammar.mly"
         ( Func.implf )
# 6373 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6378 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6384 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6390 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6401 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_268 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 557 "src/parser/grammar.mly"
          ( Func.implf )
# 6412 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6417 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6423 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6429 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6440 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_269 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 558 "src/parser/grammar.mly"
         ( Func.sequentf )
# 6451 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6456 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6462 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6468 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6479 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_270 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 559 "src/parser/grammar.mly"
         ( Func.from_string "/" )
# 6490 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6495 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6501 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6507 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6518 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_271 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x =
            let x = 
# 560 "src/parser/grammar.mly"
         ( Func.andf )
# 6529 "src/parser/grammar.ml"
             in
            
# 563 "src/parser/grammar.mly"
                                          ( x )
# 6534 "src/parser/grammar.ml"
            
          in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6540 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6546 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6557 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_272 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 564 "src/parser/grammar.mly"
         ( Func.andf )
# 6567 "src/parser/grammar.ml"
           in
          
# 571 "src/parser/grammar.mly"
                                   ( x )
# 6572 "src/parser/grammar.ml"
          
        in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6578 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6589 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_273 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x = 
# 572 "src/parser/grammar.mly"
         ( Func.rimplf )
# 6598 "src/parser/grammar.ml"
         in
        
# 519 "src/parser/grammar.mly"
                           ( x )
# 6603 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 434 "src/parser/grammar.mly"
                                 ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6614 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_274 =
    fun _endpos_r_ _endpos_s_ _startpos_s_ r s ->
      let s =
        let x = 
# 575 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6623 "src/parser/grammar.ml"
         in
        
# 512 "src/parser/grammar.mly"
                        ( x )
# 6628 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_s_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 435 "src/parser/grammar.mly"
                       ( mkAppF (loc _loc) (loc _loc_s_,s) [r] )
# 6638 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_275 =
    fun _endpos_s_ _startpos_l_ _startpos_s_ l s ->
      let s =
        let x = 
# 578 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6647 "src/parser/grammar.ml"
         in
        
# 515 "src/parser/grammar.mly"
                         ( x )
# 6652 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_s_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 436 "src/parser/grammar.mly"
                         ( mkAppF (loc _loc) (loc _loc_s_,s) [l] )
# 6662 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_276 =
    fun _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_hd_ in
      let _loc_hd_ = (_startpos_hd_, _endpos_hd_) in
      let _loc = (_startpos, _endpos) in
      (
# 439 "src/parser/grammar.mly"
                                                                 ( desugar_multi_binder (loc _loc) @@ mkApp (loc _loc) (mkConst (loc _loc_hd_) (Func.from_string "pi") :: binder args b) )
# 6674 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_277 =
    fun _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_hd_ in
      let _loc_hd_ = (_startpos_hd_, _endpos_hd_) in
      let _loc = (_startpos, _endpos) in
      (
# 440 "src/parser/grammar.mly"
                                                                    ( desugar_multi_binder (loc _loc) @@ mkApp (loc _loc) (mkConst (loc _loc_hd_) (Func.from_string "sigma") :: binder args b) )
# 6686 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_278 =
    fun _endpos_b_ _startpos_hd_ args b hd ->
      let _endpos = _endpos_b_ in
      let _startpos = _startpos_hd_ in
      let _loc = (_startpos, _endpos) in
      (
# 441 "src/parser/grammar.mly"
                                                                                   (
    let args = binder1 args b in
    let t = mkApp (loc _loc) (hd :: args) in
    t
)
# 6701 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_279 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 530 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6710 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6715 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6725 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_280 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 531 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6734 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6739 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6749 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_281 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 532 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6758 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6763 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6773 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_282 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 533 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6782 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6787 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6797 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_283 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 534 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6806 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6811 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6821 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_284 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 535 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6830 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6835 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6845 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_285 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 536 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6854 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6859 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6869 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_286 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 537 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6878 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6883 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6893 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_287 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 538 "src/parser/grammar.mly"
                  ( Func.from_string s )
# 6902 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6907 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6917 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_288 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 539 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6926 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6931 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6941 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_289 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 540 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6950 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6955 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6965 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_290 =
    fun _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s ->
      let s =
        let x = 
# 541 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 6974 "src/parser/grammar.ml"
         in
        
# 522 "src/parser/grammar.mly"
                       ( x )
# 6979 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 6989 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_291 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 544 "src/parser/grammar.mly"
         ( Func.consf )
# 6999 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7004 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7010 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7021 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_292 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 545 "src/parser/grammar.mly"
         ( Func.eqf )
# 7031 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7036 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7042 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7053 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_293 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 546 "src/parser/grammar.mly"
         ( Func.from_string "-" )
# 7063 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7068 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7074 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7085 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_294 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 547 "src/parser/grammar.mly"
         ( Func.from_string "r-" )
# 7095 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7100 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7106 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7117 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_295 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 548 "src/parser/grammar.mly"
         ( Func.from_string "i-" )
# 7127 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7132 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7138 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7149 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_296 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 549 "src/parser/grammar.mly"
         ( Func.from_string "s-" )
# 7159 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7164 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7170 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7181 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_297 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 550 "src/parser/grammar.mly"
         ( Func.from_string "==" )
# 7191 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7196 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7202 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7213 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_298 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 551 "src/parser/grammar.mly"
         ( Func.orf )
# 7223 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7228 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7234 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7245 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_299 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 552 "src/parser/grammar.mly"
         ( Func.isf )
# 7255 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7260 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7266 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7277 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_300 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 553 "src/parser/grammar.mly"
         ( Func.from_string "mod" )
# 7287 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7292 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7298 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7309 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_301 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 554 "src/parser/grammar.mly"
         ( Func.from_string "div" )
# 7319 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7324 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7330 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7341 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_302 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 555 "src/parser/grammar.mly"
         ( Func.arrowf )
# 7351 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7356 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7362 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7373 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_303 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 556 "src/parser/grammar.mly"
         ( Func.implf )
# 7383 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7388 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7394 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7405 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_304 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 557 "src/parser/grammar.mly"
          ( Func.implf )
# 7415 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7420 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7426 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7437 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_305 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 558 "src/parser/grammar.mly"
         ( Func.sequentf )
# 7447 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7452 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7458 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7469 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_306 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 559 "src/parser/grammar.mly"
         ( Func.from_string "/" )
# 7479 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7484 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7490 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7501 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_307 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x =
          let x = 
# 560 "src/parser/grammar.mly"
         ( Func.andf )
# 7511 "src/parser/grammar.ml"
           in
          
# 567 "src/parser/grammar.mly"
                                          ( x )
# 7516 "src/parser/grammar.ml"
          
        in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7522 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7533 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_308 =
    fun _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r ->
      let s =
        let x = 
# 568 "src/parser/grammar.mly"
         ( Func.rimplf )
# 7542 "src/parser/grammar.ml"
         in
        
# 523 "src/parser/grammar.mly"
                                  ( x )
# 7547 "src/parser/grammar.ml"
        
      in
      let (_endpos_s_, _startpos_s_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 446 "src/parser/grammar.mly"
                                                      ( mkAppF (loc _loc) (loc _loc_s_,s) [l;r] )
# 7558 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_309 =
    fun _endpos_r_ _endpos_s_ _startpos_s_ r s ->
      let s =
        let x = 
# 575 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 7567 "src/parser/grammar.ml"
         in
        
# 512 "src/parser/grammar.mly"
                        ( x )
# 7572 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_r_ in
      let _startpos = _startpos_s_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 447 "src/parser/grammar.mly"
                              ( mkAppF (loc _loc) (loc _loc_s_,s) [r] )
# 7582 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_310 =
    fun _endpos_s_ _startpos_l_ _startpos_s_ l s ->
      let s =
        let x = 
# 578 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 7591 "src/parser/grammar.ml"
         in
        
# 515 "src/parser/grammar.mly"
                         ( x )
# 7596 "src/parser/grammar.ml"
        
      in
      let _endpos = _endpos_s_ in
      let _startpos = _startpos_l_ in
      let _loc_s_ = (_startpos_s_, _endpos_s_) in
      let _loc = (_startpos, _endpos) in
      (
# 448 "src/parser/grammar.mly"
                                ( mkAppF (loc _loc) (loc _loc_s_,s) [l] )
# 7606 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_311 =
    fun () ->
      (
# 111 "<standard.mly>"
    ( None )
# 7614 "src/parser/grammar.ml"
       : (unit option))
  
  let _menhir_action_312 =
    fun x ->
      (
# 114 "<standard.mly>"
    ( Some x )
# 7622 "src/parser/grammar.ml"
       : (unit option))
  
  let _menhir_action_313 =
    fun () ->
      (
# 111 "<standard.mly>"
    ( None )
# 7630 "src/parser/grammar.ml"
       : (string option))
  
  let _menhir_action_314 =
    fun x ->
      (
# 114 "<standard.mly>"
    ( Some x )
# 7638 "src/parser/grammar.ml"
       : (string option))
  
  let _menhir_action_315 =
    fun () ->
      (
# 111 "<standard.mly>"
    ( None )
# 7646 "src/parser/grammar.ml"
       : ((Ast.Loc.t * Ast.Term.typ option * Ast.Term.t) option))
  
  let _menhir_action_316 =
    fun x ->
      (
# 114 "<standard.mly>"
    ( Some x )
# 7654 "src/parser/grammar.ml"
       : ((Ast.Loc.t * Ast.Term.typ option * Ast.Term.t) option))
  
  let _menhir_action_317 =
    fun () ->
      (
# 111 "<standard.mly>"
    ( None )
# 7662 "src/parser/grammar.ml"
       : (Ast.Term.t Ast.Chr.sequent list option))
  
  let _menhir_action_318 =
    fun x ->
      let x = 
# 188 "<standard.mly>"
    ( x )
# 7670 "src/parser/grammar.ml"
       in
      (
# 114 "<standard.mly>"
    ( Some x )
# 7675 "src/parser/grammar.ml"
       : (Ast.Term.t Ast.Chr.sequent list option))
  
  let _menhir_action_319 =
    fun () ->
      (
# 111 "<standard.mly>"
    ( None )
# 7683 "src/parser/grammar.ml"
       : (Ast.Term.t Ast.Chr.sequent option))
  
  let _menhir_action_320 =
    fun x ->
      let x = 
# 188 "<standard.mly>"
    ( x )
# 7691 "src/parser/grammar.ml"
       in
      (
# 114 "<standard.mly>"
    ( Some x )
# 7696 "src/parser/grammar.ml"
       : (Ast.Term.t Ast.Chr.sequent option))
  
  let _menhir_action_321 =
    fun () ->
      (
# 111 "<standard.mly>"
    ( None )
# 7704 "src/parser/grammar.ml"
       : (Ast.Term.t option))
  
  let _menhir_action_322 =
    fun x ->
      let x = 
# 188 "<standard.mly>"
    ( x )
# 7712 "src/parser/grammar.ml"
       in
      (
# 114 "<standard.mly>"
    ( Some x )
# 7717 "src/parser/grammar.ml"
       : (Ast.Term.t option))
  
  let _menhir_action_323 =
    fun () ->
      (
# 111 "<standard.mly>"
    ( None )
# 7725 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t option))
  
  let _menhir_action_324 =
    fun x ->
      (
# 114 "<standard.mly>"
    ( Some x )
# 7733 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t option))
  
  let _menhir_action_325 =
    fun s ->
      let x =
        let x = 
# 578 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 7742 "src/parser/grammar.ml"
         in
        
# 515 "src/parser/grammar.mly"
                         ( x )
# 7747 "src/parser/grammar.ml"
        
      in
      (
# 509 "src/parser/grammar.mly"
              ( x )
# 7753 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_326 =
    fun _endpos_attributes_ _endpos_xs_ _startpos__2_ _startpos_attributes_ attributes name xs ->
      let args = 
# 241 "<standard.mly>"
    ( xs )
# 7761 "src/parser/grammar.ml"
       in
      let _endpos_args_ = _endpos_xs_ in
      let _endpos = _endpos_args_ in
      let _startpos = _startpos_attributes_ in
      let _symbolstartpos = if _startpos_attributes_ != _endpos_attributes_ then
        _startpos_attributes_
      else
        _startpos__2_ in
      let _loc = (_startpos, _endpos) in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 209 "src/parser/grammar.mly"
                                                                 (
   { Type.loc=loc _sloc; name; attributes; ty = { tloc = loc _loc; tit = TPred ([], args) } }
 )
# 7777 "src/parser/grammar.ml"
       : ((Ast.raw_attribute list, Ast.raw_attribute list) Ast.Type.t))
  
  let _menhir_action_327 =
    fun _endpos_attributes_ _endpos_xs_inlined1_ _startpos__2_ _startpos_attributes_ attributes name xs xs_inlined1 ->
      let out_args =
        let xs = xs_inlined1 in
        
# 241 "<standard.mly>"
    ( xs )
# 7787 "src/parser/grammar.ml"
        
      in
      let _endpos_out_args_ = _endpos_xs_inlined1_ in
      let in_args = 
# 241 "<standard.mly>"
    ( xs )
# 7794 "src/parser/grammar.ml"
       in
      let _endpos = _endpos_out_args_ in
      let _startpos = _startpos_attributes_ in
      let _symbolstartpos = if _startpos_attributes_ != _endpos_attributes_ then
        _startpos_attributes_
      else
        _startpos__2_ in
      let _loc = (_startpos, _endpos) in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 213 "src/parser/grammar.mly"
                                                                                                                  (
    let args = List.map (fun x -> Mode.Input,x) in_args @ List.map (fun x -> Mode.Output,x) out_args in
    { Type.loc=loc _sloc; name; attributes; ty = { tloc = loc _loc; tit = TPred ([Functional], args) } }
  )
# 7810 "src/parser/grammar.ml"
       : ((Ast.raw_attribute list, Ast.raw_attribute list) Ast.Type.t))
  
  let _menhir_action_328 =
    fun _endpos_attributes_ _endpos_xs_ _startpos__2_ _startpos_attributes_ attributes name xs ->
      let in_args = 
# 241 "<standard.mly>"
    ( xs )
# 7818 "src/parser/grammar.ml"
       in
      let _endpos_in_args_ = _endpos_xs_ in
      let _endpos = _endpos_in_args_ in
      let _startpos = _startpos_attributes_ in
      let _symbolstartpos = if _startpos_attributes_ != _endpos_attributes_ then
        _startpos_attributes_
      else
        _startpos__2_ in
      let _loc = (_startpos, _endpos) in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 218 "src/parser/grammar.mly"
                                                              (
  let args = List.map (fun x -> Mode.Input,x) in_args in
  { Type.loc=loc _sloc; name; attributes; ty = { tloc = loc _loc; tit = TPred ([Functional], args) } }
)
# 7835 "src/parser/grammar.ml"
       : ((Ast.raw_attribute list, Ast.raw_attribute list) Ast.Type.t))
  
  let _menhir_action_329 =
    fun io ty ->
      (
# 224 "src/parser/grammar.mly"
                                ( (mode_of_IO io,ty) )
# 7843 "src/parser/grammar.ml"
       : (Ast.Mode.t * Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_330 =
    fun s ->
      let x =
        let x = 
# 575 "src/parser/grammar.mly"
                   ( Func.from_string s )
# 7852 "src/parser/grammar.ml"
         in
        
# 512 "src/parser/grammar.mly"
                        ( x )
# 7857 "src/parser/grammar.ml"
        
      in
      (
# 506 "src/parser/grammar.mly"
             ( x )
# 7863 "src/parser/grammar.ml"
       : (Ast.Func.t))
  
  let _menhir_action_331 =
    fun () ->
      (
# 161 "src/parser/grammar.mly"
      ( [] )
# 7871 "src/parser/grammar.ml"
       : (Ast.Program.t))
  
  let _menhir_action_332 =
    fun d p ->
      (
# 162 "src/parser/grammar.mly"
                        ( d :: p )
# 7879 "src/parser/grammar.ml"
       : (Ast.Program.t))
  
  let _menhir_action_333 =
    fun x ->
      (
# 250 "<standard.mly>"
    ( [ x ] )
# 7887 "src/parser/grammar.ml"
       : (Ast.raw_attribute list))
  
  let _menhir_action_334 =
    fun x xs ->
      (
# 253 "<standard.mly>"
    ( x :: xs )
# 7895 "src/parser/grammar.ml"
       : (Ast.raw_attribute list))
  
  let _menhir_action_335 =
    fun x ->
      (
# 250 "<standard.mly>"
    ( [ x ] )
# 7903 "src/parser/grammar.ml"
       : (Ast.Func.t list))
  
  let _menhir_action_336 =
    fun x xs ->
      (
# 253 "<standard.mly>"
    ( x :: xs )
# 7911 "src/parser/grammar.ml"
       : (Ast.Func.t list))
  
  let _menhir_action_337 =
    fun x ->
      (
# 250 "<standard.mly>"
    ( [ x ] )
# 7919 "src/parser/grammar.ml"
       : (string list))
  
  let _menhir_action_338 =
    fun x xs ->
      (
# 253 "<standard.mly>"
    ( x :: xs )
# 7927 "src/parser/grammar.ml"
       : (string list))
  
  let _menhir_action_339 =
    fun x ->
      (
# 250 "<standard.mly>"
    ( [ x ] )
# 7935 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t list))
  
  let _menhir_action_340 =
    fun x xs ->
      (
# 253 "<standard.mly>"
    ( x :: xs )
# 7943 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t list))
  
  let _menhir_action_341 =
    fun x ->
      (
# 250 "<standard.mly>"
    ( [ x ] )
# 7951 "src/parser/grammar.ml"
       : ((string * string) list list))
  
  let _menhir_action_342 =
    fun x xs ->
      (
# 253 "<standard.mly>"
    ( x :: xs )
# 7959 "src/parser/grammar.ml"
       : ((string * string) list list))
  
  let _menhir_action_343 =
    fun x ->
      (
# 250 "<standard.mly>"
    ( [ x ] )
# 7967 "src/parser/grammar.ml"
       : ((Ast.Mode.t * Ast.raw_attribute list Ast.TypeExpression.t) list))
  
  let _menhir_action_344 =
    fun x xs ->
      (
# 253 "<standard.mly>"
    ( x :: xs )
# 7975 "src/parser/grammar.ml"
       : ((Ast.Mode.t * Ast.raw_attribute list Ast.TypeExpression.t) list))
  
  let _menhir_action_345 =
    fun _endpos_t_ _startpos_t_ t ->
      let _endpos = _endpos_t_ in
      let _startpos = _startpos_t_ in
      let _loc = (_startpos, _endpos) in
      (
# 320 "src/parser/grammar.mly"
                  (
    let context, conclusion = decode_sequent (loc _loc) t in
    { Chr.eigen = underscore (loc _loc); context; conclusion }
  )
# 7989 "src/parser/grammar.ml"
       : (Ast.Term.t Ast.Chr.sequent))
  
  let _menhir_action_346 =
    fun _endpos__5_ _startpos__1_ c t ->
      let _endpos = _endpos__5_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 324 "src/parser/grammar.mly"
                                                  (
    let context, conclusion = decode_sequent (loc _loc) t in
    { Chr.eigen = c; context; conclusion }
  )
# 8003 "src/parser/grammar.ml"
       : (Ast.Term.t Ast.Chr.sequent))
  
  let _menhir_action_347 =
    fun l ->
      (
# 305 "src/parser/grammar.mly"
           (
     List.map Func.(fun (x,y) -> from_string x, from_string y) l
  )
# 8013 "src/parser/grammar.ml"
       : ((Ast.Func.t * Ast.Func.t) list))
  
  let _menhir_action_348 =
    fun name ->
      (
# 314 "src/parser/grammar.mly"
                  ( [Func.show name, Func.show name] )
# 8021 "src/parser/grammar.ml"
       : ((string * string) list))
  
  let _menhir_action_349 =
    fun l prefix ->
      (
# 315 "src/parser/grammar.mly"
                                                                                         (
    List.map (fun (p,x) -> Func.show prefix ^ "." ^ p, x) (List.flatten l)
)
# 8031 "src/parser/grammar.ml"
       : ((string * string) list))
  
  let _menhir_action_350 =
    fun t ->
      (
# 371 "src/parser/grammar.mly"
                ( t )
# 8039 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_351 =
    fun t ->
      (
# 372 "src/parser/grammar.mly"
                  ( t )
# 8047 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_352 =
    fun t ->
      (
# 373 "src/parser/grammar.mly"
                  ( t )
# 8055 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_353 =
    fun t ->
      (
# 376 "src/parser/grammar.mly"
                       ( t )
# 8063 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_354 =
    fun t ->
      (
# 377 "src/parser/grammar.mly"
                         ( t )
# 8071 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_355 =
    fun t ->
      (
# 378 "src/parser/grammar.mly"
                  ( t )
# 8079 "src/parser/grammar.ml"
       : (Ast.Term.t))
  
  let _menhir_action_356 =
    fun c l ->
      (
# 310 "src/parser/grammar.mly"
                                                                                    (
    List.map (fun (p,x) -> Func.show c ^ "." ^ p, x) (List.flatten l)
)
# 8089 "src/parser/grammar.ml"
       : ((string * string) list))
  
  let _menhir_action_357 =
    fun _endpos_attributes_ _endpos_t_ _startpos__2_ _startpos_attributes_ attributes names t ->
      let _endpos = _endpos_t_ in
      let _symbolstartpos = if _startpos_attributes_ != _endpos_attributes_ then
        _startpos_attributes_
      else
        _startpos__2_ in
      let _sloc = (_symbolstartpos, _endpos) in
      (
# 244 "src/parser/grammar.mly"
                                                                      (
    names |> List.map (fun n->
     { Type.loc=loc _sloc; attributes; name = n; ty = t })
  )
# 8106 "src/parser/grammar.ml"
       : ((Ast.raw_attribute list, Ast.raw_attribute list) Ast.Type.t list))
  
  let _menhir_action_358 =
    fun fo ->
      (
# 259 "src/parser/grammar.mly"
                   ( fo )
# 8114 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_359 =
    fun _endpos_t_ _startpos_hd_ hd t ->
      let _endpos = _endpos_t_ in
      let _startpos = _startpos_hd_ in
      let _loc = (_startpos, _endpos) in
      (
# 260 "src/parser/grammar.mly"
                                         ( { tloc = loc _loc; tit = TArr (hd, t) } )
# 8125 "src/parser/grammar.ml"
       : (Ast.raw_attribute list Ast.TypeExpression.t))
  
  let _menhir_action_360 =
    fun _endpos_t_ _startpos__1_ a t ->
      let _endpos = _endpos_t_ in
      let _startpos = _startpos__1_ in
      let _loc = (_startpos, _endpos) in
      (
# 273 "src/parser/grammar.mly"
                                            (
    let name, args = a in
    let nparams = List.length args in
    let mkLam (n,l) body =  TypeAbbreviation.Lam (n, l, body) in
    let value = List.fold_right mkLam args (Ty t) in
    { TypeAbbreviation.name = name;
      nparams = nparams;
      value = value;
      loc = loc _loc }
  )
# 8145 "src/parser/grammar.ml"
       : ((Ast.Func.t, Ast.raw_attribute list Ast.TypeExpression.t)
        Ast.TypeAbbreviation.t))
  
  let _menhir_print_token : token -> string =
    fun _tok ->
      match _tok with
      | Elpi_lexer_config.Tokens.ACCUMULATE ->
          "ACCUMULATE"
      | Elpi_lexer_config.Tokens.ACCUM_SIG ->
          "ACCUM_SIG"
      | Elpi_lexer_config.Tokens.AFTER ->
          "AFTER"
      | Elpi_lexer_config.Tokens.ARROW ->
          "ARROW"
      | Elpi_lexer_config.Tokens.AS ->
          "AS"
      | Elpi_lexer_config.Tokens.BEFORE ->
          "BEFORE"
      | Elpi_lexer_config.Tokens.BIND ->
          "BIND"
      | Elpi_lexer_config.Tokens.CLOSED ->
          "CLOSED"
      | Elpi_lexer_config.Tokens.COLON ->
          "COLON"
      | Elpi_lexer_config.Tokens.CONJ ->
          "CONJ"
      | Elpi_lexer_config.Tokens.CONJ2 ->
          "CONJ2"
      | Elpi_lexer_config.Tokens.CONS ->
          "CONS"
      | Elpi_lexer_config.Tokens.CONSTANT _ ->
          "CONSTANT"
      | Elpi_lexer_config.Tokens.CONSTRAINT ->
          "CONSTRAINT"
      | Elpi_lexer_config.Tokens.CUT ->
          "CUT"
      | Elpi_lexer_config.Tokens.DARROW ->
          "DARROW"
      | Elpi_lexer_config.Tokens.DDARROW ->
          "DDARROW"
      | Elpi_lexer_config.Tokens.DIV ->
          "DIV"
      | Elpi_lexer_config.Tokens.EOF ->
          "EOF"
      | Elpi_lexer_config.Tokens.EQ ->
          "EQ"
      | Elpi_lexer_config.Tokens.EQ2 ->
          "EQ2"
      | Elpi_lexer_config.Tokens.EXPORTDEF ->
          "EXPORTDEF"
      | Elpi_lexer_config.Tokens.EXTERNAL ->
          "EXTERNAL"
      | Elpi_lexer_config.Tokens.FAMILY_AND _ ->
          "FAMILY_AND"
      | Elpi_lexer_config.Tokens.FAMILY_BTICK _ ->
          "FAMILY_BTICK"
      | Elpi_lexer_config.Tokens.FAMILY_EQ _ ->
          "FAMILY_EQ"
      | Elpi_lexer_config.Tokens.FAMILY_EXP _ ->
          "FAMILY_EXP"
      | Elpi_lexer_config.Tokens.FAMILY_GT _ ->
          "FAMILY_GT"
      | Elpi_lexer_config.Tokens.FAMILY_LT _ ->
          "FAMILY_LT"
      | Elpi_lexer_config.Tokens.FAMILY_MINUS _ ->
          "FAMILY_MINUS"
      | Elpi_lexer_config.Tokens.FAMILY_OR _ ->
          "FAMILY_OR"
      | Elpi_lexer_config.Tokens.FAMILY_PLUS _ ->
          "FAMILY_PLUS"
      | Elpi_lexer_config.Tokens.FAMILY_QMARK _ ->
          "FAMILY_QMARK"
      | Elpi_lexer_config.Tokens.FAMILY_SHARP _ ->
          "FAMILY_SHARP"
      | Elpi_lexer_config.Tokens.FAMILY_TICK _ ->
          "FAMILY_TICK"
      | Elpi_lexer_config.Tokens.FAMILY_TILDE _ ->
          "FAMILY_TILDE"
      | Elpi_lexer_config.Tokens.FAMILY_TIMES _ ->
          "FAMILY_TIMES"
      | Elpi_lexer_config.Tokens.FIXITY _ ->
          "FIXITY"
      | Elpi_lexer_config.Tokens.FLOAT _ ->
          "FLOAT"
      | Elpi_lexer_config.Tokens.FRESHUV ->
          "FRESHUV"
      | Elpi_lexer_config.Tokens.FULLSTOP ->
          "FULLSTOP"
      | Elpi_lexer_config.Tokens.FUNC ->
          "FUNC"
      | Elpi_lexer_config.Tokens.FUNCTIONAL ->
          "FUNCTIONAL"
      | Elpi_lexer_config.Tokens.IF ->
          "IF"
      | Elpi_lexer_config.Tokens.IFF ->
          "IFF"
      | Elpi_lexer_config.Tokens.IMPORT ->
          "IMPORT"
      | Elpi_lexer_config.Tokens.INDEX ->
          "INDEX"
      | Elpi_lexer_config.Tokens.INTEGER _ ->
          "INTEGER"
      | Elpi_lexer_config.Tokens.IO _ ->
          "IO"
      | Elpi_lexer_config.Tokens.IO_COLON _ ->
          "IO_COLON"
      | Elpi_lexer_config.Tokens.IS ->
          "IS"
      | Elpi_lexer_config.Tokens.KIND ->
          "KIND"
      | Elpi_lexer_config.Tokens.LBRACKET ->
          "LBRACKET"
      | Elpi_lexer_config.Tokens.LCURLY ->
          "LCURLY"
      | Elpi_lexer_config.Tokens.LOCAL ->
          "LOCAL"
      | Elpi_lexer_config.Tokens.LOCALKIND ->
          "LOCALKIND"
      | Elpi_lexer_config.Tokens.LPAREN ->
          "LPAREN"
      | Elpi_lexer_config.Tokens.MACRO ->
          "MACRO"
      | Elpi_lexer_config.Tokens.MINUS ->
          "MINUS"
      | Elpi_lexer_config.Tokens.MINUSi ->
          "MINUSi"
      | Elpi_lexer_config.Tokens.MINUSr ->
          "MINUSr"
      | Elpi_lexer_config.Tokens.MINUSs ->
          "MINUSs"
      | Elpi_lexer_config.Tokens.MOD ->
          "MOD"
      | Elpi_lexer_config.Tokens.MODULE ->
          "MODULE"
      | Elpi_lexer_config.Tokens.NAME ->
          "NAME"
      | Elpi_lexer_config.Tokens.NAMESPACE ->
          "NAMESPACE"
      | Elpi_lexer_config.Tokens.NIL ->
          "NIL"
      | Elpi_lexer_config.Tokens.OR ->
          "OR"
      | Elpi_lexer_config.Tokens.PI ->
          "PI"
      | Elpi_lexer_config.Tokens.PIPE ->
          "PIPE"
      | Elpi_lexer_config.Tokens.PRED ->
          "PRED"
      | Elpi_lexer_config.Tokens.QDASH ->
          "QDASH"
      | Elpi_lexer_config.Tokens.QUOTED _ ->
          "QUOTED"
      | Elpi_lexer_config.Tokens.RBRACKET ->
          "RBRACKET"
      | Elpi_lexer_config.Tokens.RCURLY ->
          "RCURLY"
      | Elpi_lexer_config.Tokens.REMOVE ->
          "REMOVE"
      | Elpi_lexer_config.Tokens.REPLACE ->
          "REPLACE"
      | Elpi_lexer_config.Tokens.RPAREN ->
          "RPAREN"
      | Elpi_lexer_config.Tokens.RTRI ->
          "RTRI"
      | Elpi_lexer_config.Tokens.RULE ->
          "RULE"
      | Elpi_lexer_config.Tokens.SHORTEN ->
          "SHORTEN"
      | Elpi_lexer_config.Tokens.SIG ->
          "SIG"
      | Elpi_lexer_config.Tokens.SIGMA ->
          "SIGMA"
      | Elpi_lexer_config.Tokens.SLASH ->
          "SLASH"
      | Elpi_lexer_config.Tokens.STRING _ ->
          "STRING"
      | Elpi_lexer_config.Tokens.TYPE ->
          "TYPE"
      | Elpi_lexer_config.Tokens.TYPEABBREV ->
          "TYPEABBREV"
      | Elpi_lexer_config.Tokens.UNTYPED ->
          "UNTYPED"
      | Elpi_lexer_config.Tokens.USEONLY ->
          "USEONLY"
      | Elpi_lexer_config.Tokens.USE_SIG ->
          "USE_SIG"
      | Elpi_lexer_config.Tokens.VDASH ->
          "VDASH"
  
  let _menhir_fail : unit -> 'a =
    fun () ->
      Printf.eprintf "Internal failure -- please contact the parser generator's developers.\n%!";
      assert false
  
  include struct
    
    [@@@ocaml.warning "-4-37"]
    
    let _menhir_goto_goal : type  ttv_stack. ttv_stack -> _ -> _menhir_box_goal =
      fun _menhir_stack _v ->
        MenhirBox_goal _v
    
    let _menhir_run_469 : type  ttv_stack. ttv_stack -> _ -> _menhir_box_program =
      fun _menhir_stack _v ->
        MenhirBox_program _v
    
    let rec _menhir_goto_program : type  ttv_stack. ttv_stack -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _v _menhir_s ->
        match _menhir_s with
        | MenhirState480 ->
            _menhir_run_481 _menhir_stack _v
        | MenhirState353 ->
            _menhir_run_469 _menhir_stack _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_481 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_decl -> _ -> _menhir_box_program =
      fun _menhir_stack _v ->
        let MenhirCell1_decl (_menhir_stack, _menhir_s, d) = _menhir_stack in
        let p = _v in
        let _v = _menhir_action_332 d p in
        _menhir_goto_program _menhir_stack _v _menhir_s
    
    let _menhir_run_422 : type  ttv_stack. ttv_stack -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_s ->
        let _v = _menhir_action_331 () in
        _menhir_goto_program _menhir_stack _v _menhir_s
    
    let rec _menhir_run_001 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_097 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_constant : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState601 ->
            _menhir_run_608 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState606 ->
            _menhir_run_608 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState486 ->
            _menhir_run_534 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState524 ->
            _menhir_run_525 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState515 ->
            _menhir_run_516 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState494 ->
            _menhir_run_495 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState423 ->
            _menhir_run_428 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState428 ->
            _menhir_run_428 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState425 ->
            _menhir_run_428 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState413 ->
            _menhir_run_416 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | MenhirState392 ->
            _menhir_run_393 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok
        | MenhirState389 ->
            _menhir_run_390 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState377 ->
            _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState383 ->
            _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState379 ->
            _menhir_run_381 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState371 ->
            _menhir_run_375 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState369 ->
            _menhir_run_370 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok
        | MenhirState361 ->
            _menhir_run_366 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState362 ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState487 ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState461 ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState419 ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState406 ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState400 ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState398 ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState359 ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState355 ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState521 ->
            _menhir_run_302 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState100 ->
            _menhir_run_302 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState107 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState286 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState283 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState112 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState279 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState220 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState236 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState270 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState274 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState276 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState272 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState268 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState266 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState264 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState262 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState252 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState260 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState254 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState258 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState256 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState250 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState248 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState246 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState240 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState244 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState242 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState238 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState222 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState230 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState234 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState232 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState226 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState228 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState224 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState118 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState597 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState595 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState593 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState591 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState589 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState587 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState585 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState583 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState581 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState579 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState577 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState575 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState573 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState571 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState569 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState567 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState565 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState563 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState560 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState558 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState556 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState554 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState552 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState550 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState548 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState546 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState544 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState542 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState540 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState538 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState536 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState529 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState509 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState497 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState394 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState396 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState000 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState095 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState099 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState299 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState293 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState106 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState215 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState213 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState124 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState206 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState140 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState204 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState142 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState160 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState194 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState198 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState200 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState202 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState196 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState192 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState190 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState188 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState186 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState176 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState184 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState178 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState182 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState180 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState174 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState172 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState170 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState164 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState168 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState166 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState162 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState144 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState154 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState158 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState156 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState150 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState152 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState146 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState125 ->
            _menhir_run_137 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState531 ->
            _menhir_run_134 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState493 ->
            _menhir_run_134 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState512 ->
            _menhir_run_134 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState504 ->
            _menhir_run_134 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState505 ->
            _menhir_run_134 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState500 ->
            _menhir_run_134 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState122 ->
            _menhir_run_134 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState135 ->
            _menhir_run_134 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState129 ->
            _menhir_run_134 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState518 ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState490 ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState363 ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState003 ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState097 ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState113 ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState108 ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState105 ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState101 ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState102 ->
            _menhir_run_104 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState082 ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState066 ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState525 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState527 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState488 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState420 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState401 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState356 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState058 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState303 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState296 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState211 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState138 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState059 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState061 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState067 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState068 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState071 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState074 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState064 ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_608 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let c = _v in
        let _v = _menhir_action_122 c in
        _menhir_goto_filename _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_filename : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_filename (_menhir_stack, _menhir_s, _v) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_CONJ (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState606 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_602 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 606)
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let x = _v in
            let _v = _menhir_action_337 x in
            _menhir_goto_separated_nonempty_list_CONJ_filename_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR 605
    
    and _menhir_run_602 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let s = _v in
        let _v = _menhir_action_123 s in
        _menhir_goto_filename _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_004 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_093 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_005 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_094 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_006 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_103 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_007 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_090 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_008 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState008 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SLASH ->
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QDASH ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.OR ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MOD ->
            _menhir_run_015 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSs ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSr ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSi ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUS ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IS ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v ->
            _menhir_run_021 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_022 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v ->
            _menhir_run_024 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_OR _v ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_LT _v ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_GT _v ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_AND _v ->
            _menhir_run_034 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.EQ2 ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.EQ ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DIV ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DDARROW ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DARROW ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONS ->
            _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ2 ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AS ->
            _menhir_run_043 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            _menhir_run_045 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 8
    
    and _menhir_run_009 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_225 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_mixfix_SYMB : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState413 ->
            _menhir_run_414 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | MenhirState521 ->
            _menhir_run_046 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState494 ->
            _menhir_run_046 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState362 ->
            _menhir_run_046 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState100 ->
            _menhir_run_046 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState059 ->
            _menhir_run_046 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState067 ->
            _menhir_run_046 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState008 ->
            _menhir_run_046 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_414 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_FIXITY -> _ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.INTEGER _v_0 ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_FIXITY (_menhir_stack, _menhir_s, f, _startpos_f_) = _menhir_stack in
            let (_endpos_c_, _startpos_c_, c, i) = (_endpos, _startpos, _v, _v_0) in
            let _v = _menhir_action_125 _endpos_c_ _startpos_c_ c f i in
            _menhir_goto_fixity _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_f_ _v _menhir_s _tok
        | _ ->
            _eRR 414
    
    and _menhir_goto_fixity : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_endpos__2_, _startpos_f_, f) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_121 _endpos__2_ _startpos_f_ f in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 478
    
    and _menhir_goto_decl : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_decl (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.USE_SIG ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.USEONLY ->
            _menhir_run_355 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.TYPEABBREV ->
            _menhir_run_361 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.SIG ->
            _menhir_run_369 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.SHORTEN ->
            _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.RCURLY ->
            _menhir_run_388 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.NAMESPACE ->
            _menhir_run_389 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.MODULE ->
            _menhir_run_392 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.MACRO ->
            _menhir_run_394 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.LOCALKIND ->
            _menhir_run_398 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.LOCAL ->
            _menhir_run_400 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_405 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.KIND ->
            _menhir_run_406 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.IMPORT ->
            _menhir_run_412 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.FIXITY _v_0 ->
            _menhir_run_413 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState480
        | Elpi_lexer_config.Tokens.EXTERNAL ->
            _menhir_run_418 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.EXPORTDEF ->
            _menhir_run_419 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.EOF ->
            _menhir_run_422 _menhir_stack MenhirState480
        | Elpi_lexer_config.Tokens.CONSTRAINT ->
            _menhir_run_423 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.COLON ->
            _menhir_run_431 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.CLOSED ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.ACCUM_SIG ->
            _menhir_run_463 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.ACCUMULATE ->
            _menhir_run_464 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState480
        | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.FAMILY_TILDE _ | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FUNC | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.PI | Elpi_lexer_config.Tokens.PRED | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.RULE | Elpi_lexer_config.Tokens.SIGMA | Elpi_lexer_config.Tokens.TYPE | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_s = MenhirState480 in
            let _v = _menhir_action_024 () in
            _menhir_goto_attributes _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _endpos _v _menhir_s _tok
        | _ ->
            _eRR 480
    
    and _menhir_run_354 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _startpos__1_ = _startpos in
        let _v = _menhir_action_010 () in
        _menhir_goto_accumulate _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_accumulate : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_accumulate (_menhir_stack, _menhir_s, _v, _startpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.STRING _v_0 ->
            _menhir_run_602 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState601
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.IO _v_1 ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState601
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.CONSTANT _v_2 ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState601
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState601
        | _ ->
            _eRR 601
    
    and _menhir_run_048 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_c_, _startpos_c_, c) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_098 c in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_c_ _startpos_c_ _v _menhir_s _tok
    
    and _menhir_run_049 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_095 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_050 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_089 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_051 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_096 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_052 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_100 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_053 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_099 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_054 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_c_, _startpos_c_, c) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_088 c in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_c_ _startpos_c_ _v _menhir_s _tok
    
    and _menhir_run_055 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_091 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_056 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_092 () in
        _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_355 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_USEONLY (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState355 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 355
    
    and _menhir_run_361 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_TYPEABBREV (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState361 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_s = MenhirState362 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.VDASH ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.SLASH ->
                _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QDASH ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.OR ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.MOD ->
                _menhir_run_015 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.MINUSs ->
                _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.MINUSr ->
                _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.MINUSi ->
                _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.MINUS ->
                _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IS ->
                _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TIMES _v ->
                _menhir_run_021 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_022 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TICK _v ->
                _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_SHARP _v ->
                _menhir_run_024 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_QMARK _v ->
                _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_PLUS _v ->
                _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_OR _v ->
                _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_MINUS _v ->
                _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_LT _v ->
                _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_GT _v ->
                _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_EXP _v ->
                _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_EQ _v ->
                _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_BTICK _v ->
                _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_AND _v ->
                _menhir_run_034 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.EQ2 ->
                _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.EQ ->
                _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.DIV ->
                _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.DDARROW ->
                _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.DARROW ->
                _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CONS ->
                _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONJ2 ->
                _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONJ ->
                _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AS ->
                _menhir_run_043 _menhir_stack _menhir_lexbuf _menhir_lexer
            | Elpi_lexer_config.Tokens.ARROW ->
                _menhir_run_045 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 362)
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 361
    
    and _menhir_run_010 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_222 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_011 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_229 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_012 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_221 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_013 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_228 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_014 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_214 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_015 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_216 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_016 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_212 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_017 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_210 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_018 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_211 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_019 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_209 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_020 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_215 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_021 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_196 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_022 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_226 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_023 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_206 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_024 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_204 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_025 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_227 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_026 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_195 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_027 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_203 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_028 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_197 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_029 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_199 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_030 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_200 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_031 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_198 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_032 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_201 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_033 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_205 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_034 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_202 s in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_run_035 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_213 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_036 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_208 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_037 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_217 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_038 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_220 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_039 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_219 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_040 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_207 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_041 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_223 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_042 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_224 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_043 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LPAREN -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let _endpos__3_ = _endpos_0 in
            let _v = _menhir_action_102 () in
            _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR 43
    
    and _menhir_run_045 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_218 () in
        _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_369 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_SIG (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState369 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 369
    
    and _menhir_run_371 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_SHORTEN (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState371 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 371
    
    and _menhir_run_388 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_117 _endpos__1_ _startpos__1_ in
        _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
    
    and _menhir_run_389 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_NAMESPACE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState389 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 389
    
    and _menhir_run_392 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_MODULE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState392 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 392
    
    and _menhir_run_394 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_MACRO (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState394 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 394
    
    and _menhir_run_002 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_x_, _startpos_x_, x) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_081 _endpos_x_ _startpos_x_ x in
        _menhir_goto_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _startpos_x_ _v _menhir_s _tok
    
    and _menhir_goto_closed_term : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState493 ->
            _menhir_run_502 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState512 ->
            _menhir_run_502 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState504 ->
            _menhir_run_502 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState505 ->
            _menhir_run_502 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState500 ->
            _menhir_run_502 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState494 ->
            _menhir_run_496 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState107 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState286 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState283 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState112 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState279 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState220 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState236 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState270 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState274 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState276 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState272 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState268 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState266 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState264 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState262 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState252 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState260 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState254 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState258 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState256 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState250 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState248 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState246 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState240 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState244 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState242 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState238 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState222 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState230 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState234 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState232 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState226 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState228 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState224 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState118 ->
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState597 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState595 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState593 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState591 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState589 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState587 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState585 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState583 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState581 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState579 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState577 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState575 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState573 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState571 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState569 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState567 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState565 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState563 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState560 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState558 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState556 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState554 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState552 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState550 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState548 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState546 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState544 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState542 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState540 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState538 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState536 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState529 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState521 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState509 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState497 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState394 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState396 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState000 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState095 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState099 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState100 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState299 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState293 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState106 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState215 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState213 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState124 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState125 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState206 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState140 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState204 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState142 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState160 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState194 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState198 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState200 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState202 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState196 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState192 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState190 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState188 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState186 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState176 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState184 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState178 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState182 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState180 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState174 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState172 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState170 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState164 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState168 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState166 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState162 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState144 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState154 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState158 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState156 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState150 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState152 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState146 ->
            _menhir_run_148 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState531 ->
            _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState122 ->
            _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState135 ->
            _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState129 ->
            _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_502 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_345 _endpos_t_ _startpos_t_ t in
        _menhir_goto_sequent _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _v _menhir_s _tok
    
    and _menhir_goto_sequent : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState512 ->
            _menhir_run_513 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState505 ->
            _menhir_run_505 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState504 ->
            _menhir_run_505 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState500 ->
            _menhir_run_500 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState493 ->
            _menhir_run_500 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_513 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_RULE, _menhir_box_program) _menhir_cell1_list_sequent_ _menhir_cell0_option_preceded_BIND_nonempty_list_sequent___ _menhir_cell0_option_preceded_PIPE_term__ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let (_endpos_x_, x) = (_endpos, _v) in
        let _v = _menhir_action_320 x in
        _menhir_goto_option_preceded_IFF_sequent__ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _tok
    
    and _menhir_goto_option_preceded_IFF_sequent__ : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_RULE, _menhir_box_program) _menhir_cell1_list_sequent_ _menhir_cell0_option_preceded_BIND_nonempty_list_sequent___ _menhir_cell0_option_preceded_PIPE_term__ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_option_preceded_PIPE_term__ (_menhir_stack, guard) = _menhir_stack in
        let MenhirCell0_option_preceded_BIND_nonempty_list_sequent___ (_menhir_stack, to_remove) = _menhir_stack in
        let MenhirCell1_list_sequent_ (_menhir_stack, _, to_match) = _menhir_stack in
        let MenhirCell1_RULE (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_attributes (_menhir_stack, _menhir_s, attributes, _startpos_attributes_, _) = _menhir_stack in
        let (_endpos_new_goal_, new_goal) = (_endpos, _v) in
        let _v = _menhir_action_037 _endpos_new_goal_ _startpos_attributes_ attributes guard new_goal to_match to_remove in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let r = _v in
            let _v = _menhir_action_106 r in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 484
    
    and _menhir_run_505 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.STRING _v_0 ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState505
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.QUOTED _v_1 ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState505
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.LCURLY ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.IO _v_2 ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState505
        | Elpi_lexer_config.Tokens.INTEGER _v_3 ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState505
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.FLOAT _v_4 ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState505
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.CONSTANT _v_5 ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState505
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState505
        | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE ->
            let x = _v in
            let _v = _menhir_action_238 x in
            _menhir_goto_nonempty_list_sequent_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 505
    
    and _menhir_run_096 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_x_, _startpos_x_, x) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_082 _endpos_x_ _startpos_x_ x in
        _menhir_goto_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _startpos_x_ _v _menhir_s _tok
    
    and _menhir_run_494 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState494 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SLASH ->
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.QDASH ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.OR ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MOD ->
            _menhir_run_015 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSs ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSr ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSi ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUS ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IS ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v ->
            _menhir_run_021 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_293 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v ->
            _menhir_run_024 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_OR _v ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_LT _v ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_GT _v ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_AND _v ->
            _menhir_run_034 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.EQ2 ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.EQ ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DIV ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DDARROW ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DARROW ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CONS ->
            _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ2 ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AS ->
            _menhir_run_043 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            _menhir_run_045 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 494
    
    and _menhir_run_101 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.IO _v ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState101
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState101
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState101
        | Elpi_lexer_config.Tokens.RPAREN ->
            let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
            let _v = _menhir_action_229 () in
            _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR 101
    
    and _menhir_run_105 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.IO _v ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState105
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState105
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Elpi_lexer_config.Tokens.RPAREN ->
            let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
            let _v = _menhir_action_228 () in
            _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR 105
    
    and _menhir_run_100 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState100 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SLASH ->
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.QDASH ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.OR ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MOD ->
            _menhir_run_015 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSs ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSr ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSi ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUS ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IS ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v ->
            _menhir_run_021 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_293 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v ->
            _menhir_run_024 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_OR _v ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_LT _v ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_GT _v ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_AND _v ->
            _menhir_run_034 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.EQ2 ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.EQ ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DIV ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DDARROW ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DARROW ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CONS ->
            _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ2 ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AS ->
            _menhir_run_043 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            _menhir_run_045 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 100
    
    and _menhir_run_106 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_LCURLY (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState106 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 106
    
    and _menhir_run_003 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState003 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 3
    
    and _menhir_run_097 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState097 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 97
    
    and _menhir_run_107 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACKET (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState107 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.RBRACKET ->
            _menhir_run_111 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PIPE ->
            _menhir_run_112 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 107
    
    and _menhir_run_108 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState108 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 108
    
    and _menhir_run_111 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_184 _endpos__1_ _startpos__1_ in
        _menhir_goto_list_items _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_list_items : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState107 ->
            _menhir_run_290 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState286 ->
            _menhir_run_288 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_290 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACKET -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_LBRACKET (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_endpos_l_, l) = (_endpos, _v) in
        let _v = _menhir_action_084 _endpos_l_ _startpos__1_ l in
        _menhir_goto_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_l_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_288 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_CONJ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_CONJ (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let (_endpos_xs_, xs) = (_endpos, _v) in
        let _v = _menhir_action_186 x xs in
        _menhir_goto_list_items _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_xs_ _v _menhir_s _tok
    
    and _menhir_run_112 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_PIPE (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState112 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 112
    
    and _menhir_run_113 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState113 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 113
    
    and _menhir_run_116 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_x_, _startpos_x_, x) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_079 _endpos_x_ _startpos_x_ x in
        _menhir_goto_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _startpos_x_ _v _menhir_s _tok
    
    and _menhir_run_117 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_x_, _startpos_x_, x) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_080 _endpos_x_ _startpos_x_ x in
        _menhir_goto_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _startpos_x_ _v _menhir_s _tok
    
    and _menhir_run_118 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState118 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 118
    
    and _menhir_run_125 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState125 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 125
    
    and _menhir_run_293 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.STRING _v_0 ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState293
        | Elpi_lexer_config.Tokens.SIGMA ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.QUOTED _v_1 ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState293
        | Elpi_lexer_config.Tokens.PI ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.LCURLY ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.IO _v_2 ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState293
        | Elpi_lexer_config.Tokens.INTEGER _v_3 ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState293
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.FLOAT _v_4 ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState293
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v_5 ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState293
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.CONSTANT _v_6 ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6 MenhirState293
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Elpi_lexer_config.Tokens.RPAREN ->
            let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_226 s in
            _menhir_goto_mixfix_SYMB _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
        | _ ->
            _eRR 293
    
    and _menhir_goto_nonempty_list_sequent_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState504 ->
            _menhir_run_507 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState505 ->
            _menhir_run_506 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_507 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_RULE, _menhir_box_program) _menhir_cell1_list_sequent_ _menhir_cell0_BIND -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_BIND (_menhir_stack, _, _) = _menhir_stack in
        let x = _v in
        let _v = _menhir_action_318 x in
        _menhir_goto_option_preceded_BIND_nonempty_list_sequent___ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
    
    and _menhir_goto_option_preceded_BIND_nonempty_list_sequent___ : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_RULE, _menhir_box_program) _menhir_cell1_list_sequent_ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let _menhir_stack = MenhirCell0_option_preceded_BIND_nonempty_list_sequent___ (_menhir_stack, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.PIPE ->
            let _menhir_s = MenhirState509 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 509)
        | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF ->
            let _v = _menhir_action_321 () in
            _menhir_goto_option_preceded_PIPE_term__ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_goto_option_preceded_PIPE_term__ : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_RULE, _menhir_box_program) _menhir_cell1_list_sequent_ _menhir_cell0_option_preceded_BIND_nonempty_list_sequent___ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let _menhir_stack = MenhirCell0_option_preceded_PIPE_term__ (_menhir_stack, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.IFF ->
            let _menhir_s = MenhirState512 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 512)
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _v = _menhir_action_319 () in
            _menhir_goto_option_preceded_IFF_sequent__ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_506 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_sequent -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_sequent (_menhir_stack, _menhir_s, x, _) = _menhir_stack in
        let xs = _v in
        let _v = _menhir_action_239 x xs in
        _menhir_goto_nonempty_list_sequent_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
    
    and _menhir_run_500 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_sequent (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.STRING _v_0 ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState500
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.QUOTED _v_1 ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState500
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.IO _v_2 ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState500
        | Elpi_lexer_config.Tokens.INTEGER _v_3 ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState500
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.FLOAT _v_4 ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState500
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.CONSTANT _v_5 ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState500
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState500
        | Elpi_lexer_config.Tokens.BIND | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE ->
            let _v_6 = _menhir_action_182 () in
            _menhir_run_501 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v_6 _tok
        | _ ->
            _eRR 500
    
    and _menhir_run_501 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_sequent -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_sequent (_menhir_stack, _menhir_s, x, _) = _menhir_stack in
        let xs = _v in
        let _v = _menhir_action_183 x xs in
        _menhir_goto_list_sequent_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
    
    and _menhir_goto_list_sequent_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState493 ->
            _menhir_run_503 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState500 ->
            _menhir_run_501 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_503 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_RULE as 'stack) -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_list_sequent_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.BIND ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_BIND (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState504 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 504)
        | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE ->
            let _v = _menhir_action_317 () in
            _menhir_goto_option_preceded_BIND_nonempty_list_sequent___ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_496 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.RTRI ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState497 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 497)
        | Elpi_lexer_config.Tokens.ACCUMULATE | Elpi_lexer_config.Tokens.ACCUM_SIG | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.BIND | Elpi_lexer_config.Tokens.CLOSED | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CONSTRAINT | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.EXPORTDEF | Elpi_lexer_config.Tokens.EXTERNAL | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TILDE _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FIXITY _ | Elpi_lexer_config.Tokens.FLOAT _ | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.FUNC | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IMPORT | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.INTEGER _ | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.IO_COLON _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.KIND | Elpi_lexer_config.Tokens.LBRACKET | Elpi_lexer_config.Tokens.LCURLY | Elpi_lexer_config.Tokens.LOCAL | Elpi_lexer_config.Tokens.LOCALKIND | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.MACRO | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.MODULE | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NAMESPACE | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PI | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.PRED | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.QUOTED _ | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.RULE | Elpi_lexer_config.Tokens.SHORTEN | Elpi_lexer_config.Tokens.SIG | Elpi_lexer_config.Tokens.SIGMA | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.STRING _ | Elpi_lexer_config.Tokens.TYPE | Elpi_lexer_config.Tokens.TYPEABBREV | Elpi_lexer_config.Tokens.UNTYPED | Elpi_lexer_config.Tokens.USEONLY | Elpi_lexer_config.Tokens.USE_SIG | Elpi_lexer_config.Tokens.VDASH ->
            let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_352 t in
            _menhir_goto_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
    
    and _menhir_goto_term : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState597 ->
            _menhir_run_598 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState595 ->
            _menhir_run_596 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState593 ->
            _menhir_run_594 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState591 ->
            _menhir_run_592 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState589 ->
            _menhir_run_590 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState587 ->
            _menhir_run_588 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState585 ->
            _menhir_run_586 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState583 ->
            _menhir_run_584 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState581 ->
            _menhir_run_582 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState579 ->
            _menhir_run_580 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState577 ->
            _menhir_run_578 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState575 ->
            _menhir_run_576 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState573 ->
            _menhir_run_574 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState571 ->
            _menhir_run_572 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState569 ->
            _menhir_run_570 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState567 ->
            _menhir_run_568 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState565 ->
            _menhir_run_566 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState563 ->
            _menhir_run_564 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState560 ->
            _menhir_run_561 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState558 ->
            _menhir_run_559 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState556 ->
            _menhir_run_557 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState554 ->
            _menhir_run_555 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState552 ->
            _menhir_run_553 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState550 ->
            _menhir_run_551 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState548 ->
            _menhir_run_549 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState546 ->
            _menhir_run_547 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState544 ->
            _menhir_run_545 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState542 ->
            _menhir_run_543 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState540 ->
            _menhir_run_541 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState538 ->
            _menhir_run_539 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState536 ->
            _menhir_run_537 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState529 ->
            _menhir_run_530 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState521 ->
            _menhir_run_522 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState509 ->
            _menhir_run_510 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState497 ->
            _menhir_run_498 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState396 ->
            _menhir_run_397 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState394 ->
            _menhir_run_395 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState000 ->
            _menhir_run_310 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState095 ->
            _menhir_run_308 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState099 ->
            _menhir_run_306 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState299 ->
            _menhir_run_300 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState494 ->
            _menhir_run_294 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState100 ->
            _menhir_run_294 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState106 ->
            _menhir_run_291 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState215 ->
            _menhir_run_216 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState213 ->
            _menhir_run_214 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState124 ->
            _menhir_run_208 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState206 ->
            _menhir_run_207 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState204 ->
            _menhir_run_205 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState202 ->
            _menhir_run_203 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState200 ->
            _menhir_run_201 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState198 ->
            _menhir_run_199 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState196 ->
            _menhir_run_197 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState194 ->
            _menhir_run_195 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState192 ->
            _menhir_run_193 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState190 ->
            _menhir_run_191 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState188 ->
            _menhir_run_189 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState186 ->
            _menhir_run_187 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState184 ->
            _menhir_run_185 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState182 ->
            _menhir_run_183 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState180 ->
            _menhir_run_181 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState178 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState176 ->
            _menhir_run_177 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState174 ->
            _menhir_run_175 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState172 ->
            _menhir_run_173 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState170 ->
            _menhir_run_171 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState168 ->
            _menhir_run_169 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState166 ->
            _menhir_run_167 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState164 ->
            _menhir_run_165 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState162 ->
            _menhir_run_163 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState160 ->
            _menhir_run_161 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState158 ->
            _menhir_run_159 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState156 ->
            _menhir_run_157 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState154 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState152 ->
            _menhir_run_153 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState150 ->
            _menhir_run_151 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState146 ->
            _menhir_run_147 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState144 ->
            _menhir_run_145 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState142 ->
            _menhir_run_143 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState140 ->
            _menhir_run_141 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState293 ->
            _menhir_run_126 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState125 ->
            _menhir_run_126 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_598 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_ARROW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_ARROW (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_068 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 598
    
    and _menhir_run_144 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_SLASH (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState144 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 144
    
    and _menhir_run_162 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_MOD (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState162 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 162
    
    and _menhir_run_164 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_MINUSs (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState164 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 164
    
    and _menhir_run_170 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_MINUSr (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState170 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 170
    
    and _menhir_run_172 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_MINUSi (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState172 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 172
    
    and _menhir_run_174 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_MINUS (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState174 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 174
    
    and _menhir_run_176 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_IS (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState176 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 176
    
    and _menhir_run_166 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_TIMES (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState166 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 166
    
    and _menhir_run_178 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_TICK (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState178 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 178
    
    and _menhir_run_146 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_SHARP (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState146 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 146
    
    and _menhir_run_127 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_275 _endpos_s_ _startpos_l_ _startpos_s_ l s in
        _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_l_ _v _menhir_s _tok
    
    and _menhir_goto_open_term : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_350 t in
        _menhir_goto_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
    
    and _menhir_run_180 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_PLUS (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState180 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 180
    
    and _menhir_run_150 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_OR (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState150 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 150
    
    and _menhir_run_154 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_MINUS (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState154 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 154
    
    and _menhir_run_186 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_LT (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState186 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 186
    
    and _menhir_run_188 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_GT (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState188 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 188
    
    and _menhir_run_182 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_EXP (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState182 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 182
    
    and _menhir_run_156 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_EQ (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState156 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 156
    
    and _menhir_run_158 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_BTICK (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState158 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 158
    
    and _menhir_run_152 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_AND (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState152 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 152
    
    and _menhir_run_190 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_EQ2 (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState190 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 190
    
    and _menhir_run_192 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_EQ (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState192 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 192
    
    and _menhir_run_168 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_DIV (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState168 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 168
    
    and _menhir_run_196 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_DARROW (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState196 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 196
    
    and _menhir_run_184 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_CONS (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState184 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 184
    
    and _menhir_run_202 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_ARROW (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState202 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 202
    
    and _menhir_goto_clause_hd_open_term : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_attributes as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_077 t in
        _menhir_goto_clause_hd_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
    
    and _menhir_goto_clause_hd_term : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_attributes as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_VDASH (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState536 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 536)
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_SLASH (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState538 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 538)
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_QDASH (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState540 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 540)
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_OR (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState542 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 542)
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_MOD (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState544 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 544)
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_MINUSs (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState546 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 546)
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_MINUSr (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState548 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 548)
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_MINUSi (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState550 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 550)
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_MINUS (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState552 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 552)
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IS (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState554 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 554)
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_90 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_90 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_TIMES (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState556 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 556)
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_100 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_100 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_TICK (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState558 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 558)
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_110 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_110 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_SHARP (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState560 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 560)
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_120 ->
            let _startpos_121 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_122 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_endpos_s_, _startpos_s_, s, _startpos_l_, l) = (_endpos_122, _startpos_121, _v_120, _startpos, _v) in
            let _v = _menhir_action_076 _endpos_s_ _startpos_l_ _startpos_s_ l s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_l_ _v _menhir_s _tok
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_123 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_123 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_PLUS (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState563 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 563)
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_133 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_133 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_OR (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState565 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 565)
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_143 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_143 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_MINUS (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState567 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 567)
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_153 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_153 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_LT (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState569 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 569)
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_163 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_163 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_GT (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState571 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 571)
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_173 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_173 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_EXP (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState573 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 573)
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_183 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_183 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_EQ (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState575 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 575)
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_193 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_193 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_BTICK (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState577 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 577)
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_203 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _v = _v_203 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FAMILY_AND (_menhir_stack, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState579 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 579)
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_EQ2 (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState581 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 581)
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_EQ (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState583 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 583)
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_DIV (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState585 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 585)
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_DDARROW (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState587 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 587)
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_DARROW (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState589 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 589)
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_CONS (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState591 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 591)
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_CONJ2 (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState593 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 593)
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_CONJ (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState595 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 595)
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_ARROW (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState597 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 597)
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let MenhirCell1_attributes (_menhir_stack, _menhir_s, attributes, _startpos_attributes_, _endpos_attributes_) = _menhir_stack in
            let (_endpos_body_, _startpos_body_, body) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_038 _endpos_attributes_ _endpos_body_ _startpos_attributes_ _startpos_body_ attributes body in
            _menhir_goto_clause _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR 535
    
    and _menhir_goto_clause : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let c = _v in
        let _v = _menhir_action_105 c in
        _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
    
    and _menhir_run_596 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_CONJ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_CONJ (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_074 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 596
    
    and _menhir_run_198 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_CONJ2 (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState198 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 198
    
    and _menhir_run_200 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_CONJ (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState200 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 200
    
    and _menhir_run_594 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_CONJ2 as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_CONJ2 (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_073 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 594
    
    and _menhir_run_592 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_CONS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_CONS (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_057 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 592
    
    and _menhir_run_590 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_DARROW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_DARROW (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_069 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 590
    
    and _menhir_run_588 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_DDARROW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_DDARROW (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_070 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 588
    
    and _menhir_run_194 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_DDARROW (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState194 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 194
    
    and _menhir_run_586 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_DIV as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_DIV (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_067 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 586
    
    and _menhir_run_584 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_EQ (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_058 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 584
    
    and _menhir_run_582 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_EQ2 as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_EQ2 (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_063 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 582
    
    and _menhir_run_580 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_AND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_AND (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_052 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 580
    
    and _menhir_run_578 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_BTICK as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_BTICK (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_055 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 578
    
    and _menhir_run_576 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_EQ (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_051 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 576
    
    and _menhir_run_574 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_EXP as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_EXP (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_048 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 574
    
    and _menhir_run_572 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_GT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_GT (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_050 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 572
    
    and _menhir_run_570 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_LT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_LT (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_049 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 570
    
    and _menhir_run_568 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_MINUS (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_047 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 568
    
    and _menhir_run_566 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_OR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_OR (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_053 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 566
    
    and _menhir_run_564 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_PLUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_PLUS (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_045 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 564
    
    and _menhir_run_561 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_SHARP as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_SHARP (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_054 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 561
    
    and _menhir_run_559 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_TICK as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_TICK (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_056 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 559
    
    and _menhir_run_557 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_FAMILY_TIMES as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_TIMES (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_046 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 557
    
    and _menhir_run_555 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_IS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_IS (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_065 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 555
    
    and _menhir_run_553 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUS (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_059 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 553
    
    and _menhir_run_551 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_MINUSi as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUSi (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_061 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 551
    
    and _menhir_run_549 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_MINUSr as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUSr (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_060 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 549
    
    and _menhir_run_547 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_MINUSs as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUSs (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_062 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 547
    
    and _menhir_run_545 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_MOD as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MOD (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_066 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 545
    
    and _menhir_run_543 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_OR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_OR (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_064 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 543
    
    and _menhir_run_160 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_OR (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState160 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 160
    
    and _menhir_run_541 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_QDASH as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let MenhirCell0_QDASH (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_071 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 541
    
    and _menhir_run_539 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_SLASH as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_SLASH (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_072 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 539
    
    and _menhir_run_537 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_clause_hd_term _menhir_cell0_VDASH as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let MenhirCell0_VDASH (_menhir_stack, _startpos_v_, _endpos_v_) = _menhir_stack in
            let MenhirCell1_clause_hd_term (_menhir_stack, _, l, _startpos_l_, _) = _menhir_stack in
            let MenhirCell1_attributes (_menhir_stack, _menhir_s, attributes, _startpos_attributes_, _endpos_attributes_) = _menhir_stack in
            let (_endpos_r_, r, v) = (_endpos, _v, ()) in
            let _v = _menhir_action_039 _endpos_attributes_ _endpos_r_ _endpos_v_ _startpos_attributes_ _startpos_l_ _startpos_v_ attributes l r v in
            _menhir_goto_clause _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR 537
    
    and _menhir_run_142 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_VDASH (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState142 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 142
    
    and _menhir_run_204 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_QDASH (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState204 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 204
    
    and _menhir_run_530 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_FAMILY_TILDE as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_075 _endpos_r_ _endpos_s_ _startpos_s_ r s in
            _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_s_ _v _menhir_s _tok
        | _ ->
            _eRR 530
    
    and _menhir_run_522 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
                let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_endpos__3_, t) = (_endpos_0, _v) in
                let _v = _menhir_action_041 _endpos__3_ _startpos__1_ t in
                _menhir_goto_clause_hd_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
            | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.FLOAT _ | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.INTEGER _ | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.LBRACKET | Elpi_lexer_config.Tokens.LCURLY | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.QUOTED _ | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.STRING _ | Elpi_lexer_config.Tokens.UNTYPED ->
                let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_endpos__3_, t) = (_endpos_0, _v) in
                let _v = _menhir_action_133 _endpos__3_ _startpos__1_ t in
                _menhir_goto_head_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
            | _ ->
                _eRR 523)
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_13 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.COLON ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_296 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 522
    
    and _menhir_goto_clause_hd_closed_term : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_attributes as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_078 t in
        _menhir_goto_clause_hd_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
    
    and _menhir_goto_head_term : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState486 ->
            _menhir_run_531 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState531 ->
            _menhir_run_133 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState493 ->
            _menhir_run_133 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState512 ->
            _menhir_run_133 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState504 ->
            _menhir_run_133 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState505 ->
            _menhir_run_133 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState500 ->
            _menhir_run_133 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState122 ->
            _menhir_run_133 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState135 ->
            _menhir_run_133 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState129 ->
            _menhir_run_133 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState597 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState595 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState593 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState591 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState589 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState587 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState585 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState583 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState581 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState579 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState577 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState575 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState573 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState571 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState569 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState567 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState565 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState563 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState560 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState558 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState556 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState554 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState552 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState550 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState548 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState546 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState544 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState542 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState540 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState538 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState536 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState529 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState521 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState509 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState497 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState494 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState394 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState396 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState000 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState095 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState099 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState100 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState299 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState293 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState106 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState215 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState213 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState124 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState206 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState140 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState204 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState142 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState160 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState194 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState198 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState200 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState202 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState196 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState192 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState190 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState188 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState186 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState176 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState184 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState178 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState182 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState180 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState174 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState172 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState170 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState164 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState168 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState166 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState162 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState144 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState154 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState158 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState156 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState150 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState152 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState146 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState125 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState107 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState286 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState283 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState112 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState279 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState220 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState236 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState270 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState274 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState276 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState272 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState268 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState266 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState264 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState262 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState252 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState260 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState254 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState258 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState256 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState250 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState248 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState246 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState240 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState244 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState242 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState238 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState222 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState230 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState234 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState232 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState226 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState228 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState224 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState118 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_531 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_attributes as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.STRING _v_0 ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState531
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.QUOTED _v_1 ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState531
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.IO _v_2 ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState531
        | Elpi_lexer_config.Tokens.INTEGER _v_3 ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState531
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.FLOAT _v_4 ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState531
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.CONSTANT _v_5 ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState531
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState531
        | _ ->
            _eRR 531
    
    and _menhir_run_133 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_087 t in
        _menhir_goto_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
    
    and _menhir_run_129 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.STRING _v_0 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState129
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.QUOTED _v_1 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState129
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.LCURLY ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.IO _v_2 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState129
        | Elpi_lexer_config.Tokens.INTEGER _v_3 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState129
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.FLOAT _v_4 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState129
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.CONSTANT _v_5 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState129
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState129
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.RTRI | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_087 t in
            _menhir_goto_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
        | _ ->
            _eRR 129
    
    and _menhir_run_122 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.STRING _v_0 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState122
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.QUOTED _v_1 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState122
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.LCURLY ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.IO _v_2 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState122
        | Elpi_lexer_config.Tokens.INTEGER _v_3 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState122
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.FLOAT _v_4 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState122
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.CONSTANT _v_5 ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState122
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_head_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState122
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_087 t in
            _menhir_goto_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
        | _ ->
            _eRR 122
    
    and _menhir_run_296 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN, ttv_result) _menhir_cell1_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell0_COLON (_menhir_stack, _startpos) in
        let _menhir_s = MenhirState296 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 296
    
    and _menhir_run_059 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState059 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SLASH ->
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QDASH ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.PRED ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.OR ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MOD ->
            _menhir_run_015 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSs ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSr ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSi ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUS ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IS ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNC ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v ->
            _menhir_run_021 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_022 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v ->
            _menhir_run_024 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_OR _v ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_LT _v ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_GT _v ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_AND _v ->
            _menhir_run_034 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.EQ2 ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.EQ ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DIV ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DDARROW ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DARROW ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CONS ->
            _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ2 ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AS ->
            _menhir_run_043 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            _menhir_run_045 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 59
    
    and _menhir_run_060 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_PRED (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.IO_COLON _v ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState060
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _v = _menhir_action_192 () in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _eRR 60
    
    and _menhir_run_061 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _menhir_stack = MenhirCell1_IO_COLON (_menhir_stack, _menhir_s, _v) in
        let _menhir_s = MenhirState061 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 61
    
    and _menhir_run_089 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN, ttv_result) _menhir_cell1_PRED -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_PRED (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_endpos_xs_, xs) = (_endpos, _v) in
        let _v = _menhir_action_011 _endpos_xs_ _startpos__1_ xs in
        _menhir_goto_anonymous_pred _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_anonymous_pred : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState059 ->
            _menhir_run_092 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState067 ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_092 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LPAREN -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_endpos__3_, t) = (_endpos, _v) in
            let _v = _menhir_action_128 t in
            _menhir_goto_fotype_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR 92
    
    and _menhir_goto_fotype_term : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState525 ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState527 ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState068 ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState074 ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState071 ->
            _menhir_run_073 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState488 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState420 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState401 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState356 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState058 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState303 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState296 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState211 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState138 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState059 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState067 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState064 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState061 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_073 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_fotype_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_CONJ (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState074 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 74)
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.RPAREN ->
            let (_endpos_x_, x) = (_endpos, _v) in
            let _v = _menhir_action_339 x in
            _menhir_goto_separated_nonempty_list_CONJ_fotype_term_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _menhir_s _tok
        | _ ->
            _eRR 73
    
    and _menhir_goto_separated_nonempty_list_CONJ_fotype_term_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState074 ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState527 ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState525 ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState071 ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState068 ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_075 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_fotype_term _menhir_cell0_CONJ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_CONJ (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_fotype_term (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let (_endpos_xs_, xs) = (_endpos, _v) in
        let _v = _menhir_action_340 x xs in
        _menhir_goto_separated_nonempty_list_CONJ_fotype_term_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_xs_ _v _menhir_s _tok
    
    and _menhir_run_069 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let (_endpos_x_, x) = (_endpos, _v) in
        let _v = _menhir_action_191 x in
        _menhir_goto_loption_separated_nonempty_list_CONJ_fotype_term__ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _menhir_s _tok
    
    and _menhir_goto_loption_separated_nonempty_list_CONJ_fotype_term__ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState527 ->
            _menhir_run_528 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState525 ->
            _menhir_run_526 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState071 ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState068 ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_528 : type  ttv_stack. ((((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_FUNC, _menhir_box_program) _menhir_cell1_constant, _menhir_box_program) _menhir_cell1_loption_separated_nonempty_list_CONJ_fotype_term__ _menhir_cell0_ARROW -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_ARROW (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_loption_separated_nonempty_list_CONJ_fotype_term__ (_menhir_stack, _, xs, _) = _menhir_stack in
        let MenhirCell1_constant (_menhir_stack, _, name, _, _) = _menhir_stack in
        let MenhirCell1_FUNC (_menhir_stack, _, _startpos__2_) = _menhir_stack in
        let MenhirCell1_attributes (_menhir_stack, _menhir_s, attributes, _startpos_attributes_, _endpos_attributes_) = _menhir_stack in
        let (_endpos_xs_inlined1_, xs_inlined1) = (_endpos, _v) in
        let _v = _menhir_action_327 _endpos_attributes_ _endpos_xs_inlined1_ _startpos__2_ _startpos_attributes_ attributes name xs xs_inlined1 in
        _menhir_goto_pred _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_pred : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let p = _v in
            let _v = _menhir_action_107 p in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 470
    
    and _menhir_run_526 : type  ttv_stack. ((((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_FUNC, _menhir_box_program) _menhir_cell1_constant as 'stack) -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_loption_separated_nonempty_list_CONJ_fotype_term__ (_menhir_stack, _menhir_s, _v, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_ARROW (_menhir_stack, _startpos, _endpos_0) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.IO _v_1 ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState527
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.CONSTANT _v_2 ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState527
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState527
            | Elpi_lexer_config.Tokens.FULLSTOP ->
                let _v_3 = _menhir_action_190 () in
                _menhir_run_528 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_0 _v_3 _tok
            | _ ->
                _eRR 527)
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let MenhirCell1_constant (_menhir_stack, _, name, _, _) = _menhir_stack in
            let MenhirCell1_FUNC (_menhir_stack, _, _startpos__2_) = _menhir_stack in
            let MenhirCell1_attributes (_menhir_stack, _menhir_s, attributes, _startpos_attributes_, _endpos_attributes_) = _menhir_stack in
            let (_endpos_xs_, xs) = (_endpos, _v) in
            let _v = _menhir_action_328 _endpos_attributes_ _endpos_xs_ _startpos__2_ _startpos_attributes_ attributes name xs in
            _menhir_goto_pred _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR 526
    
    and _menhir_run_072 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_LPAREN, ttv_result) _menhir_cell1_FUNC, ttv_result) _menhir_cell1_loption_separated_nonempty_list_CONJ_fotype_term__ _menhir_cell0_ARROW -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_ARROW (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_loption_separated_nonempty_list_CONJ_fotype_term__ (_menhir_stack, _, xs, _) = _menhir_stack in
        let MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_endpos_xs_inlined1_, xs_inlined1) = (_endpos, _v) in
        let _v = _menhir_action_012 _endpos_xs_inlined1_ _startpos__1_ xs xs_inlined1 in
        _menhir_goto_anonymous_pred _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_070 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_LPAREN, ttv_result) _menhir_cell1_FUNC as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_loption_separated_nonempty_list_CONJ_fotype_term__ (_menhir_stack, _menhir_s, _v, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_ARROW (_menhir_stack, _startpos, _endpos_0) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.IO _v_1 ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState071
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.CONSTANT _v_2 ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState071
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
            | Elpi_lexer_config.Tokens.RPAREN ->
                let _v_3 = _menhir_action_190 () in
                _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_0 _v_3 _tok
            | _ ->
                _eRR 71)
        | Elpi_lexer_config.Tokens.RPAREN ->
            let MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_endpos_xs_, xs) = (_endpos, _v) in
            let _v = _menhir_action_013 _endpos_xs_ _startpos__1_ xs in
            _menhir_goto_anonymous_pred _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR 70
    
    and _menhir_run_063 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_fotype_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_ARROW (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState064 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 64)
        | Elpi_lexer_config.Tokens.BIND | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IO_COLON _ | Elpi_lexer_config.Tokens.RPAREN ->
            let (_endpos_fo_, fo) = (_endpos, _v) in
            let _v = _menhir_action_358 fo in
            _menhir_goto_type_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_fo_ _v _menhir_s _tok
        | _ ->
            _eRR 63
    
    and _menhir_goto_type_term : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState488 ->
            _menhir_run_489 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState420 ->
            _menhir_run_421 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok
        | MenhirState401 ->
            _menhir_run_402 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState367 ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState356 ->
            _menhir_run_357 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok
        | MenhirState303 ->
            _menhir_run_304 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState296 ->
            _menhir_run_297 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState211 ->
            _menhir_run_212 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState138 ->
            _menhir_run_139 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState058 ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState059 ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState067 ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState064 ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState061 ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_489 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_TYPE, _menhir_box_program) _menhir_cell1_separated_nonempty_list_CONJ_constant_ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _, names, _) = _menhir_stack in
        let MenhirCell1_TYPE (_menhir_stack, _, _startpos__2_, _) = _menhir_stack in
        let MenhirCell1_attributes (_menhir_stack, _menhir_s, attributes, _startpos_attributes_, _endpos_attributes_) = _menhir_stack in
        let (_endpos_t_, t) = (_endpos, _v) in
        let _v = _menhir_action_357 _endpos_attributes_ _endpos_t_ _startpos__2_ _startpos_attributes_ attributes names t in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let t = _v in
            let _v = _menhir_action_108 t in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 467
    
    and _menhir_run_421 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_EXPORTDEF, _menhir_box_program) _menhir_cell1_separated_nonempty_list_CONJ_constant_ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok ->
        let MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _, _, _) = _menhir_stack in
        let MenhirCell1_EXPORTDEF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let _endpos__3_ = _endpos in
        let _ = _menhir_action_139 _endpos__3_ _startpos__1_ in
        _menhir_goto_ignored _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _menhir_s _tok
    
    and _menhir_goto_ignored : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_startpos__1_, _endpos__2_) = (_startpos, _endpos) in
            let _v = _menhir_action_120 _endpos__2_ _startpos__1_ in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 476
    
    and _menhir_run_402 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_LOCAL, _menhir_box_program) _menhir_cell1_separated_nonempty_list_CONJ_constant_ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let x = _v in
        let _ = _menhir_action_324 x in
        _menhir_goto_option_type_term_ _menhir_stack _menhir_lexbuf _menhir_lexer _tok
    
    and _menhir_goto_option_type_term_ : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_LOCAL, _menhir_box_program) _menhir_cell1_separated_nonempty_list_CONJ_constant_ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _, l, _) = _menhir_stack in
            let MenhirCell1_LOCAL (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let _endpos__4_ = _endpos in
            let _v = _menhir_action_119 _endpos__4_ _startpos__1_ l in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 403
    
    and _menhir_run_368 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_TYPEABBREV, _menhir_box_program) _menhir_cell1_abbrevform -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_abbrevform (_menhir_stack, _, a) = _menhir_stack in
        let MenhirCell1_TYPEABBREV (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_endpos_t_, t) = (_endpos, _v) in
        let _v = _menhir_action_360 _endpos_t_ _startpos__1_ a t in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let a = _v in
            let _v = _menhir_action_115 a in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 465
    
    and _menhir_run_357 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_USEONLY, _menhir_box_program) _menhir_cell1_separated_nonempty_list_CONJ_constant_ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok ->
        let MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _, _, _) = _menhir_stack in
        let MenhirCell1_USEONLY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let _endpos__3_ = _endpos in
        let _ = _menhir_action_142 _endpos__3_ _startpos__1_ in
        _menhir_goto_ignored _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _menhir_s _tok
    
    and _menhir_run_304 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_LPAREN, ttv_result) _menhir_cell1_constant _menhir_cell0_COLON as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_constant (_menhir_stack, _, t, _startpos_t_, _endpos_t_) = _menhir_stack in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_endpos__5_, ty) = (_endpos_0, _v) in
            let _v = _menhir_action_134 _endpos__5_ _endpos_t_ _startpos__1_ _startpos_t_ t ty in
            _menhir_goto_head_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__5_ _startpos__1_ _v _menhir_s _tok
        | Elpi_lexer_config.Tokens.BIND ->
            let _menhir_stack = MenhirCell1_type_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_140 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 304
    
    and _menhir_run_140 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_constant _menhir_cell0_COLON, ttv_result) _menhir_cell1_type_term -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_BIND (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState140 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 140
    
    and _menhir_run_297 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN, ttv_result) _menhir_cell1_term _menhir_cell0_COLON -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _, t, _, _) = _menhir_stack in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_endpos__5_, ty) = (_endpos_0, _v) in
            let _v = _menhir_action_135 _endpos__5_ _startpos__1_ t ty in
            _menhir_goto_head_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__5_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR 297
    
    and _menhir_run_212 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_constant _menhir_cell0_COLON as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_type_term (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.BIND ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_BIND (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState213 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 213)
        | _ ->
            _eRR 212
    
    and _menhir_run_139 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_constant _menhir_cell0_COLON as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_type_term (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.BIND ->
            _menhir_run_140 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 139
    
    and _menhir_run_094 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_nonempty_list_constant_w_loc_, ttv_result) _menhir_cell1_COLON as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_type_term (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.BIND ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_BIND (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState095 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 95)
        | _ ->
            _eRR 94
    
    and _menhir_run_090 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LPAREN -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_endpos__3_, t) = (_endpos_0, _v) in
            let _v = _menhir_action_129 t in
            _menhir_goto_fotype_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR 90
    
    and _menhir_run_076 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LPAREN -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _) = _menhir_stack in
            let (_endpos__3_, t) = (_endpos_0, _v) in
            let _v = _menhir_action_028 t in
            _menhir_goto_atype_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _v _menhir_s _tok
        | _ ->
            _eRR 76
    
    and _menhir_goto_atype_term : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.IO _v_0 ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState082
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState082
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_atype_term (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.BIND | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IO_COLON _ | Elpi_lexer_config.Tokens.RPAREN ->
            let (_endpos_x_, x) = (_endpos, _v) in
            let _v = _menhir_action_230 x in
            _menhir_goto_nonempty_list_atype_term_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _menhir_s _tok
        | _ ->
            _eRR 82
    
    and _menhir_run_067 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState067 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SLASH ->
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QDASH ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.PRED ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.OR ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MOD ->
            _menhir_run_015 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSs ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSr ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSi ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUS ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IS ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNC ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v ->
            _menhir_run_021 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_022 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v ->
            _menhir_run_024 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_OR _v ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_LT _v ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_GT _v ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_AND _v ->
            _menhir_run_034 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.EQ2 ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.EQ ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DIV ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DDARROW ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DARROW ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CONS ->
            _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ2 ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AS ->
            _menhir_run_043 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            _menhir_run_045 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 67
    
    and _menhir_run_068 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState068
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState068
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState068
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.RPAREN ->
            let _v = _menhir_action_190 () in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v MenhirState068 _tok
        | _ ->
            _eRR 68
    
    and _menhir_goto_nonempty_list_atype_term_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState082 ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState066 ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_083 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_atype_term -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_atype_term (_menhir_stack, _menhir_s, x, _) = _menhir_stack in
        let (_endpos_xs_, xs) = (_endpos, _v) in
        let _v = _menhir_action_231 x xs in
        _menhir_goto_nonempty_list_atype_term_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_xs_ _v _menhir_s _tok
    
    and _menhir_run_080 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_constant -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_constant (_menhir_stack, _menhir_s, hd, _startpos_hd_, _) = _menhir_stack in
        let (_endpos_args_, args) = (_endpos, _v) in
        let _v = _menhir_action_127 _endpos_args_ _startpos_hd_ args hd in
        _menhir_goto_fotype_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_args_ _startpos_hd_ _v _menhir_s _tok
    
    and _menhir_run_065 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_fotype_term _menhir_cell0_ARROW -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_ARROW (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_fotype_term (_menhir_stack, _menhir_s, hd, _startpos_hd_, _) = _menhir_stack in
        let (_endpos_t_, t) = (_endpos, _v) in
        let _v = _menhir_action_359 _endpos_t_ _startpos_hd_ hd t in
        _menhir_goto_type_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _v _menhir_s _tok
    
    and _menhir_run_062 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_IO_COLON -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_IO_COLON (_menhir_stack, _menhir_s, io) = _menhir_stack in
        let (_endpos_ty_, ty) = (_endpos, _v) in
        let _v = _menhir_action_329 io ty in
        let _endpos = _endpos_ty_ in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_pred_item (_menhir_stack, _menhir_s, _v, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let x = () in
            let _v = _menhir_action_312 x in
            _menhir_goto_option_CONJ_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | Elpi_lexer_config.Tokens.IO_COLON _ ->
            let _menhir_stack = MenhirCell1_pred_item (_menhir_stack, _menhir_s, _v, _endpos) in
            let _v = _menhir_action_311 () in
            _menhir_goto_option_CONJ_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.RPAREN ->
            let (_endpos_x_, x) = (_endpos, _v) in
            let _v = _menhir_action_343 x in
            _menhir_goto_separated_nonempty_list_option_CONJ__pred_item_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _menhir_s _tok
        | _ ->
            _eRR 85
    
    and _menhir_goto_option_CONJ_ : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_pred_item -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let _menhir_stack = MenhirCell0_option_CONJ_ (_menhir_stack, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.IO_COLON _v_0 ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState087
        | _ ->
            _eRR 87
    
    and _menhir_goto_separated_nonempty_list_option_CONJ__pred_item_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState087 ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState516 ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState060 ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_088 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_pred_item _menhir_cell0_option_CONJ_ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_option_CONJ_ (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_pred_item (_menhir_stack, _menhir_s, x, _) = _menhir_stack in
        let (_endpos_xs_, xs) = (_endpos, _v) in
        let _v = _menhir_action_344 x xs in
        _menhir_goto_separated_nonempty_list_option_CONJ__pred_item_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_xs_ _v _menhir_s _tok
    
    and _menhir_run_084 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let (_endpos_x_, x) = (_endpos, _v) in
        let _v = _menhir_action_193 x in
        _menhir_goto_loption_separated_nonempty_list_option_CONJ__pred_item__ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _menhir_s _tok
    
    and _menhir_goto_loption_separated_nonempty_list_option_CONJ__pred_item__ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState516 ->
            _menhir_run_517 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState060 ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_517 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_PRED, _menhir_box_program) _menhir_cell1_constant -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_constant (_menhir_stack, _, name, _, _) = _menhir_stack in
        let MenhirCell1_PRED (_menhir_stack, _, _startpos__2_) = _menhir_stack in
        let MenhirCell1_attributes (_menhir_stack, _menhir_s, attributes, _startpos_attributes_, _endpos_attributes_) = _menhir_stack in
        let (_endpos_xs_, xs) = (_endpos, _v) in
        let _v = _menhir_action_326 _endpos_attributes_ _endpos_xs_ _startpos__2_ _startpos_attributes_ attributes name xs in
        _menhir_goto_pred _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_078 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LPAREN -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _) = _menhir_stack in
            let (_endpos__3_, t) = (_endpos, _v) in
            let _v = _menhir_action_029 t in
            _menhir_goto_atype_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _v _menhir_s _tok
        | _ ->
            _eRR 78
    
    and _menhir_run_510 : type  ttv_stack. ((((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_RULE, _menhir_box_program) _menhir_cell1_list_sequent_ _menhir_cell0_option_preceded_BIND_nonempty_list_sequent___ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF ->
            let x = _v in
            let _v = _menhir_action_322 x in
            _menhir_goto_option_preceded_PIPE_term__ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _eRR 510
    
    and _menhir_run_498 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_LPAREN, _menhir_box_program) _menhir_cell1_closed_term as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_closed_term (_menhir_stack, _, c, _, _) = _menhir_stack in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (t, _endpos__5_) = (_v, _endpos_0) in
            let _v = _menhir_action_346 _endpos__5_ _startpos__1_ c t in
            _menhir_goto_sequent _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__5_ _v _menhir_s _tok
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_13 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 498
    
    and _menhir_run_397 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_MACRO, _menhir_box_program) _menhir_cell1_term _menhir_cell0_VDASH as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let MenhirCell0_VDASH (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _, m, _, _) = _menhir_stack in
            let MenhirCell1_MACRO (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_endpos_b_, b) = (_endpos, _v) in
            let _v = _menhir_action_194 _endpos_b_ _startpos__1_ b m in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let m = _v in
            let _v = _menhir_action_110 m in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 397
    
    and _menhir_run_395 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_MACRO as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_VDASH (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState396 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 396)
        | Elpi_lexer_config.Tokens.SLASH ->
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.QDASH ->
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_9 ->
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_10 ->
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_11 ->
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_12 ->
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_13 ->
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_14 ->
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_14
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_15 ->
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_15
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_16 ->
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_16
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_17 ->
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_17
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_18 ->
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_18
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_19 ->
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_19
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_20 ->
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_20
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_21 ->
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_21
        | Elpi_lexer_config.Tokens.EQ2 ->
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 395
    
    and _menhir_run_310 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_goal) _menhir_state -> _ -> _menhir_box_goal =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let g = _v in
            let _v = _menhir_action_131 g in
            _menhir_goto_goal _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_13 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EOF ->
            let g = _v in
            let _v = _menhir_action_130 g in
            _menhir_goto_goal _menhir_stack _v
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 310
    
    and _menhir_run_308 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_nonempty_list_constant_w_loc_, ttv_result) _menhir_cell1_COLON, ttv_result) _menhir_cell1_type_term _menhir_cell0_BIND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN ->
            let MenhirCell0_BIND (_menhir_stack, _startpos_bind_, _endpos_bind_) = _menhir_stack in
            let MenhirCell1_type_term (_menhir_stack, _, ty, _) = _menhir_stack in
            let MenhirCell1_COLON (_menhir_stack, _menhir_s, _) = _menhir_stack in
            let (_endpos_b_, b, bind) = (_endpos, _v, ()) in
            let _v = _menhir_action_031 _endpos_bind_ _startpos_bind_ b bind ty in
            _menhir_goto_binder_body _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _v _menhir_s _tok
        | _ ->
            _eRR 308
    
    and _menhir_goto_binder_body : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_nonempty_list_constant_w_loc_ as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState519 ->
            _menhir_run_520 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState491 ->
            _menhir_run_492 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState057 ->
            _menhir_run_309 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState098 ->
            _menhir_run_307 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState114 ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState109 ->
            _menhir_run_110 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_520 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_PI, _menhir_box_program) _menhir_cell1_nonempty_list_constant_w_loc_ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _, args) = _menhir_stack in
        let MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos_hd_, _endpos_hd_) = _menhir_stack in
        let (_endpos_b_, b, hd) = (_endpos, _v, ()) in
        let _v = _menhir_action_042 _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd in
        _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_hd_ _v _menhir_s _tok
    
    and _menhir_run_492 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_SIGMA, _menhir_box_program) _menhir_cell1_nonempty_list_constant_w_loc_ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _, args) = _menhir_stack in
        let MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos_hd_, _endpos_hd_) = _menhir_stack in
        let (_endpos_b_, b, hd) = (_endpos, _v, ()) in
        let _v = _menhir_action_043 _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd in
        _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_hd_ _v _menhir_s _tok
    
    and _menhir_run_309 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_SIGMA, ttv_result) _menhir_cell1_nonempty_list_constant_w_loc_ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _, args) = _menhir_stack in
        let MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos_hd_, _endpos_hd_) = _menhir_stack in
        let (_endpos_b_, b, hd) = (_endpos, _v, ()) in
        let _v = _menhir_action_241 _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd in
        _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_hd_ _v _menhir_s _tok
    
    and _menhir_run_307 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_PI, ttv_result) _menhir_cell1_nonempty_list_constant_w_loc_ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _, args) = _menhir_stack in
        let MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos_hd_, _endpos_hd_) = _menhir_stack in
        let (_endpos_b_, b, hd) = (_endpos, _v, ()) in
        let _v = _menhir_action_240 _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd in
        _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_hd_ _v _menhir_s _tok
    
    and _menhir_run_115 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_PI, ttv_result) _menhir_cell1_nonempty_list_constant_w_loc_ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _, args) = _menhir_stack in
        let MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos_hd_, _endpos_hd_) = _menhir_stack in
        let (_endpos_b_, b, hd) = (_endpos, _v, ()) in
        let _v = _menhir_action_276 _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd in
        _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_hd_ _v _menhir_s _tok
    
    and _menhir_goto_open_term_noconj : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_353 t in
        _menhir_goto_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
    
    and _menhir_goto_term_noconj : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState283 ->
            _menhir_run_284 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState286 ->
            _menhir_run_281 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState107 ->
            _menhir_run_281 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState279 ->
            _menhir_run_280 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState276 ->
            _menhir_run_277 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState274 ->
            _menhir_run_275 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState272 ->
            _menhir_run_273 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState270 ->
            _menhir_run_271 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState268 ->
            _menhir_run_269 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState266 ->
            _menhir_run_267 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState264 ->
            _menhir_run_265 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState262 ->
            _menhir_run_263 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState260 ->
            _menhir_run_261 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState258 ->
            _menhir_run_259 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState256 ->
            _menhir_run_257 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState254 ->
            _menhir_run_255 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState252 ->
            _menhir_run_253 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState250 ->
            _menhir_run_251 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState248 ->
            _menhir_run_249 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState246 ->
            _menhir_run_247 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState244 ->
            _menhir_run_245 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState242 ->
            _menhir_run_243 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState240 ->
            _menhir_run_241 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState238 ->
            _menhir_run_239 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState236 ->
            _menhir_run_237 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState234 ->
            _menhir_run_235 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState232 ->
            _menhir_run_233 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState230 ->
            _menhir_run_231 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState228 ->
            _menhir_run_229 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState226 ->
            _menhir_run_227 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState224 ->
            _menhir_run_225 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState222 ->
            _menhir_run_223 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState220 ->
            _menhir_run_221 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState112 ->
            _menhir_run_219 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState118 ->
            _menhir_run_119 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_284 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_220 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.RBRACKET ->
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
            let (xs, _endpos__4_) = (_v, _endpos_1) in
            let _v = _menhir_action_188 x xs in
            _menhir_goto_list_items_tail _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__4_ _v _menhir_s _tok
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_279 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_236 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_264 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_11 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_12 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_13 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_14 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_14
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_266 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_268 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_270 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_274 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_276 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 284
    
    and _menhir_run_220 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_VDASH (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState220 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 220
    
    and _menhir_run_222 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_SLASH (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState222 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 222
    
    and _menhir_goto_list_items_tail : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState107 ->
            _menhir_run_289 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState286 ->
            _menhir_run_287 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_289 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACKET -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_LBRACKET (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_endpos_l_, l) = (_endpos, _v) in
        let _v = _menhir_action_085 _endpos_l_ _startpos__1_ l in
        _menhir_goto_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_l_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_287 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_CONJ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_CONJ (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let (_endpos_xs_, xs) = (_endpos, _v) in
        let _v = _menhir_action_189 x xs in
        _menhir_goto_list_items_tail _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_xs_ _v _menhir_s _tok
    
    and _menhir_run_279 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_QDASH (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState279 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 279
    
    and _menhir_run_236 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_OR (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState236 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 236
    
    and _menhir_run_238 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_MOD (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState238 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 238
    
    and _menhir_run_240 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_MINUSs (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState240 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 240
    
    and _menhir_run_246 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_MINUSr (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState246 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 246
    
    and _menhir_run_248 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_MINUSi (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState248 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 248
    
    and _menhir_run_250 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_MINUS (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState250 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 250
    
    and _menhir_run_252 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_IS (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState252 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 252
    
    and _menhir_run_242 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_TIMES (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState242 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 242
    
    and _menhir_run_254 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_TICK (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState254 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 254
    
    and _menhir_run_224 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_SHARP (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState224 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 224
    
    and _menhir_run_120 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_310 _endpos_s_ _startpos_l_ _startpos_s_ l s in
        _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_l_ _v _menhir_s _tok
    
    and _menhir_run_256 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_PLUS (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState256 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 256
    
    and _menhir_run_226 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_OR (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState226 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 226
    
    and _menhir_run_230 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_MINUS (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState230 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 230
    
    and _menhir_run_262 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_LT (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState262 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 262
    
    and _menhir_run_264 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_GT (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState264 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 264
    
    and _menhir_run_258 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_EXP (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState258 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 258
    
    and _menhir_run_232 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_EQ (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState232 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 232
    
    and _menhir_run_234 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_BTICK (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState234 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 234
    
    and _menhir_run_228 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_FAMILY_AND (_menhir_stack, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState228 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 228
    
    and _menhir_run_266 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_EQ2 (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState266 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 266
    
    and _menhir_run_268 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_EQ (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState268 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 268
    
    and _menhir_run_244 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_DIV (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState244 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 244
    
    and _menhir_run_270 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_DDARROW (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState270 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 270
    
    and _menhir_run_272 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_DARROW (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState272 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 272
    
    and _menhir_run_260 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_CONS (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState260 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 260
    
    and _menhir_run_274 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_CONJ2 (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState274 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 274
    
    and _menhir_run_276 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_term_noconj -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_ARROW (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState276 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 276
    
    and _menhir_run_281 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_220 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.RBRACKET ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (x, _endpos_r_, _startpos_r_, r) = (_v, _endpos_1, _startpos_0, ()) in
            let _v = _menhir_action_185 _endpos_r_ _startpos_r_ r x in
            _menhir_goto_list_items _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _v _menhir_s _tok
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_279 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.PIPE ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState283 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 283)
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_236 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_11 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_12 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_13 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_14 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_14
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_15 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_15
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_16 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _v_16
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_17 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_264 _menhir_stack _menhir_lexbuf _menhir_lexer _v_17
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_18 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_18
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_19 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_19
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_20 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_20
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_21 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_21
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_266 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_268 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_270 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_274 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_CONJ (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState286 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.RBRACKET ->
                _menhir_run_111 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PIPE ->
                _menhir_run_112 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_118 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 286)
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_276 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 281
    
    and _menhir_run_280 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_QDASH as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_236 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_264 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_266 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_268 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_270 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_274 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_276 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.RBRACKET ->
            let MenhirCell0_QDASH (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_305 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 280
    
    and _menhir_run_277 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_ARROW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_264 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_266 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_268 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_276 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_ARROW (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_302 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 277
    
    and _menhir_run_275 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_CONJ2 as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_264 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_266 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_268 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_274 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_276 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_CONJ2 (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_307 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 275
    
    and _menhir_run_273 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_DARROW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_264 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_266 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_268 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_DARROW (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_303 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 273
    
    and _menhir_run_271 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_DDARROW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_264 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_266 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_268 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_270 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_274 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_276 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_DDARROW (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_304 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 271
    
    and _menhir_run_269 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_EQ (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_292 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 269
    
    and _menhir_run_267 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_EQ2 as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_EQ2 (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_297 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 267
    
    and _menhir_run_265 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_GT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_GT (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_284 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 265
    
    and _menhir_run_263 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_LT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_LT (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_283 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 263
    
    and _menhir_run_261 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_CONS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_CONS (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_291 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 261
    
    and _menhir_run_259 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_EXP as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_EXP (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_282 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 259
    
    and _menhir_run_257 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_PLUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_PLUS (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_279 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 257
    
    and _menhir_run_255 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_TICK as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_TICK (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_290 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 255
    
    and _menhir_run_253 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_IS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_IS (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_299 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 253
    
    and _menhir_run_251 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUS (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_293 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 251
    
    and _menhir_run_249 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_MINUSi as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUSi (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_295 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 249
    
    and _menhir_run_247 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_MINUSr as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUSr (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_294 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 247
    
    and _menhir_run_245 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_DIV as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_DIV (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_301 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 245
    
    and _menhir_run_243 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_TIMES as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_TIMES (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_280 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 243
    
    and _menhir_run_241 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_MINUSs as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUSs (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_296 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 241
    
    and _menhir_run_239 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_MOD as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MOD (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_300 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 239
    
    and _menhir_run_237 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_OR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_236 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_264 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_266 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_268 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_270 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_274 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_276 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_OR (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_298 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 237
    
    and _menhir_run_235 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_BTICK as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_BTICK (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_289 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 235
    
    and _menhir_run_233 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_EQ (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_285 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 233
    
    and _menhir_run_231 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_MINUS (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_281 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 231
    
    and _menhir_run_229 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_AND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_AND (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_286 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 229
    
    and _menhir_run_227 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_OR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_OR (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_287 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 227
    
    and _menhir_run_225 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_FAMILY_SHARP as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_SHARP (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_288 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 225
    
    and _menhir_run_223 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_SLASH as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_SLASH (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_306 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 223
    
    and _menhir_run_221 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term_noconj _menhir_cell0_VDASH as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_236 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_264 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_266 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_268 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_270 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_274 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_276 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.RBRACKET ->
            let MenhirCell0_VDASH (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term_noconj (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_308 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 221
    
    and _menhir_run_219 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_PIPE as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_220 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.RBRACKET ->
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_PIPE (_menhir_stack, _menhir_s) = _menhir_stack in
            let (x, _endpos__3_) = (_v, _endpos_1) in
            let _v = _menhir_action_187 x in
            _menhir_goto_list_items_tail _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _v _menhir_s _tok
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_279 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_236 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_3 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_254 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_4 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_224 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_5 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_6 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_7 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_8 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_230 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_9 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_10 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_264 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_11 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_258 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_12 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_13 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_14 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v_14
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_266 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_268 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_270 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_274 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_276 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 219
    
    and _menhir_run_119 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_FAMILY_TILDE as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_0 ->
            let _menhir_stack = MenhirCell1_term_noconj (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_309 _endpos_r_ _endpos_s_ _startpos_s_ r s in
            _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_s_ _v _menhir_s _tok
        | _ ->
            _eRR 119
    
    and _menhir_run_110 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_SIGMA, ttv_result) _menhir_cell1_nonempty_list_constant_w_loc_ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _, args) = _menhir_stack in
        let MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos_hd_, _endpos_hd_) = _menhir_stack in
        let (_endpos_b_, b, hd) = (_endpos, _v, ()) in
        let _v = _menhir_action_277 _endpos_b_ _endpos_hd_ _startpos_hd_ args b hd in
        _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_hd_ _v _menhir_s _tok
    
    and _menhir_run_306 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_nonempty_list_constant_w_loc_, ttv_result) _menhir_cell1_BIND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN ->
            let MenhirCell1_BIND (_menhir_stack, _menhir_s, _startpos_bind_, _endpos_bind_) = _menhir_stack in
            let (_endpos_b_, b, bind) = (_endpos, _v, ()) in
            let _v = _menhir_action_030 _endpos_bind_ _startpos_bind_ b bind in
            _menhir_goto_binder_body _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _v _menhir_s _tok
        | _ ->
            _eRR 306
    
    and _menhir_run_300 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_LPAREN, ttv_result) _menhir_cell1_term _menhir_cell0_AS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_AS (_menhir_stack, _startpos_a_, _endpos_a_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _, t, _, _) = _menhir_stack in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (c, _endpos__5_, a) = (_v, _endpos_0, ()) in
            let _v = _menhir_action_083 _endpos__5_ _endpos_a_ _startpos__1_ _startpos_a_ a c t in
            _menhir_goto_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__5_ _startpos__1_ _v _menhir_s _tok
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_13 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 300
    
    and _menhir_run_294 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_endpos__3_, t) = (_endpos_0, _v) in
            let _v = _menhir_action_133 _endpos__3_ _startpos__1_ t in
            _menhir_goto_head_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_13 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.COLON ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_296 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_AS (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState299 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 299)
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 294
    
    and _menhir_run_291 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LCURLY as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.RCURLY ->
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LCURLY (_menhir_stack, _menhir_s, _startpos_l_, _endpos_l_) = _menhir_stack in
            let (_endpos__3_, t, l) = (_endpos_1, _v, ()) in
            let _v = _menhir_action_086 _endpos__3_ _endpos_l_ _startpos_l_ l t in
            _menhir_goto_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_l_ _v _menhir_s _tok
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_13 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_14 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_14
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR 291
    
    and _menhir_run_216 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_constant _menhir_cell0_BIND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.RBRACKET ->
            let MenhirCell0_BIND (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_constant (_menhir_stack, _menhir_s, t, _startpos_t_, _) = _menhir_stack in
            let (_endpos_b_, b) = (_endpos, _v) in
            let _v = _menhir_action_035 _endpos_b_ _startpos_t_ b t in
            _menhir_goto_binder_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_t_ _v _menhir_s _tok
        | _ ->
            _eRR 216
    
    and _menhir_goto_binder_term_noconj : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_354 t in
        _menhir_goto_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
    
    and _menhir_run_214 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_constant _menhir_cell0_COLON, ttv_result) _menhir_cell1_type_term _menhir_cell0_BIND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.RBRACKET ->
            let MenhirCell0_BIND (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_type_term (_menhir_stack, _, ty, _) = _menhir_stack in
            let MenhirCell0_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_constant (_menhir_stack, _menhir_s, t, _startpos_t_, _) = _menhir_stack in
            let (_endpos_b_, b) = (_endpos, _v) in
            let _v = _menhir_action_036 _endpos_b_ _startpos_t_ b t ty in
            _menhir_goto_binder_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_t_ _v _menhir_s _tok
        | _ ->
            _eRR 214
    
    and _menhir_run_208 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_head_term, ttv_result) _menhir_cell1_nonempty_list_closed_term_, ttv_result) _menhir_cell1_BIND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN ->
            let MenhirCell1_BIND (_menhir_stack, _menhir_s, _startpos_bind_, _endpos_bind_) = _menhir_stack in
            let (_endpos_b_, b, bind) = (_endpos, _v, ()) in
            let _v = _menhir_action_032 _endpos_bind_ _startpos_bind_ b bind in
            let _endpos = _endpos_b_ in
            let (_endpos_x_, x) = (_endpos, _v) in
            let _v = _menhir_action_316 x in
            _menhir_goto_option_binder_body_no_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _menhir_s _tok
        | _ ->
            _eRR 208
    
    and _menhir_goto_option_binder_body_no_ty_ : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_head_term, ttv_result) _menhir_cell1_nonempty_list_closed_term_ as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState532 ->
            _menhir_run_533 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState123 ->
            _menhir_run_209 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState130 ->
            _menhir_run_131 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_533 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_head_term, _menhir_box_program) _menhir_cell1_nonempty_list_closed_term_ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_nonempty_list_closed_term_ (_menhir_stack, _, args) = _menhir_stack in
        let MenhirCell1_head_term (_menhir_stack, _menhir_s, hd, _startpos_hd_, _) = _menhir_stack in
        let (_endpos_b_, b) = (_endpos, _v) in
        let _v = _menhir_action_044 _endpos_b_ _startpos_hd_ args b hd in
        _menhir_goto_clause_hd_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_hd_ _v _menhir_s _tok
    
    and _menhir_run_209 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_head_term, ttv_result) _menhir_cell1_nonempty_list_closed_term_ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_nonempty_list_closed_term_ (_menhir_stack, _, args) = _menhir_stack in
        let MenhirCell1_head_term (_menhir_stack, _menhir_s, hd, _startpos_hd_, _) = _menhir_stack in
        let (_endpos_b_, b) = (_endpos, _v) in
        let _v = _menhir_action_278 _endpos_b_ _startpos_hd_ args b hd in
        _menhir_goto_open_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_hd_ _v _menhir_s _tok
    
    and _menhir_run_131 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_head_term, ttv_result) _menhir_cell1_nonempty_list_closed_term_ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_nonempty_list_closed_term_ (_menhir_stack, _, args) = _menhir_stack in
        let MenhirCell1_head_term (_menhir_stack, _menhir_s, hd, _startpos_hd_, _) = _menhir_stack in
        let (_endpos_b_, b) = (_endpos, _v) in
        let _v = _menhir_action_242 _endpos_b_ _startpos_hd_ args b hd in
        _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_hd_ _v _menhir_s _tok
    
    and _menhir_run_207 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_constant _menhir_cell0_BIND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN ->
            let MenhirCell0_BIND (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_constant (_menhir_stack, _menhir_s, t, _startpos_t_, _) = _menhir_stack in
            let (_endpos_b_, b) = (_endpos, _v) in
            let _v = _menhir_action_033 _endpos_b_ _startpos_t_ b t in
            _menhir_goto_binder_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_t_ _v _menhir_s _tok
        | _ ->
            _eRR 207
    
    and _menhir_goto_binder_term : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_351 t in
        _menhir_goto_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
    
    and _menhir_run_205 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_QDASH as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN ->
            let MenhirCell0_QDASH (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_269 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 205
    
    and _menhir_run_203 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_ARROW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_ARROW (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_266 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 203
    
    and _menhir_run_201 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_CONJ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_CONJ (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_272 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 201
    
    and _menhir_run_199 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_CONJ2 as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_CONJ2 (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_271 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 199
    
    and _menhir_run_197 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_DARROW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_DARROW (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_267 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 197
    
    and _menhir_run_195 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_DDARROW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_DDARROW (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_268 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 195
    
    and _menhir_run_193 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_EQ (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_256 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 193
    
    and _menhir_run_191 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_EQ2 as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_EQ2 (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_261 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 191
    
    and _menhir_run_189 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_GT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_GT (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_248 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 189
    
    and _menhir_run_187 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_LT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_LT (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_247 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 187
    
    and _menhir_run_185 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_CONS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_CONS (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_255 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 185
    
    and _menhir_run_183 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_EXP as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_EXP (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_246 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 183
    
    and _menhir_run_181 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_PLUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_PLUS (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_243 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 181
    
    and _menhir_run_179 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_TICK as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_TICK (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_254 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 179
    
    and _menhir_run_177 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_IS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_IS (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_263 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 177
    
    and _menhir_run_175 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUS (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_257 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 175
    
    and _menhir_run_173 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_MINUSi as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUSi (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_259 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 173
    
    and _menhir_run_171 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_MINUSr as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUSr (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_258 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 171
    
    and _menhir_run_169 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_DIV as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_DIV (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_265 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 169
    
    and _menhir_run_167 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_TIMES as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_TIMES (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_244 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 167
    
    and _menhir_run_165 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_MINUSs as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MINUSs (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_260 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 165
    
    and _menhir_run_163 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_MOD as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_MOD (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_264 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 163
    
    and _menhir_run_161 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_OR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_OR (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_262 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 161
    
    and _menhir_run_159 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_BTICK as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_BTICK (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_253 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 159
    
    and _menhir_run_157 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_EQ (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_249 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 157
    
    and _menhir_run_155 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_MINUS (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_245 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 155
    
    and _menhir_run_153 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_AND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_AND (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_250 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 153
    
    and _menhir_run_151 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_OR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_OR (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_251 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 151
    
    and _menhir_run_147 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_FAMILY_SHARP as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_FAMILY_SHARP (_menhir_stack, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_252 _endpos_r_ _endpos_s_ _startpos_l_ _startpos_s_ l r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 147
    
    and _menhir_run_145 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_SLASH as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell0_SLASH (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_270 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 145
    
    and _menhir_run_143 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_term _menhir_cell0_VDASH as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN ->
            let MenhirCell0_VDASH (_menhir_stack, _startpos__1_, _endpos__1_) = _menhir_stack in
            let MenhirCell1_term (_menhir_stack, _menhir_s, l, _startpos_l_, _) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_273 _endpos__1_ _endpos_r_ _startpos__1_ _startpos_l_ l r in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_l_ _v _menhir_s _tok
        | _ ->
            _eRR 143
    
    and _menhir_run_141 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_constant _menhir_cell0_COLON, ttv_result) _menhir_cell1_type_term _menhir_cell0_BIND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.SLASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_144 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.OR ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MOD ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_162 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.IS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v_1 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_178 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v_2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_3 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v_4 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_180 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4
        | Elpi_lexer_config.Tokens.FAMILY_OR _v_5 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_150 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v_6 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | Elpi_lexer_config.Tokens.FAMILY_LT _v_7 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7
        | Elpi_lexer_config.Tokens.FAMILY_GT _v_8 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v_9 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v_10 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_156 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v_11 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11
        | Elpi_lexer_config.Tokens.FAMILY_AND _v_12 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.EQ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DIV ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.DARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONS ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN ->
            let MenhirCell0_BIND (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_type_term (_menhir_stack, _, ty, _) = _menhir_stack in
            let MenhirCell0_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_constant (_menhir_stack, _menhir_s, t, _startpos_t_, _) = _menhir_stack in
            let (_endpos_b_, b) = (_endpos, _v) in
            let _v = _menhir_action_034 _endpos_b_ _startpos_t_ b t ty in
            _menhir_goto_binder_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_t_ _v _menhir_s _tok
        | _ ->
            _eRR 141
    
    and _menhir_run_126 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_FAMILY_TILDE as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v_0 ->
            let _menhir_stack = MenhirCell1_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_127 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, s, _startpos_s_, _endpos_s_) = _menhir_stack in
            let (_endpos_r_, r) = (_endpos, _v) in
            let _v = _menhir_action_274 _endpos_r_ _endpos_s_ _startpos_s_ r s in
            _menhir_goto_open_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_s_ _v _menhir_s _tok
        | _ ->
            _eRR 126
    
    and _menhir_run_217 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_355 t in
        _menhir_goto_term_noconj _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
    
    and _menhir_run_148 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_352 t in
        _menhir_goto_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
    
    and _menhir_run_135 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.STRING _v_0 ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState135
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.QUOTED _v_1 ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState135
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.LCURLY ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.IO _v_2 ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState135
        | Elpi_lexer_config.Tokens.INTEGER _v_3 ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState135
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.FLOAT _v_4 ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState135
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.CONSTANT _v_5 ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState135
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_closed_term (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState135
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.BIND | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let x = _v in
            let _v = _menhir_action_232 x in
            _menhir_goto_nonempty_list_closed_term_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 135
    
    and _menhir_goto_nonempty_list_closed_term_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState531 ->
            _menhir_run_532 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState135 ->
            _menhir_run_136 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState129 ->
            _menhir_run_130 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState122 ->
            _menhir_run_123 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_532 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_head_term as 'stack) -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_nonempty_list_closed_term_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.BIND ->
            _menhir_run_124 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState532
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let _v_0 = _menhir_action_315 () in
            _menhir_run_533 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v_0 _tok
        | _ ->
            _eRR 532
    
    and _menhir_run_124 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_head_term, ttv_result) _menhir_cell1_nonempty_list_closed_term_ as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_BIND (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState124 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 124
    
    and _menhir_run_136 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_closed_term -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_closed_term (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let xs = _v in
        let _v = _menhir_action_233 x xs in
        _menhir_goto_nonempty_list_closed_term_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
    
    and _menhir_run_130 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_head_term as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_nonempty_list_closed_term_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.BIND ->
            _menhir_run_124 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState130
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let _v_0 = _menhir_action_315 () in
            _menhir_run_131 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v_0 _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_123 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_head_term as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_nonempty_list_closed_term_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.BIND ->
            _menhir_run_124 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState123
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let _v_0 = _menhir_action_315 () in
            _menhir_run_209 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v_0 _tok
        | _ ->
            _eRR 123
    
    and _menhir_run_398 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LOCALKIND (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState398 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 398
    
    and _menhir_run_400 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LOCAL (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState400 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 400
    
    and _menhir_run_405 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_116 _endpos__1_ _startpos__1_ in
        _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
    
    and _menhir_run_406 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_KIND (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState406 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 406
    
    and _menhir_run_412 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _startpos__1_ = _startpos in
        let _v = _menhir_action_008 () in
        _menhir_goto_accumulate _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_413 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_FIXITY (_menhir_stack, _menhir_s, _v, _startpos) in
        let _menhir_s = MenhirState413 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SLASH ->
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QDASH ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.OR ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MOD ->
            _menhir_run_015 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSs ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSr ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUSi ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MINUS ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IS ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v ->
            _menhir_run_021 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_022 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v ->
            _menhir_run_024 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_OR _v ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_LT _v ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_GT _v ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_AND _v ->
            _menhir_run_034 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.EQ2 ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.EQ ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DIV ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DDARROW ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.DARROW ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CONS ->
            _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ2 ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONJ ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.ARROW ->
            _menhir_run_045 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 413
    
    and _menhir_run_418 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
        let _v = _menhir_action_025 () in
        _menhir_goto_attributes _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_attributes : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_attributes (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | Elpi_lexer_config.Tokens.TYPE ->
            let _menhir_s = MenhirState486 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_TYPE (_menhir_stack, _menhir_s, _startpos, _endpos) in
            let _menhir_s = MenhirState487 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 487)
        | Elpi_lexer_config.Tokens.SIGMA ->
            let _menhir_s = MenhirState486 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_SIGMA (_menhir_stack, _menhir_s, _startpos, _endpos) in
            let _menhir_s = MenhirState490 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 490)
        | Elpi_lexer_config.Tokens.RULE ->
            let _endpos_8 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_RULE (_menhir_stack, MenhirState486) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.STRING _v_9 ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9 MenhirState493
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.QUOTED _v_10 ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10 MenhirState493
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.IO _v_11 ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11 MenhirState493
            | Elpi_lexer_config.Tokens.INTEGER _v_12 ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12 MenhirState493
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.FLOAT _v_13 ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13 MenhirState493
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.CONSTANT _v_14 ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_14 MenhirState493
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState493
            | Elpi_lexer_config.Tokens.BIND | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.PIPE ->
                let _v_15 = _menhir_action_182 () in
                _menhir_run_503 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_8 _v_15 MenhirState493 _tok
            | _ ->
                _eRR 493)
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | Elpi_lexer_config.Tokens.PRED ->
            let _menhir_s = MenhirState486 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell1_PRED (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_s = MenhirState515 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 515)
        | Elpi_lexer_config.Tokens.PI ->
            let _menhir_s = MenhirState486 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_PI (_menhir_stack, _menhir_s, _startpos, _endpos) in
            let _menhir_s = MenhirState518 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 518)
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_s = MenhirState486 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_s = MenhirState521 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.VDASH ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SLASH ->
                _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.QDASH ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.OR ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.MOD ->
                _menhir_run_015 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.MINUSs ->
                _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.MINUSr ->
                _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.MINUSi ->
                _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.MINUS ->
                _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IS ->
                _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TIMES _v ->
                _menhir_run_021 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_293 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TICK _v ->
                _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_SHARP _v ->
                _menhir_run_024 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_QMARK _v ->
                _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_PLUS _v ->
                _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_OR _v ->
                _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_MINUS _v ->
                _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_LT _v ->
                _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_GT _v ->
                _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_EXP _v ->
                _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_EQ _v ->
                _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_BTICK _v ->
                _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_AND _v ->
                _menhir_run_034 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.EQ2 ->
                _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.EQ ->
                _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.DIV ->
                _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.DDARROW ->
                _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.DARROW ->
                _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CONS ->
                _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONJ2 ->
                _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONJ ->
                _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AS ->
                _menhir_run_043 _menhir_stack _menhir_lexbuf _menhir_lexer
            | Elpi_lexer_config.Tokens.ARROW ->
                _menhir_run_045 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 521)
        | Elpi_lexer_config.Tokens.IO _v_44 ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_44 MenhirState486
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | Elpi_lexer_config.Tokens.FUNC ->
            let _menhir_s = MenhirState486 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_s = MenhirState524 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 524)
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v_48 ->
            let (_v, _menhir_s) = (_v_48, MenhirState486) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_FAMILY_TILDE (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState529 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 529)
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | Elpi_lexer_config.Tokens.CONSTANT _v_58 ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_58 MenhirState486
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState486
        | _ ->
            _eRR 486
    
    and _menhir_run_419 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_EXPORTDEF (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState419 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 419
    
    and _menhir_run_423 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_CONSTRAINT (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState423
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState423
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState423
        | Elpi_lexer_config.Tokens.LCURLY | Elpi_lexer_config.Tokens.QDASH ->
            let _v = _menhir_action_180 () in
            _menhir_run_424 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState423 _tok
        | _ ->
            _eRR 423
    
    and _menhir_run_424 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_CONSTRAINT as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.QDASH ->
            let _menhir_stack = MenhirCell1_list_constant_ (_menhir_stack, _menhir_s, _v) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_QDASH (_menhir_stack, _startpos, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.IO _v_0 ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState425
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState425
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState425
            | Elpi_lexer_config.Tokens.LCURLY ->
                let _v_2 = _menhir_action_180 () in
                _menhir_run_426 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 _tok
            | _ ->
                _eRR 425)
        | Elpi_lexer_config.Tokens.LCURLY ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_CONSTRAINT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_endpos__3_, cl) = (_endpos, _v) in
            let _v = _menhir_action_112 _endpos__3_ _startpos__1_ cl in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_426 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_CONSTRAINT, _menhir_box_program) _menhir_cell1_list_constant_ _menhir_cell0_QDASH -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.LCURLY ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_QDASH (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_list_constant_ (_menhir_stack, _, hyps) = _menhir_stack in
            let MenhirCell1_CONSTRAINT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_endpos__5_, cl) = (_endpos, _v) in
            let _v = _menhir_action_111 _endpos__5_ _startpos__1_ cl hyps in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 426
    
    and _menhir_run_431 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_COLON (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState431 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_432 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_433 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_437 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_449 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_451 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.EXTERNAL ->
            _menhir_run_452 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_455 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 431
    
    and _menhir_run_432 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _endpos__1_ = _endpos in
        let _v = _menhir_action_022 () in
        _menhir_goto_attribute _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_attribute : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.COLON ->
            let _menhir_stack = MenhirCell1_attribute (_menhir_stack, _menhir_s, _v, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_COLON (_menhir_stack, _startpos) in
            let _menhir_s = MenhirState459 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_432 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_433 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_435 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_437 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_449 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_451 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.EXTERNAL ->
                _menhir_run_452 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_455 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 459)
        | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.FAMILY_TILDE _ | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FUNC | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.PI | Elpi_lexer_config.Tokens.PRED | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.RULE | Elpi_lexer_config.Tokens.SIGMA | Elpi_lexer_config.Tokens.TYPE | Elpi_lexer_config.Tokens.UNTYPED ->
            let (_endpos_x_, x) = (_endpos, _v) in
            let _v = _menhir_action_333 x in
            _menhir_goto_separated_nonempty_list_COLON_attribute_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _menhir_s _tok
        | _ ->
            _eRR 458
    
    and _menhir_run_433 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.STRING _v ->
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_endpos_s_, s) = (_endpos_1, _v) in
            let _v = _menhir_action_018 s in
            _menhir_goto_attribute _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _v _menhir_s _tok
        | _ ->
            _eRR 433
    
    and _menhir_run_435 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.STRING _v ->
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_endpos_s_, s) = (_endpos_1, _v) in
            let _v = _menhir_action_019 s in
            _menhir_goto_attribute _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _v _menhir_s _tok
        | _ ->
            _eRR 435
    
    and _menhir_run_437 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.STRING _v ->
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_endpos_s_, s) = (_endpos_1, _v) in
            let _v = _menhir_action_015 s in
            _menhir_goto_attribute _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _v _menhir_s _tok
        | _ ->
            _eRR 437
    
    and _menhir_run_439 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_INDEX (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_LPAREN (_menhir_stack, _startpos) in
            let _menhir_s = MenhirState440 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_441 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_442 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 440)
        | _ ->
            _eRR 439
    
    and _menhir_run_441 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let i = _v in
        let _v = _menhir_action_145 i in
        _menhir_goto_indexing _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_indexing : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.INTEGER _v_0 ->
            let _menhir_stack = MenhirCell1_indexing (_menhir_stack, _menhir_s, _v) in
            _menhir_run_441 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState447
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_indexing (_menhir_stack, _menhir_s, _v) in
            _menhir_run_442 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState447
        | Elpi_lexer_config.Tokens.RPAREN ->
            let x = _v in
            let _v = _menhir_action_236 x in
            _menhir_goto_nonempty_list_indexing_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR 447
    
    and _menhir_run_442 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _v = _menhir_action_144 () in
        _menhir_goto_indexing _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_nonempty_list_indexing_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState447 ->
            _menhir_run_448 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState440 ->
            _menhir_run_443 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_448 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_indexing -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_indexing (_menhir_stack, _menhir_s, x) = _menhir_stack in
        let xs = _v in
        let _v = _menhir_action_237 x xs in
        _menhir_goto_nonempty_list_indexing_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_443 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_INDEX _menhir_cell0_LPAREN as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _menhir_stack = MenhirCell1_nonempty_list_indexing_ (_menhir_stack, _menhir_s, _v) in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_RPAREN (_menhir_stack, _endpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.STRING _v_0 ->
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_endpos_x_, x) = (_endpos_1, _v_0) in
            let _v = _menhir_action_314 x in
            _menhir_goto_option_STRING_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _tok
        | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.FAMILY_TILDE _ | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FUNC | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.PI | Elpi_lexer_config.Tokens.PRED | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.RULE | Elpi_lexer_config.Tokens.SIGMA | Elpi_lexer_config.Tokens.TYPE | Elpi_lexer_config.Tokens.UNTYPED ->
            let _v = _menhir_action_313 () in
            _menhir_goto_option_STRING_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _eRR 444
    
    and _menhir_goto_option_STRING_ : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_INDEX _menhir_cell0_LPAREN, _menhir_box_program) _menhir_cell1_nonempty_list_indexing_ _menhir_cell0_RPAREN -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_RPAREN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_nonempty_list_indexing_ (_menhir_stack, _, l) = _menhir_stack in
        let MenhirCell0_LPAREN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_INDEX (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        let (_endpos_o_, o) = (_endpos, _v) in
        let _v = _menhir_action_023 l o in
        _menhir_goto_attribute _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_o_ _v _menhir_s _tok
    
    and _menhir_run_449 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.STRING _v ->
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_endpos_s_, s) = (_endpos_1, _v) in
            let _v = _menhir_action_014 s in
            _menhir_goto_attribute _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _v _menhir_s _tok
        | _ ->
            _eRR 449
    
    and _menhir_run_451 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _endpos__1_ = _endpos in
        let _v = _menhir_action_021 () in
        _menhir_goto_attribute _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_run_452 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _endpos__1_ = _endpos in
        let _v = _menhir_action_020 () in
        _menhir_goto_attribute _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_run_453 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.STRING _v ->
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_endpos_s_, s) = (_endpos_1, _v) in
            let _v = _menhir_action_017 s in
            _menhir_goto_attribute _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _v _menhir_s _tok
        | _ ->
            _eRR 453
    
    and _menhir_run_455 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.STRING _v ->
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_endpos_s_, s) = (_endpos_1, _v) in
            let _v = _menhir_action_016 s in
            _menhir_goto_attribute _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _v _menhir_s _tok
        | _ ->
            _eRR 455
    
    and _menhir_goto_separated_nonempty_list_COLON_attribute_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState459 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState431 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_460 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_attribute _menhir_cell0_COLON -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_COLON (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_attribute (_menhir_stack, _menhir_s, x, _) = _menhir_stack in
        let (_endpos_xs_, xs) = (_endpos, _v) in
        let _v = _menhir_action_334 x xs in
        _menhir_goto_separated_nonempty_list_COLON_attribute_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_xs_ _v _menhir_s _tok
    
    and _menhir_run_457 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_COLON -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_COLON (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_endpos_l_, l) = (_endpos, _v) in
        let _v = _menhir_action_026 l in
        _menhir_goto_attributes _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_l_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_461 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_CLOSED (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState461 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 461
    
    and _menhir_run_463 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _startpos__1_ = _startpos in
        let _v = _menhir_action_009 () in
        _menhir_goto_accumulate _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_464 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _startpos__1_ = _startpos in
        let _v = _menhir_action_007 () in
        _menhir_goto_accumulate _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_046 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LPAREN -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (s, _endpos__3_) = (_v, _endpos_0) in
            let _v = _menhir_action_101 s in
            _menhir_goto_constant _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR 46
    
    and _menhir_goto_separated_nonempty_list_CONJ_filename_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState606 ->
            _menhir_run_607 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState601 ->
            _menhir_run_603 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_607 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_filename _menhir_cell0_CONJ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell0_CONJ (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_filename (_menhir_stack, _menhir_s, x) = _menhir_stack in
        let xs = _v in
        let _v = _menhir_action_338 x xs in
        _menhir_goto_separated_nonempty_list_CONJ_filename_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_603 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_accumulate -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_accumulate (_menhir_stack, _menhir_s, ext, _startpos_ext_) = _menhir_stack in
        let (_endpos__3_, l) = (_endpos, _v) in
        let _v = _menhir_action_118 _endpos__3_ _startpos_ext_ ext l in
        _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
    
    and _menhir_run_534 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_attributes as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.VDASH ->
            let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_040 _endpos_t_ _startpos_t_ t in
            _menhir_goto_clause_hd_closed_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
        | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.FLOAT _ | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.INTEGER _ | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.LBRACKET | Elpi_lexer_config.Tokens.LCURLY | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.QUOTED _ | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.STRING _ | Elpi_lexer_config.Tokens.UNTYPED ->
            let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_132 _endpos_t_ _startpos_t_ t in
            _menhir_goto_head_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
        | _ ->
            _eRR 534
    
    and _menhir_run_525 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_FUNC as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.IO _v_0 ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState525
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState525
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _v_2 = _menhir_action_190 () in
            _menhir_run_526 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v_2 MenhirState525 _tok
        | _ ->
            _eRR 525
    
    and _menhir_run_516 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_PRED as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.IO_COLON _v_0 ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState516
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _v_1 = _menhir_action_192 () in
            _menhir_run_517 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v_1 _tok
        | _ ->
            _eRR 516
    
    and _menhir_run_495 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.COLON ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_303 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.BIND ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_206 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FLOAT _ | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.INTEGER _ | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.LBRACKET | Elpi_lexer_config.Tokens.LCURLY | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.QUOTED _ | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.RTRI | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.STRING _ | Elpi_lexer_config.Tokens.UNTYPED | Elpi_lexer_config.Tokens.VDASH ->
            let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_132 _endpos_t_ _startpos_t_ t in
            _menhir_goto_head_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
        | _ ->
            _eRR 495
    
    and _menhir_run_303 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN, ttv_result) _menhir_cell1_constant -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell0_COLON (_menhir_stack, _startpos) in
        let _menhir_s = MenhirState303 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 303
    
    and _menhir_run_206 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_constant -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_BIND (_menhir_stack, _startpos, _endpos) in
        let _menhir_s = MenhirState206 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 206
    
    and _menhir_run_428 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.IO _v_0 ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState428
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState428
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState428
        | Elpi_lexer_config.Tokens.LCURLY | Elpi_lexer_config.Tokens.QDASH ->
            let _v_2 = _menhir_action_180 () in
            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 _tok
        | _ ->
            _eRR 428
    
    and _menhir_run_429 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_constant -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_constant (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let xs = _v in
        let _v = _menhir_action_181 x xs in
        _menhir_goto_list_constant_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_list_constant_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState428 ->
            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState425 ->
            _menhir_run_426 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState423 ->
            _menhir_run_424 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_416 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_FIXITY -> _ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.INTEGER _v_0 ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_FIXITY (_menhir_stack, _menhir_s, f, _startpos_f_) = _menhir_stack in
            let (_endpos_c_, _startpos_c_, c, i) = (_endpos, _startpos, _v, _v_0) in
            let _v = _menhir_action_124 _endpos_c_ _startpos_c_ c f i in
            _menhir_goto_fixity _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_f_ _v _menhir_s _tok
        | _ ->
            _eRR 416
    
    and _menhir_run_393 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_MODULE -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok ->
        let MenhirCell1_MODULE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let _endpos__2_ = _endpos in
        let _ = _menhir_action_136 _endpos__2_ _startpos__1_ in
        _menhir_goto_ignored _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _menhir_s _tok
    
    and _menhir_run_390 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_NAMESPACE -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.LCURLY ->
            let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_NAMESPACE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_endpos, c, _endpos__3_) = (_endpos_1, _v, _endpos_1) in
            let _v = _menhir_action_113 _endpos__3_ _startpos__1_ c in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 390
    
    and _menhir_run_381 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FULLSTOP (_menhir_stack, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.LCURLY ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
                let _menhir_stack = MenhirCell0_LCURLY (_menhir_stack, _startpos, _endpos) in
                let _menhir_s = MenhirState383 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Elpi_lexer_config.Tokens.UNTYPED ->
                    _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.REPLACE ->
                    _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.REMOVE ->
                    _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.NIL ->
                    _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.NAME ->
                    _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.LPAREN ->
                    _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.IO _v ->
                    _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Elpi_lexer_config.Tokens.INDEX ->
                    _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.IF ->
                    _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                    _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.FRESHUV ->
                    _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.CUT ->
                    _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.CONSTANT _v ->
                    _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Elpi_lexer_config.Tokens.BEFORE ->
                    _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.AFTER ->
                    _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR 383)
            | _ ->
                _eRR 382)
        | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.RCURLY ->
            let name = _v in
            let _v = _menhir_action_348 name in
            _menhir_goto_subtrie _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR 381
    
    and _menhir_goto_subtrie : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_subtrie (_menhir_stack, _menhir_s, _v) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_CONJ (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState379 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 379)
        | Elpi_lexer_config.Tokens.RCURLY ->
            let x = _v in
            let _v = _menhir_action_341 x in
            _menhir_goto_separated_nonempty_list_CONJ_subtrie_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR 378
    
    and _menhir_goto_separated_nonempty_list_CONJ_subtrie_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState377 ->
            _menhir_run_386 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState383 ->
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState379 ->
            _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_386 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_SHORTEN, _menhir_box_program) _menhir_cell1_constant _menhir_cell0_FULLSTOP _menhir_cell0_LCURLY -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell0_LCURLY (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell0_FULLSTOP (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_constant (_menhir_stack, _, c, _, _) = _menhir_stack in
        let l = _v in
        let _v = _menhir_action_356 c l in
        let l = _v in
        let _v = _menhir_action_347 l in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_SHORTEN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (s, _endpos__3_) = (_v, _endpos) in
            let _v = _menhir_action_114 _endpos__3_ _startpos__1_ s in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _eRR 373
    
    and _menhir_run_384 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_constant _menhir_cell0_FULLSTOP _menhir_cell0_LCURLY -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell0_LCURLY (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell0_FULLSTOP (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_constant (_menhir_stack, _menhir_s, prefix, _, _) = _menhir_stack in
        let l = _v in
        let _v = _menhir_action_349 l prefix in
        _menhir_goto_subtrie _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_380 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_subtrie _menhir_cell0_CONJ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell0_CONJ (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_subtrie (_menhir_stack, _menhir_s, x) = _menhir_stack in
        let xs = _v in
        let _v = _menhir_action_342 x xs in
        _menhir_goto_separated_nonempty_list_CONJ_subtrie_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_375 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_SHORTEN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_FULLSTOP (_menhir_stack, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.LCURLY ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
                let _menhir_stack = MenhirCell0_LCURLY (_menhir_stack, _startpos, _endpos) in
                let _menhir_s = MenhirState377 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Elpi_lexer_config.Tokens.UNTYPED ->
                    _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.REPLACE ->
                    _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.REMOVE ->
                    _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.NIL ->
                    _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.NAME ->
                    _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.LPAREN ->
                    _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.IO _v ->
                    _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Elpi_lexer_config.Tokens.INDEX ->
                    _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.IF ->
                    _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                    _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.FRESHUV ->
                    _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.CUT ->
                    _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.CONSTANT _v ->
                    _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Elpi_lexer_config.Tokens.BEFORE ->
                    _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Elpi_lexer_config.Tokens.AFTER ->
                    _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR 377)
            | _ ->
                _eRR 376)
        | _ ->
            _eRR 375
    
    and _menhir_run_370 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_SIG -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok ->
        let MenhirCell1_SIG (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let _endpos__2_ = _endpos in
        let _ = _menhir_action_137 _endpos__2_ _startpos__1_ in
        _menhir_goto_ignored _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _menhir_s _tok
    
    and _menhir_run_366 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_TYPEABBREV as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let c = _v in
        let _v = _menhir_action_005 c in
        _menhir_goto_abbrevform _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_abbrevform : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_TYPEABBREV as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_abbrevform (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.IO _v_0 ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState367
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState367
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState367
        | _ ->
            _eRR 367
    
    and _menhir_run_363 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_TYPEABBREV, _menhir_box_program) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.IO _v_0 ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState363
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState363
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState363
        | _ ->
            _eRR 363
    
    and _menhir_run_358 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.CONJ ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_CONJ (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState359 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 359)
        | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.TYPE | Elpi_lexer_config.Tokens.UNTYPED ->
            let (_endpos_x_, x) = (_endpos, _v) in
            let _v = _menhir_action_335 x in
            _menhir_goto_separated_nonempty_list_CONJ_constant_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _menhir_s _tok
        | _ ->
            _eRR 358
    
    and _menhir_goto_separated_nonempty_list_CONJ_constant_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState487 ->
            _menhir_run_488 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState461 ->
            _menhir_run_462 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok
        | MenhirState419 ->
            _menhir_run_420 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState406 ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState400 ->
            _menhir_run_401 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState398 ->
            _menhir_run_399 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok
        | MenhirState359 ->
            _menhir_run_360 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState355 ->
            _menhir_run_356 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_488 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_TYPE as 'stack) -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.IO _v_0 ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState488
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState488
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState488
        | _ ->
            _eRR 488
    
    and _menhir_run_462 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_CLOSED -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok ->
        let MenhirCell1_CLOSED (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let _endpos__2_ = _endpos in
        let _ = _menhir_action_143 _endpos__2_ _startpos__1_ in
        _menhir_goto_ignored _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _menhir_s _tok
    
    and _menhir_run_420 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_EXPORTDEF as 'stack) -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.IO _v_0 ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState420
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState420
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState420
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let MenhirCell1_EXPORTDEF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let _endpos__2_ = _endpos in
            let _ = _menhir_action_138 _endpos__2_ _startpos__1_ in
            _menhir_goto_ignored _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _menhir_s _tok
        | _ ->
            _eRR 420
    
    and _menhir_run_407 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_KIND as 'stack) -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.TYPE ->
            _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState407
        | _ ->
            _eRR 407
    
    and _menhir_run_408 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_TYPE (_menhir_stack, _menhir_s, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_ARROW (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState409 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.TYPE ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 409)
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let (_endpos__1_, _startpos__1_) = (_endpos, _startpos) in
            let _v = _menhir_action_178 _endpos__1_ _startpos__1_ in
            _menhir_goto_kind_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s
        | _ ->
            _eRR 408
    
    and _menhir_goto_kind_term : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_program) _menhir_state -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s ->
        match _menhir_s with
        | MenhirState407 ->
            _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v
        | MenhirState409 ->
            _menhir_run_410 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_411 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_KIND, _menhir_box_program) _menhir_cell1_separated_nonempty_list_CONJ_constant_ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v ->
        let MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _, names, _) = _menhir_stack in
        let MenhirCell1_KIND (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_endpos_k_, k) = (_endpos, _v) in
        let _v = _menhir_action_177 _endpos_k_ _startpos__1_ k names in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let t = _v in
        let _v = _menhir_action_109 t in
        _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
    
    and _menhir_run_410 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_TYPE _menhir_cell0_ARROW -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v ->
        let MenhirCell0_ARROW (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_TYPE (_menhir_stack, _menhir_s, _startpos_x_, _endpos_x_) = _menhir_stack in
        let (x, _endpos_t_, t) = ((), _endpos, _v) in
        let _v = _menhir_action_179 _endpos_t_ _endpos_x_ _startpos_x_ t x in
        _menhir_goto_kind_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _v _menhir_s
    
    and _menhir_run_401 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_LOCAL as 'stack) -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.IO _v_0 ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState401
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState401
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState401
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let _ = _menhir_action_323 () in
            _menhir_goto_option_type_term_ _menhir_stack _menhir_lexbuf _menhir_lexer _tok
        | _ ->
            _eRR 401
    
    and _menhir_run_399 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_LOCALKIND -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _tok ->
        let MenhirCell1_LOCALKIND (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let _endpos__2_ = _endpos in
        let _ = _menhir_action_140 _endpos__2_ _startpos__1_ in
        _menhir_goto_ignored _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _menhir_s _tok
    
    and _menhir_run_360 : type  ttv_stack. (ttv_stack, _menhir_box_program) _menhir_cell1_constant _menhir_cell0_CONJ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_CONJ (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_constant (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let (_endpos_xs_, xs) = (_endpos, _v) in
        let _v = _menhir_action_336 x xs in
        _menhir_goto_separated_nonempty_list_CONJ_constant_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_xs_ _v _menhir_s _tok
    
    and _menhir_run_356 : type  ttv_stack. ((ttv_stack, _menhir_box_program) _menhir_cell1_USEONLY as 'stack) -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.IO _v_0 ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState356
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState356
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_separated_nonempty_list_CONJ_constant_ (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState356
        | Elpi_lexer_config.Tokens.FULLSTOP ->
            let MenhirCell1_USEONLY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let _endpos__2_ = _endpos in
            let _ = _menhir_action_141 _endpos__2_ _startpos__1_ in
            _menhir_goto_ignored _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _menhir_s _tok
        | _ ->
            _eRR 356
    
    and _menhir_run_302 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAREN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.COLON ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_303 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.BIND ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_206 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FLOAT _ | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.INTEGER _ | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.LBRACKET | Elpi_lexer_config.Tokens.LCURLY | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.QUOTED _ | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.STRING _ | Elpi_lexer_config.Tokens.UNTYPED | Elpi_lexer_config.Tokens.VDASH ->
            let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_132 _endpos_t_ _startpos_t_ t in
            _menhir_goto_head_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
        | _ ->
            _eRR 302
    
    and _menhir_run_210 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.COLON ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_COLON (_menhir_stack, _startpos) in
            let _menhir_s = MenhirState211 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 211)
        | Elpi_lexer_config.Tokens.BIND ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_BIND (_menhir_stack, _startpos, _endpos) in
            let _menhir_s = MenhirState215 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.STRING _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.SIGMA ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.QUOTED _v ->
                _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.PI ->
                _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LCURLY ->
                _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LBRACKET ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INTEGER _v ->
                _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FLOAT _v ->
                _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
                _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 215)
        | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FLOAT _ | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.INTEGER _ | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.LBRACKET | Elpi_lexer_config.Tokens.LCURLY | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.QUOTED _ | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.STRING _ | Elpi_lexer_config.Tokens.UNTYPED | Elpi_lexer_config.Tokens.VDASH ->
            let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_132 _endpos_t_ _startpos_t_ t in
            _menhir_goto_head_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
        | _ ->
            _eRR 210
    
    and _menhir_run_137 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.COLON ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_COLON (_menhir_stack, _startpos) in
            let _menhir_s = MenhirState138 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Elpi_lexer_config.Tokens.UNTYPED ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REPLACE ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.REMOVE ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NIL ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.NAME ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.LPAREN ->
                _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IO _v ->
                _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.INDEX ->
                _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.IF ->
                _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FUNCTIONAL ->
                _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.FRESHUV ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CUT ->
                _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.CONSTANT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Elpi_lexer_config.Tokens.BEFORE ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Elpi_lexer_config.Tokens.AFTER ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR 138)
        | Elpi_lexer_config.Tokens.BIND ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_206 _menhir_stack _menhir_lexbuf _menhir_lexer
        | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.AS | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.CONJ2 | Elpi_lexer_config.Tokens.CONS | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.DARROW | Elpi_lexer_config.Tokens.DDARROW | Elpi_lexer_config.Tokens.DIV | Elpi_lexer_config.Tokens.EOF | Elpi_lexer_config.Tokens.EQ | Elpi_lexer_config.Tokens.EQ2 | Elpi_lexer_config.Tokens.FAMILY_AND _ | Elpi_lexer_config.Tokens.FAMILY_BTICK _ | Elpi_lexer_config.Tokens.FAMILY_EQ _ | Elpi_lexer_config.Tokens.FAMILY_EXP _ | Elpi_lexer_config.Tokens.FAMILY_GT _ | Elpi_lexer_config.Tokens.FAMILY_LT _ | Elpi_lexer_config.Tokens.FAMILY_MINUS _ | Elpi_lexer_config.Tokens.FAMILY_OR _ | Elpi_lexer_config.Tokens.FAMILY_PLUS _ | Elpi_lexer_config.Tokens.FAMILY_QMARK _ | Elpi_lexer_config.Tokens.FAMILY_SHARP _ | Elpi_lexer_config.Tokens.FAMILY_TICK _ | Elpi_lexer_config.Tokens.FAMILY_TIMES _ | Elpi_lexer_config.Tokens.FLOAT _ | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.IFF | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.INTEGER _ | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.IS | Elpi_lexer_config.Tokens.LBRACKET | Elpi_lexer_config.Tokens.LCURLY | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.MINUS | Elpi_lexer_config.Tokens.MINUSi | Elpi_lexer_config.Tokens.MINUSr | Elpi_lexer_config.Tokens.MINUSs | Elpi_lexer_config.Tokens.MOD | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.OR | Elpi_lexer_config.Tokens.PIPE | Elpi_lexer_config.Tokens.QDASH | Elpi_lexer_config.Tokens.QUOTED _ | Elpi_lexer_config.Tokens.RBRACKET | Elpi_lexer_config.Tokens.RCURLY | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.RPAREN | Elpi_lexer_config.Tokens.SLASH | Elpi_lexer_config.Tokens.STRING _ | Elpi_lexer_config.Tokens.UNTYPED | Elpi_lexer_config.Tokens.VDASH ->
            let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_132 _endpos_t_ _startpos_t_ t in
            _menhir_goto_head_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
        | _ ->
            _eRR 137
    
    and _menhir_run_134 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_t_, _startpos_t_, t) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_132 _endpos_t_ _startpos_t_ t in
        _menhir_goto_head_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_t_ _v _menhir_s _tok
    
    and _menhir_run_104 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_c_, _startpos_c_, c) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_104 _endpos_c_ _startpos_c_ c in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.IO _v_0 ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState102
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState102
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, _v) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState102
        | Elpi_lexer_config.Tokens.BIND | Elpi_lexer_config.Tokens.COLON | Elpi_lexer_config.Tokens.RPAREN ->
            let x = _v in
            let _v = _menhir_action_234 x in
            _menhir_goto_nonempty_list_constant_w_loc_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR 102
    
    and _menhir_goto_nonempty_list_constant_w_loc_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState518 ->
            _menhir_run_519 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState490 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState363 ->
            _menhir_run_364 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState113 ->
            _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState108 ->
            _menhir_run_109 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState102 ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState105 ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState097 ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState101 ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState003 ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_519 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_PI as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.COLON ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState519
        | Elpi_lexer_config.Tokens.BIND ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState519
        | _ ->
            _eRR 519
    
    and _menhir_run_058 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_nonempty_list_constant_w_loc_ as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_COLON (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState058 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 58
    
    and _menhir_run_099 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_nonempty_list_constant_w_loc_ as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_BIND (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState099 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 99
    
    and _menhir_run_491 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_attributes, _menhir_box_program) _menhir_cell1_SIGMA as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_program) _menhir_state -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.COLON ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState491
        | Elpi_lexer_config.Tokens.BIND ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState491
        | _ ->
            _eRR 491
    
    and _menhir_run_364 : type  ttv_stack. (((ttv_stack, _menhir_box_program) _menhir_cell1_TYPEABBREV, _menhir_box_program) _menhir_cell1_LPAREN, _menhir_box_program) _menhir_cell1_constant -> _ -> _ -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.RPAREN ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_constant (_menhir_stack, _, hd, _, _) = _menhir_stack in
            let MenhirCell1_LPAREN (_menhir_stack, _menhir_s, _) = _menhir_stack in
            let args = _v in
            let _v = _menhir_action_006 args hd in
            _menhir_goto_abbrevform _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR 364
    
    and _menhir_run_114 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_PI as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.COLON ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState114
        | Elpi_lexer_config.Tokens.BIND ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState114
        | _ ->
            _eRR 114
    
    and _menhir_run_109 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_SIGMA as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.COLON ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState109
        | Elpi_lexer_config.Tokens.BIND ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState109
        | _ ->
            _eRR 109
    
    and _menhir_run_103 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_constant_w_loc -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_constant_w_loc (_menhir_stack, _menhir_s, x) = _menhir_stack in
        let xs = _v in
        let _v = _menhir_action_235 x xs in
        _menhir_goto_nonempty_list_constant_w_loc_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_098 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_PI as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.COLON ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | Elpi_lexer_config.Tokens.BIND ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState098
        | _ ->
            _eRR 98
    
    and _menhir_run_057 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_SIGMA as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_nonempty_list_constant_w_loc_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.COLON ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Elpi_lexer_config.Tokens.BIND ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | _ ->
            _eRR 57
    
    and _menhir_run_081 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_c_, _startpos_c_, c) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_027 _endpos_c_ _startpos_c_ c in
        _menhir_goto_atype_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_c_ _v _menhir_s _tok
    
    and _menhir_run_066 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.REPLACE ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.REMOVE ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.NIL ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.NAME ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.LPAREN ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.IO _v_0 ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState066
        | Elpi_lexer_config.Tokens.INDEX ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.IF ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.FRESHUV ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.CUT ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.CONSTANT _v_1 ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState066
        | Elpi_lexer_config.Tokens.BEFORE ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.AFTER ->
            let _menhir_stack = MenhirCell1_constant (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState066
        | Elpi_lexer_config.Tokens.ARROW | Elpi_lexer_config.Tokens.BIND | Elpi_lexer_config.Tokens.CONJ | Elpi_lexer_config.Tokens.FULLSTOP | Elpi_lexer_config.Tokens.IO_COLON _ | Elpi_lexer_config.Tokens.RPAREN ->
            let (_endpos_c_, _startpos_c_, c) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_126 _endpos_c_ _startpos_c_ c in
            _menhir_goto_fotype_term _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_c_ _startpos_c_ _v _menhir_s _tok
        | _ ->
            _eRR 66
    
    let _menhir_run_000 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_goal =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _menhir_s = MenhirState000 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.UNTYPED ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.STRING _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.SIGMA ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REPLACE ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.REMOVE ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.QUOTED _v ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.PI ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NIL ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAME ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LPAREN ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_106 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LBRACKET ->
            _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IO _v ->
            _menhir_run_048 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INTEGER _v ->
            _menhir_run_116 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.INDEX ->
            _menhir_run_049 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IF ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FUNCTIONAL ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FRESHUV ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FLOAT _v ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            _menhir_run_125 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.CUT ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CONSTANT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.BEFORE ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR 0
    
    let _menhir_goto_infix_SYMB : type  ttv_stack. ttv_stack -> _ -> _menhir_box_infix_SYMB =
      fun _menhir_stack _v ->
        MenhirBox_infix_SYMB _v
    
    let _menhir_run_314 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_infix_SYMB =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.VDASH ->
            let _v = _menhir_action_176 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.SLASH ->
            let _v = _menhir_action_173 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.QDASH ->
            let _v = _menhir_action_172 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.OR ->
            let _v = _menhir_action_165 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.MOD ->
            let _v = _menhir_action_167 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.MINUSs ->
            let _v = _menhir_action_163 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.MINUSr ->
            let _v = _menhir_action_161 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.MINUSi ->
            let _v = _menhir_action_162 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.MINUS ->
            let _v = _menhir_action_160 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.IS ->
            let _v = _menhir_action_166 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_TIMES _v ->
            let s = _v in
            let _v = _menhir_action_147 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_TICK _v ->
            let s = _v in
            let _v = _menhir_action_157 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_SHARP _v ->
            let s = _v in
            let _v = _menhir_action_155 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_PLUS _v ->
            let s = _v in
            let _v = _menhir_action_146 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_OR _v ->
            let s = _v in
            let _v = _menhir_action_154 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_MINUS _v ->
            let s = _v in
            let _v = _menhir_action_148 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_LT _v ->
            let s = _v in
            let _v = _menhir_action_150 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_GT _v ->
            let s = _v in
            let _v = _menhir_action_151 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_EXP _v ->
            let s = _v in
            let _v = _menhir_action_149 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_EQ _v ->
            let s = _v in
            let _v = _menhir_action_152 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_BTICK _v ->
            let s = _v in
            let _v = _menhir_action_156 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.FAMILY_AND _v ->
            let s = _v in
            let _v = _menhir_action_153 s in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.EQ2 ->
            let _v = _menhir_action_164 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.EQ ->
            let _v = _menhir_action_159 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.DIV ->
            let _v = _menhir_action_168 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.DDARROW ->
            let _v = _menhir_action_171 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.DARROW ->
            let _v = _menhir_action_170 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.CONS ->
            let _v = _menhir_action_158 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.CONJ2 ->
            let _v = _menhir_action_174 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.CONJ ->
            let _v = _menhir_action_175 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | Elpi_lexer_config.Tokens.ARROW ->
            let _v = _menhir_action_169 () in
            _menhir_goto_infix_SYMB _menhir_stack _v
        | _ ->
            _eRR 314
    
    let _menhir_run_347 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_postfix_SYMB =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_QMARK _v ->
            let s = _v in
            let _v = _menhir_action_325 s in
            MenhirBox_postfix_SYMB _v
        | _ ->
            _eRR 347
    
    let _menhir_run_350 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_prefix_SYMB =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.FAMILY_TILDE _v ->
            let s = _v in
            let _v = _menhir_action_330 s in
            MenhirBox_prefix_SYMB _v
        | _ ->
            _eRR 350
    
    let _menhir_run_353 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_program =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _menhir_s = MenhirState353 in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Elpi_lexer_config.Tokens.USE_SIG ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.USEONLY ->
            _menhir_run_355 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.TYPEABBREV ->
            _menhir_run_361 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SIG ->
            _menhir_run_369 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.SHORTEN ->
            _menhir_run_371 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.RCURLY ->
            _menhir_run_388 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.NAMESPACE ->
            _menhir_run_389 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MODULE ->
            _menhir_run_392 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.MACRO ->
            _menhir_run_394 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LOCALKIND ->
            _menhir_run_398 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LOCAL ->
            _menhir_run_400 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.LCURLY ->
            _menhir_run_405 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.KIND ->
            _menhir_run_406 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.IMPORT ->
            _menhir_run_412 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.FIXITY _v ->
            _menhir_run_413 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Elpi_lexer_config.Tokens.EXTERNAL ->
            _menhir_run_418 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.EXPORTDEF ->
            _menhir_run_419 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.EOF ->
            _menhir_run_422 _menhir_stack _menhir_s
        | Elpi_lexer_config.Tokens.CONSTRAINT ->
            _menhir_run_423 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.COLON ->
            _menhir_run_431 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.CLOSED ->
            _menhir_run_461 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.ACCUM_SIG ->
            _menhir_run_463 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.ACCUMULATE ->
            _menhir_run_464 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Elpi_lexer_config.Tokens.AFTER | Elpi_lexer_config.Tokens.BEFORE | Elpi_lexer_config.Tokens.CONSTANT _ | Elpi_lexer_config.Tokens.CUT | Elpi_lexer_config.Tokens.FAMILY_TILDE _ | Elpi_lexer_config.Tokens.FRESHUV | Elpi_lexer_config.Tokens.FUNC | Elpi_lexer_config.Tokens.FUNCTIONAL | Elpi_lexer_config.Tokens.IF | Elpi_lexer_config.Tokens.INDEX | Elpi_lexer_config.Tokens.IO _ | Elpi_lexer_config.Tokens.LPAREN | Elpi_lexer_config.Tokens.NAME | Elpi_lexer_config.Tokens.NIL | Elpi_lexer_config.Tokens.PI | Elpi_lexer_config.Tokens.PRED | Elpi_lexer_config.Tokens.REMOVE | Elpi_lexer_config.Tokens.REPLACE | Elpi_lexer_config.Tokens.RULE | Elpi_lexer_config.Tokens.SIGMA | Elpi_lexer_config.Tokens.TYPE | Elpi_lexer_config.Tokens.UNTYPED ->
            let _v = _menhir_action_024 () in
            _menhir_goto_attributes _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _endpos _v _menhir_s _tok
        | _ ->
            _eRR 353
    
  end
  
  let program =
    fun _menhir_lexer _menhir_lexbuf ->
      let _menhir_stack = () in
      let MenhirBox_program v = _menhir_run_353 _menhir_stack _menhir_lexbuf _menhir_lexer in
      v
  
  let prefix_SYMB =
    fun _menhir_lexer _menhir_lexbuf ->
      let _menhir_stack = () in
      let MenhirBox_prefix_SYMB v = _menhir_run_350 _menhir_stack _menhir_lexbuf _menhir_lexer in
      v
  
  let postfix_SYMB =
    fun _menhir_lexer _menhir_lexbuf ->
      let _menhir_stack = () in
      let MenhirBox_postfix_SYMB v = _menhir_run_347 _menhir_stack _menhir_lexbuf _menhir_lexer in
      v
  
  let infix_SYMB =
    fun _menhir_lexer _menhir_lexbuf ->
      let _menhir_stack = () in
      let MenhirBox_infix_SYMB v = _menhir_run_314 _menhir_stack _menhir_lexbuf _menhir_lexer in
      v
  
  let goal =
    fun _menhir_lexer _menhir_lexbuf ->
      let _menhir_stack = () in
      let MenhirBox_goal v = _menhir_run_000 _menhir_stack _menhir_lexbuf _menhir_lexer in
      v
  
end
